# include "StdAfx.h"
# include "Resource.h"
# include "ChangeType.h"
# include "Types.h"
# include "Load.h"
# include "Selection.h"
# include "UndoRedo.h"
# include "Label.h"
# include "Globals.h"

class CChangeTypeAction : public CEditAction {
  public:
	CChangeTypeAction (layer_t * _pLayer, rgns_t::iterator _iRgn, USHORT _ushType) :
		m_pLayer (_pLayer),
		m_pRgn (& * _iRgn),
		m_ushPrevType (_iRgn->ushType),
		m_ushNewType (_ushType)
	{}

	virtual void GetName (std::string & _str) const;
	virtual bool CanUndo () const;
	virtual bool Do ();
	virtual void Undo ();
	virtual void Redo ();

  private:
	layer_t * m_pLayer;
	rgn_t *   m_pRgn;

	USHORT m_ushNewType;
	USHORT m_ushPrevType;
};

void CChangeTypeAction::GetName (std::string & _str) const {
	CCritSec::CLock lock (g_cs);

	// TODO: apply initial type.
	_str.append ("change type [");
	GetHintText (* m_pRgn, _str);
	_str.append ("]");
}

bool CChangeTypeAction::CanUndo () const {return true;}

bool CChangeTypeAction::Do () {
	CCritSec::CLock lock (g_cs);

	// Change type.
	assert (m_pLayer);
	m_pRgn->ushType = m_ushNewType;
	m_pRgn->pTypeDesc = GetTypeDesc (m_pRgn->ushType, m_pRgn->ushKind);

	return true;
}

void CChangeTypeAction::Undo () {
	CCritSec::CLock lock (g_cs);

	// Revert type.
	assert (m_pLayer);
	m_pRgn->ushType = m_ushPrevType;
	m_pRgn->pTypeDesc = GetTypeDesc (m_pRgn->ushType, m_pRgn->ushKind);
}

void CChangeTypeAction::Redo () {
	CChangeTypeAction::Do ();
}

/////////////////////////////////////////////////////////////////

int CALLBACK ChangeTypeProc (HWND _hWnd, UINT _uMsg, WPARAM _wParam, LPARAM);

void ShowChangeType (HWND _hWnd) {
	if (g_bLoading)
		return;

	if (g_Selection.size () != 1 || ! g_Selection [0].pLayer)
		return;

	::DialogBox (g_hInst, MAKEINTRESOURCE (IDD_CHANGE_TYPE), _hWnd, & ChangeTypeProc);
}

static
void AddColumn (HWND _hTypes, int _i, const char * _strText, int _w) {
	LVCOLUMN clm;
	clm.mask = LVCF_FMT | LVCF_TEXT | LVCF_WIDTH;
	clm.fmt = LVCFMT_LEFT;
	clm.cx = _w;
	clm.pszText = const_cast<char *> (_strText);
	clm.cchTextMax = strlen (_strText);
	ListView_InsertColumn (_hTypes, _i, & clm);
}

static HWND g_hTypes;
static int g_iSortColumn = -1;
static bool g_bAccent = false;

static
int CALLBACK CompareItems (LPARAM _lParam1, LPARAM _lParam2, LPARAM) {
	char str1 [64], str2 [64];
	ListView_GetItemText (g_hTypes, _lParam1, g_iSortColumn, str1, 64); 
	ListView_GetItemText (g_hTypes, _lParam2, g_iSortColumn, str2, 64);

	if (g_iSortColumn == 4 || g_iSortColumn == 5) {
		// Compare as integers.
		return g_bAccent ? atoi (str1) - atoi (str2) : atoi (str2) - atoi (str1);
	} else
		// Compare as strings.
		return g_bAccent ? strcmp (str1, str2) : strcmp (str2, str1);
}

static
void ApplyChangeTypes () {
	const int iSel = ListView_GetSelectionMark (g_hTypes);
	if (iSel == -1)
		return;
	LVITEM lvi;
	lvi.mask = LVIF_PARAM;
	lvi.iItem = iSel;
	lvi.iSubItem = 0;
	ListView_GetItem (g_hTypes, & lvi);
	const USHORT ushType = lvi.lParam;

	if (g_Selection [0].pLayer)
		PerformEditAction (new CChangeTypeAction (g_Selection [0].pLayer, g_Selection [0].iRgn, ushType));
} 

int CALLBACK ChangeTypeProc (HWND _hWnd, UINT _uMsg, WPARAM _wParam, LPARAM _lParam) {
	switch (_uMsg) {
		case WM_INITDIALOG: {
			char strBuf [64];
			
			// Prepare list view control for types.			
			g_hTypes = ::GetDlgItem (_hWnd, IDC_TYPES);

			ListView_SetExtendedListViewStyleEx (g_hTypes, LVS_EX_FULLROWSELECT, LVS_EX_FULLROWSELECT);

			AddColumn (g_hTypes, 0, "Code", 50);
			AddColumn (g_hTypes, 1, "Description", 200);

			LVITEM lvi;
			lvi.mask = LVIF_TEXT | LVIF_STATE;// | LVIF_PARAM;
			lvi.state = 0;
			lvi.stateMask = 0;
			lvi.iItem = 0;
		    lvi.pszText = strBuf;
			lvi.cchTextMax = sizeof (strBuf)/sizeof (strBuf [0]);

			const USHORT ushType = g_Selection [0].iRgn->ushType;
			const USHORT ushKind = g_Selection [0].iRgn->ushKind;
			assert (ushKind < 4);

			// For each type.
			// NOTE: omit 0-th type ("unknown").
			for (size_t c = 1; c < g_cTypes; ++ c) {
				if (g_types [c].ushKind != ushKind)
					continue;

				// Code
				::sprintf (strBuf, "0x%04x", g_types [c].ushType);
				lvi.mask = LVIF_TEXT | LVIF_STATE | LVIF_PARAM;
				lvi.iSubItem = 0;
				lvi.lParam = g_types [c].ushType;
				ListView_InsertItem (g_hTypes, & lvi);

				// Description
				::sprintf (strBuf, "%s", g_types [c].strTypeName);
				lvi.mask = LVIF_TEXT | LVIF_STATE;
				lvi.iSubItem = 1;
				ListView_SetItem (g_hTypes, & lvi);

				if (g_types [c].ushType == ushType) {
					ListView_SetItemState (g_hTypes, lvi.iItem, LVIS_SELECTED, LVIS_SELECTED);
					ListView_SetSelectionMark (g_hTypes, lvi.iItem);
				}

				++ lvi.iItem;
			}

			return TRUE;
		}

		case WM_NOTIFY: {
			const NMHDR * const pInfo = reinterpret_cast<const NMHDR *> (_lParam);
			switch (pInfo->code) {
				case LVN_COLUMNCLICK: {
					const NMLISTVIEW * const pInfo2 = reinterpret_cast<const NMLISTVIEW *> (_lParam);

					if (g_iSortColumn != pInfo2->iSubItem)
						g_iSortColumn = pInfo2->iSubItem;
					else
						g_bAccent = ! g_bAccent;
					
					ListView_SortItemsEx (g_hTypes, CompareItems, 0);

					break;
				}

				case NM_CLICK:
				case LVN_KEYDOWN:
					::EnableWindow (::GetDlgItem (_hWnd, IDOK), TRUE);
					::EnableWindow (::GetDlgItem (_hWnd, IDAPPLY), TRUE);
					break;
			}
			break;
		}

		case WM_COMMAND:
			switch (LOWORD (_wParam)) {
				case IDAPPLY:
					ApplyChangeTypes ();
					break;

				case IDOK:
					ApplyChangeTypes ();
				case IDCANCEL:
					// On closing dialog.
					::EndDialog (_hWnd, 0);
					return TRUE;
			}
	}

	return FALSE;
}
