# include "StdAfx.h"
# include "CombinedAction.h"
# include "UndoRedo.h"

class CBookmarkAction : public CEditAction {
  public:
	CBookmarkAction (int _iDepth) : m_iDepth (_iDepth) {}

	virtual void GetName (std::string & _str) const {_str = "------";}
	virtual bool CanUndo () const {return true;}
	virtual bool Do () {return true;}
	virtual void Undo () {}
	virtual void Redo () {}

	virtual int GetDepthLevel () const {return m_iDepth;}

  private:
	const int m_iDepth;
};

class CCombinedAction : public CEditAction {
  public:
	CCombinedAction (int _iDepth, const char * _strName) :
		m_iDepth (_iDepth),
		m_strName (_strName)
	{}
	~CCombinedAction ();

	virtual void GetName (std::string & _str) const {_str = m_strName;}
	virtual bool CanUndo () const {return true;}
	virtual bool Do () {return true;}
	virtual void Undo ();
	virtual void Redo ();

	virtual int GetDepthLevel () const {return -m_iDepth;}

	void AddAction (CEditAction * _pAction) {m_actionsUndo.push_back (_pAction);}

  private:
	const int m_iDepth;
	const std::string m_strName;

	actions_t m_actionsUndo;
	actions_t m_actionsRedo;
};

CCombinedAction::~CCombinedAction () {
	actions_t::iterator i;
	for (i = m_actionsUndo.begin (); i != m_actionsUndo.end (); ++ i)
		delete * i;

	for (i = m_actionsRedo.begin (); i != m_actionsRedo.end (); ++ i)
		delete * i;
}

void CCombinedAction::Undo () {
	while (! m_actionsUndo.empty ()) {
		CEditAction * const pAction = m_actionsUndo.back ();
		m_actionsUndo.pop_back ();

		pAction->Undo ();

		m_actionsRedo.push_back (pAction);
	}
}

void CCombinedAction::Redo () {
	while (! m_actionsRedo.empty ()) {
		CEditAction * const pAction = m_actionsRedo.back ();
		m_actionsRedo.pop_back ();

		pAction->Redo ();

		m_actionsUndo.push_back (pAction);
	}
}

///////////////////////////////////

static int g_iDepth;

int BeginCombinedAction () {
	++ g_iDepth;
	PerformEditAction (new CBookmarkAction (g_iDepth));
	return g_iDepth;
}

void EndCombinedAction (int _iDepth, const char * _strName) {
	// Find the bookmark.
	actions_t::iterator i = g_Undo.end ();
	do {
		-- i;
		if ((* i)->GetDepthLevel () == g_iDepth)
			break;
	} while (i != g_Undo.begin ());

	if ((* i)->GetDepthLevel () != g_iDepth) {
		// The bookmark is not found.
		assert (false);
		return;
	}

	// Remove bookmark.
	const actions_t::iterator iBookmark = i;
	++ i;
	g_Undo.erase (iBookmark);

	CCombinedAction * pCA = new CCombinedAction (g_iDepth, _strName);

	// Move actions from g_Undo (from bookmark till last one) to combined action.
	while (i != g_Undo.end ()) {
		actions_t::iterator i_next = i;
		++ i_next;

		pCA->AddAction (* i);
		g_Undo.erase (i);

		i = i_next;
	}

	PerformEditAction (pCA);
	
	-- g_iDepth;
}
