# include "stdafx.h"
# include "Datum.h"
# include "Globals.h"

//
// Based on the GOST R 51794-2001 of Russian Federation
//

	// Arc seconds per radian.
const float c_ro  = 206264.8062f;

	// Krassovsky ellipsoid (Pulkovo 1942)
const float c_aP  = 6378245.f; //  
const float c_alP = 1/298.3f; // 
const float c_e2P = 2*c_alP - c_alP*c_alP; //  

	// GRS80 ellipsoid (WGS84)
const float c_aW  = 6378137.f; //  
const float c_alW = 1/298.257223563f; // 
const float c_e2W = 2*c_alW - c_alW*c_alW; //  

	//     
const float c_a   = (c_aP + c_aW)/2;
const float c_e2  = (c_e2P + c_e2W)/2;
const float c_da  = c_aW - c_aP;
const float c_de2 = c_e2W - c_e2P;

	//   ,  
const float c_dx  = 23.92f;
const float c_dy  = -141.27f;
const float c_dz  = -80.9f;
	//   ,  
const float c_wx  = 0;
const float c_wy  = -0.35f;
const float c_wz  = -0.82f;
	//   
const float c_ms  = -0.12e-6f;

void ToWGS84 (datum_t _Datum, point_t & _point, float _fHeight) {
	if (_Datum != dtPulkovo42)
		return;

	const float & Ld = _point.x;
	const float & Bd = _point.y;

	const float B = Bd*c_PI/180;
	const float L = Ld*c_PI/180;

	const float fSinB  = ::sin (B);
	const float fCosB  = ::cos (B);
	const float fCos2B = ::cos (2*B);
	const float fTanB  = fSinB/fCosB;
	const float fSinL  = ::sin (L);
	const float fCosL  = ::cos (L);

	const float fNu = 1 - c_e2*fSinB*fSinB;
	const float fSqrtNu = ::sqrt (fNu);

	const float M = c_a*(1 - c_e2)/(fNu*fSqrtNu);
	const float N = c_a/fSqrtNu;

	//
	// Latitude.
	//

	const float dB = c_ro/(M + _fHeight)*(N/c_a*c_e2*fSinB*fCosB*c_da
		+ (N*N/(c_a*c_a) + 1)*N*fSinB*fCosB*c_de2/2
		- (c_dx*fCosL + c_dy*fSinL)*fSinB + c_dz*fCosB)
		- c_wx*fSinL*(1 + c_e2*fCos2B)
		+ c_wy*fCosL*(1 + c_e2*fCos2B)
		- c_ro*c_ms * c_e2 * fSinB * fCosB;

	const float fWGS84Lat = Bd + dB/3600;

	//
	// Longitude.
	//

	const float dL = c_ro/((N + _fHeight)*fCosB)*(-c_dx*fSinL + c_dy*fCosL)
		+ fTanB*(1 - c_e2)*(c_wx*fCosL + c_wy*fSinL) - c_wz;

	const float fWGS84Long = Ld + dL/3600;

	//
	// Height.
	//

	const float dH = -c_a/N*c_da + N*fSinB*fSinB*c_de2/2
		+ (c_dx*fCosL + c_dy*fSinL)*fCosB + c_dz*fSinB
		- N*c_e2*fSinB*fCosB*(c_wx/c_ro*fSinL - c_wy/c_ro*fCosL)
		+ (c_a*c_a/N + _fHeight)*c_ms;

	const float fWGS84Alt = _fHeight + dH;

	_point.x = fWGS84Long;
	_point.y = fWGS84Lat;
}

void FromWGS84 (datum_t _Datum, point_t & _point, float _fHeight) {

}
