# include "stdafx.h"
# include "UndoRedo.h"
# include "Map.h"
# include "DeleteElement.h"
# include "Label.h"
# include "Selection.h"
# include "Globals.h"

class CDeleteElementAction : public CEditAction {
  public:
	CDeleteElementAction (layer_t * _pLayer, rgns_t::iterator _iRgn, size_t _cElementOrOrigin) :
		m_pLayer (_pLayer),
		m_pRgn (& * _iRgn),
		m_cElementOrOrigin (_cElementOrOrigin)
	{}

	virtual void GetName (std::string & _str) const;
	virtual bool CanUndo () const;
	virtual bool Do ();
	virtual void Undo ();
	virtual void Redo ();

  private:
	layer_t * const m_pLayer;
	rgn_t *   const m_pRgn;
	size_t    const m_cElementOrOrigin;
};

void DeleteElement (layer_t * _pLayer, rgns_t::iterator _iRgn, size_t _cElementOrOrigin) {
	PerformEditAction (new CDeleteElementAction (_pLayer, _iRgn, _cElementOrOrigin));
}

void CDeleteElementAction::GetName (std::string & _str) const {
	CCritSec::CLock lock (g_cs);

	assert (m_pRgn);
	_str.append ("delete element of [");
	GetHintText (* m_pRgn, _str);
	_str.append ("]");
}

bool CDeleteElementAction::CanUndo () const {return true;}

bool CDeleteElementAction::Do () {
	CCritSec::CLock lock (g_cs);

	assert (m_pLayer);
	if (m_pRgn->origins.empty ()) {
		assert (m_cElementOrOrigin < m_pRgn->elements.size ());
		m_pRgn->elements [m_cElementOrOrigin].bDeleted = true;
	} else {
		assert (m_cElementOrOrigin < m_pRgn->origins.size ());
		m_pRgn->origins [m_cElementOrOrigin].bDeleted = true;
	}

	return true;
}

void CDeleteElementAction::Undo () {
	CCritSec::CLock lock (g_cs);

	assert (m_pLayer);
	if (m_pRgn->origins.empty ()) {
		assert (m_cElementOrOrigin < m_pRgn->elements.size ());
		m_pRgn->elements [m_cElementOrOrigin].bDeleted = false;
	} else {
		assert (m_cElementOrOrigin < m_pRgn->origins.size ());
		m_pRgn->origins [m_cElementOrOrigin].bDeleted = false;
	}
}

void CDeleteElementAction::Redo () {
	CDeleteElementAction::Do ();
}
