# include "stdafx.h"
# include "UndoRedo.h"
# include "Map.h"
# include "Selection.h"
# include "Globals.h"

class CDeleteImageAction : public CEditAction {
  public:
	CDeleteImageAction (image_t * _pImage) : m_pImage (_pImage) {}

	virtual void GetName (std::string & _str) const;
	virtual bool CanUndo () const;
	virtual bool Do ();
	virtual void Undo ();
	virtual void Redo ();

  private:
	image_t * const m_pImage;
};

void DeleteImage (image_t * _pImage) {
	PerformEditAction (new CDeleteImageAction (_pImage));
}

void CDeleteImageAction::GetName (std::string & _str) const {
	CCritSec::CLock lock (g_cs);

	assert (m_pImage);
	_str.append ("delete [");
	_str.append (m_pImage->strFileName);
	_str.append ("]");
}

bool CDeleteImageAction::CanUndo () const {return true;}

bool CDeleteImageAction::Do () {
	CCritSec::CLock lock (g_cs);

	assert (m_pImage);
	m_pImage->bDeleted = true;

	return true;
}

void CDeleteImageAction::Undo () {
	CCritSec::CLock lock (g_cs);

	assert (m_pImage);
	m_pImage->bDeleted = false;
}

void CDeleteImageAction::Redo () {
	CDeleteImageAction::Do ();
}

