# include "stdafx.h"
# include "UndoRedo.h"
# include "Map.h"
# include "DeleteRegion.h"
# include "Label.h"
# include "Selection.h"
# include "Globals.h"

class CDeleteRegionAction : public CEditAction {
  public:
	CDeleteRegionAction (layer_t * _pLayer, rgns_t::iterator _iRgn) :
		m_pLayer (_pLayer),
		m_pRgn (& * _iRgn)
	{}

	virtual void GetName (std::string & _str) const;
	virtual bool CanUndo () const;
	virtual bool Do ();
	virtual void Undo ();
	virtual void Redo ();

  private:
	layer_t * const m_pLayer;
	rgn_t *   const m_pRgn;
};

void DeleteRegion (layer_t * _pLayer, rgns_t::iterator _iRgn) {
	PerformEditAction (new CDeleteRegionAction (_pLayer, _iRgn));
}

void CDeleteRegionAction::GetName (std::string & _str) const {
	CCritSec::CLock lock (g_cs);

	assert (m_pRgn);
	_str.append ("delete [");
	GetHintText (* m_pRgn, _str);
	_str.append ("]");
}

bool CDeleteRegionAction::CanUndo () const {return true;}

bool CDeleteRegionAction::Do () {
	CCritSec::CLock lock (g_cs);

	assert (m_pLayer);
	++ m_pLayer->cDeletedRgns;
	m_pRgn->dwFlags |= rgn_t::fDeleted;

	return true;
}

void CDeleteRegionAction::Undo () {
	CCritSec::CLock lock (g_cs);

	assert (m_pLayer);
	m_pRgn->dwFlags &= ~rgn_t::fDeleted;
	-- m_pLayer->cDeletedRgns;
}

void CDeleteRegionAction::Redo () {
	CDeleteRegionAction::Do ();
}

