# include "stdafx.h"
# include "Map.h"
# include "Globals.h"

void image_t::MakeSmall () {
	assert (hSmallBitmap == NULL && hSmallMemDC == NULL);

	lSmallWidth  = lWidth/2;
	lSmallHeight = lHeight/2;

	hSmallBitmap = ::CreateCompatibleBitmap (hMemDC, lSmallWidth, lSmallHeight);
	hSmallMemDC  = ::CreateCompatibleDC (hMemDC);
	::SelectObject (hSmallMemDC, hSmallBitmap);

	::SetStretchBltMode (hSmallMemDC, HALFTONE);
	::StretchBlt (hSmallMemDC, 0, 0, lSmallWidth, lSmallHeight, hMemDC, 0, 0, lWidth, lHeight, SRCCOPY);
}

void image_t::clear () {
	if (hMemDC) {
		::DeleteDC (hMemDC);
		hMemDC = NULL;
	}
	if (hBitmap) {
		::DeleteObject (hBitmap);
		hBitmap = NULL;
	}
	if (hSmallMemDC) {
		::DeleteDC (hSmallMemDC);
		hSmallMemDC = NULL;
	}
	if (hSmallBitmap) {
		::DeleteObject (hSmallBitmap);
		hSmallBitmap = NULL;
	}

	rect = rect_t ();
	lWidth = 0;
	lHeight = 0;
	lSmallWidth = 0;
	lSmallHeight = 0;

	strFileName = "";
}

void image_t::swap (image_t & _other) {
	rect.swap (_other.rect);

	std::swap (hBitmap,      _other.hBitmap);
	std::swap (hMemDC,       _other.hMemDC);
	std::swap (lWidth,       _other.lWidth);
	std::swap (lHeight,      _other.lHeight);

	std::swap (hSmallBitmap, _other.hSmallBitmap);
	std::swap (hSmallMemDC,  _other.hSmallMemDC);
	std::swap (lSmallWidth,  _other.lSmallWidth);
	std::swap (lSmallHeight, _other.lSmallHeight);

	std::swap (strFileName,  _other.strFileName);
	std::swap (bOverlay,     _other.bOverlay);
	std::swap (bDeleted,     _other.bDeleted);
}

void JoinImages (images_t & _images1, images_t & _images2) {
	if (_images1.empty ())
		_images1.swap (_images2);
	else {
		for (images_t::iterator iImage = _images2.begin (); iImage != _images2.end (); ++ iImage) {
			if (iImage->bOverlay || iImage->bDeleted)
				continue;

			// Move new images to the end of the list.
			images_t::iterator iLast = _images1.insert (_images1.end (), image_t ());
			iLast->swap (* iImage);
			image_t & image1 = * iLast;

			// Create overlay images.
			for (images_t::const_iterator iImage2 = _images1.begin (); iImage2 != iLast; ++ iImage2) {
				const image_t & image2 = * iImage2;
				if (image2.bDeleted)
					continue;

				if (image1.rect.Intersects (image2.rect)) {
					image_t overlay;
					overlay.bOverlay = true;

					rect_t::Intersection (overlay.rect, image1.rect, image2.rect);
					assert (! overlay.rect.Invalid ());

					// Coordinates of overlaped part inside image1.
					const long lX10 = fast_floor ((overlay.rect.x0 - image1.rect.x0)/image1.rect.Width ()*image1.lWidth + .5);
					const long lX11 = fast_floor ((overlay.rect.x1 - image1.rect.x0)/image1.rect.Width ()*image1.lWidth + .5);
					const long lY10 = fast_floor ((image1.rect.y1 - overlay.rect.y1)/image1.rect.Height ()*image1.lHeight + .5);
					const long lY11 = fast_floor ((image1.rect.y1 - overlay.rect.y0)/image1.rect.Height ()*image1.lHeight + .5);

					// Coordinates of overlaped part inside image2.
					const long lX20 = fast_floor ((overlay.rect.x0 - image2.rect.x0)/image2.rect.Width ()*image2.lWidth + .5);
					const long lX21 = fast_floor ((overlay.rect.x1 - image2.rect.x0)/image2.rect.Width ()*image2.lWidth + .5);
					const long lY20 = fast_floor ((image2.rect.y1 - overlay.rect.y1)/image2.rect.Height ()*image2.lHeight + .5);
					const long lY21 = fast_floor ((image2.rect.y1 - overlay.rect.y0)/image2.rect.Height ()*image2.lHeight + .5);

					// Create bitmap for the overlay.
					overlay.lWidth  = lX11 - lX10 > lX21 - lX20 ? lX11 - lX10 : lX21 - lX20;
					overlay.lHeight = lY11 - lY10 > lY21 - lY20 ? lY11 - lY10 : lY21 - lY20;
					overlay.hBitmap = ::CreateCompatibleBitmap (image1.hMemDC, overlay.lWidth, overlay.lHeight);
					overlay.hMemDC = ::CreateCompatibleDC (image1.hMemDC);
					::SelectObject (overlay.hMemDC, overlay.hBitmap);

					// Render new image's part.
					::SetStretchBltMode (overlay.hMemDC, HALFTONE);
					::StretchBlt (overlay.hMemDC, 0, 0, overlay.lWidth, overlay.lHeight, image1.hMemDC, lX10, lY10, lX11 - lX10, lY11 - lY10, SRCCOPY);
				
					// Render old image's part using alpha blending.
					if (g_fpAlphaBlend) {
						::SetStretchBltMode (overlay.hMemDC, COLORONCOLOR);
						BLENDFUNCTION bf = {AC_SRC_OVER, 0, 96, 0};
						g_fpAlphaBlend (overlay.hMemDC, 0, 0, overlay.lWidth, overlay.lHeight, image2.hMemDC, lX20, lY20, lX21 - lX20, lY21 - lY20, bf);
					}

					overlay.MakeSmall ();

					_images1.push_back (image_t ());
					_images1.back ().swap (overlay);
				}
			}
		}

		_images2.clear ();
	}
}