# include "stdafx.h"
# include "Resource.h"
# include "Levels.h"
# include "Status.h"
# include "Scale.h"
# include "Map.h"
# include "MapView.h"
# include "Menu.h"
# include "Globals.h"

static size_t g_cVisibleLevel = -1;

const float fScales [9] = {
	0,
		// zoom = 0
	c_Scale_1_2km,
		// zoom = 1
	c_Scale_3km,
		// zoom = 2
	c_Scale_8km,
		// zoom = 3
	c_Scale_12km,
		// zoom = 4
	c_Scale_30km,
		// zoom = 5
	c_Scale_50km,
		// zoom = 6
	c_Scale_120km,
		// zoom = 7
	c_Scale_300km,
};

size_t GetZoomLevel (float _fScale) {
	if (g_cVisibleLevel != -1)
		return g_cVisibleLevel;

	CCritSec::CLock lock (g_cs);

	if (g_map.Zoom.size () <= 1)
		return -1;

	// Find zoom level for then given scale.
	for (size_t cZoom = 1; cZoom < 9; ++ cZoom)
		if (_fScale <= fScales [cZoom])
			break;
	-- cZoom;

	// Find map level for the given scale.
	for (size_t cLevel = 0; cLevel < g_map.Zoom.size () - 1; ++ cLevel)
		if (g_map.Zoom [cLevel] == cZoom)
			return cLevel;

	if (cZoom < g_map.Zoom [0])
		// NOTE: Show most detailed level up to 0 to avoid empty map.
		return 0;
	if (cZoom > g_map.Zoom [g_map.Zoom.size () - 2])
		// NOTE: Show less detailed level up to infinite to avoid empty map.
		return g_map.Zoom.size () - 2;

	return -1;
}

void GetZoomLevel (float _fScale, char * _strOut) {
	if (g_cVisibleLevel != -1) {
		::sprintf (_strOut, " *Level%d", g_cVisibleLevel);
		return;
	}

	CCritSec::CLock lock (g_cs);

	if (g_map.Zoom.size () <= 1) {
		::sprintf (_strOut, " No data");
		return;
	}

	// Find zoom level for then given scale.
	for (size_t cZoom = 1; cZoom < 9; ++ cZoom)
		if (_fScale <= fScales [cZoom])
			break;
	-- cZoom;

	// Find map level for the given scale.
	for (size_t cLevel = 0; cLevel < g_map.Zoom.size () - 1; ++ cLevel)
		if (g_map.Zoom [cLevel] == cZoom) {
			::sprintf (_strOut, " =Level%d", cLevel);
			return;
		}

	if (cZoom < g_map.Zoom [0]) {
		// NOTE: Show most detailed level up to 0 to avoid empty map.
		::sprintf (_strOut, "<Level0");
		return;
	}
	if (cZoom > g_map.Zoom [g_map.Zoom.size () - 2]) {
		// NOTE: Show less detailed level up to infinite to avoid empty map.
		::sprintf (_strOut, ">Level%d", g_map.Zoom.size () - 2);
		return;
	}

	::sprintf (_strOut, " No data");
}

bool LevelIsVisible (size_t _cLevel) {
	if (g_cVisibleLevel != -1)
		return _cLevel == g_cVisibleLevel;

	CCritSec::CLock lock (g_cs);

	if (g_map.Zoom.empty ())
		return true;

	if (_cLevel >= g_map.Zoom.size ())
		return false;

	const size_t cZoom = g_map.Zoom [_cLevel];
	assert (cZoom <= 7);

	if (_cLevel >= g_map.Zoom.size () - 2)
		// NOTE: Show less detailed level up to infinite to avoid empty map.
		return fScales [cZoom] < g_fScale;
	else if (_cLevel == 0)
		// NOTE: Show most detailed level up to 0 to avoid empty map.
		return g_fScale <= fScales [cZoom + 1];

	return fScales [cZoom] < g_fScale && g_fScale <= fScales [cZoom + 1];
}

void SetVisibleLevel (size_t _cLevel) {
	if (g_cVisibleLevel == _cLevel)
		return;

	size_t cLevels;
	{
		CCritSec::CLock lock (g_cs);
		cLevels = g_map.Levels.size ();
	}

	if (_cLevel != -1 && _cLevel >= cLevels) {
		MakeLevelsMenu ();
		UpdateMenu ();
		return;
	}

	g_cVisibleLevel = _cLevel;

	Refresh ();
	UpdateMenu ();
	SetScaleStatus (g_fScale);
}

void MakeLevelsMenu () {
	const HMENU hMenu = ::GetMenu (g_hWnd);
	const HMENU hViewMenu = ::GetSubMenu (hMenu, 2);
	const HMENU hmenuLevels = ::GetSubMenu (hViewMenu, 6);

	size_t cLevels;
	{
		CCritSec::CLock lock (g_cs);
		cLevels = g_map.Levels.size ();
	}

	for (size_t cLevel = 0; cLevel < 100; ++ cLevel)
		::RemoveMenu (hmenuLevels, 2300 + cLevel, MF_BYCOMMAND);
	for (cLevel = 0; cLevel < cLevels; ++ cLevel) {
		char strText [32];
		::sprintf (strText, "Level&%d", cLevel);
		::AppendMenu (hmenuLevels, MF_STRING, 2300 + cLevel, strText);
	}
	::EnableMenuItem (hmenuLevels, 2300 + cLevels - 1, MF_BYCOMMAND | MF_GRAYED); 
	::CheckMenuRadioItem (hMenu, ID_VIEW_ALL_LEVELS, ID_VIEW_ALL_LEVELS, ID_VIEW_ALL_LEVELS, MF_BYCOMMAND);
}

void UpdateLevelsMenu () {
	const HMENU hMenu = ::GetMenu (g_hWnd);

	if (g_cVisibleLevel == -1) {
		for (size_t cLevel = 0; cLevel < 100; ++ cLevel)
			::CheckMenuItem (hMenu, 2300 + cLevel, MF_BYCOMMAND | MF_UNCHECKED);
		::CheckMenuRadioItem (hMenu, ID_VIEW_ALL_LEVELS, ID_VIEW_ALL_LEVELS, ID_VIEW_ALL_LEVELS, MF_BYCOMMAND);
	} else {
		::CheckMenuItem (hMenu, ID_VIEW_ALL_LEVELS, MF_BYCOMMAND | MF_UNCHECKED);
		::CheckMenuRadioItem (hMenu, 2300, 2399, 2300 + g_cVisibleLevel, MF_BYCOMMAND);
	}
}