# include "stdafx.h"
# include "Resource.h"
# include "Menu.h"
# include "Map.h"
# include "Scale.h"
# include "Levels.h"
# include "NewMap.h"
# include "Load.h"
# include "Log.h"
# include "Status.h"
# include "Position.h"
# include "AddMap.h"
# include "RecentFiles.h"
# include "MapView.h"
# include "MouseMode.h"
# include "Version.h"
# include "Globals.h"

	// Find end of line.
const char * FindNextLine (const char * _pBegin, const char * _pEnd) {
	for (const char * p = _pBegin; p < _pEnd; ++ p)
		if (* p == '\n' || * p == '\r')
			return p + 1 < _pEnd ? p + 1 : _pEnd;

	return _pEnd;
}

bool CheckAnsiFile (const char * _pData, size_t _cFileSize, bool _bSilent) {
	if (_cFileSize < 2)
		return true;

	if (_cFileSize >= 2 && _pData [0] == '\xFF' && _pData [1] == '\xFE') {
		if (! _bSilent)
			ReportText ("ERROR: Input file is in Unicode.");
		return false;
	}
	if (_cFileSize >= 2 && _pData [0] == '\xEF' && _pData [1] == '\xBB') {
		if (! _bSilent)
			ReportText ("ERROR: Input file is in UTF-8.");
		return false;
	}

	return true;
}

bool ParseFloat (const char * _pValue, const char * _pEnd, const char ** _ppEnd, float & _fRet) {
	float r = 0;
	bool bMinus = false;
	bool bHasDigits = false;
	const char * p = _pValue;

	// Omit non-digits.
	while (p < _pEnd && ! (('0' <= * p && * p <= '9') || * p == '-' || * p == '.'))
		++ p;

	// Get sign.
	if (p < _pEnd && * p == '-') {
		++ p;
		bMinus = true;
	}

	// Get whole part of the digit.
	while (p < _pEnd && '0' <= * p && * p <= '9') {
		r *= 10;
		r += (* p) - '0';
		++ p;
		bHasDigits = true;
	}

	// Get decimal dot.
	if (p < _pEnd && * p != '.') {
		* _ppEnd = p;
		_fRet = bMinus ? -r : r;
		return bHasDigits;
	}
	++ p;
	const char * pFraction = p;

	// Get fractional part of the digit.
	float f = 0;
	float pw = 0.1f;
	while (p < _pEnd && '0' <= * p && * p <= '9') {
		f += ((* p) - '0')*pw;
		++ p;
		pw *= .1f;
		bHasDigits = true;
	}

	* _ppEnd = p;
	_fRet = bMinus ? -(r + f) : (r + f);

	return bHasDigits;
}

//////////////////////////////////////////////

CMapLoader::CMetaData * CMapLoader::CMetaData::s_pFirst;
size_t CMapLoader::CMetaData::s_cMaxIndex = 0;

CMapLoader * CMapLoader::CreateLoader (const char * _pData, size_t _cFileSize, const char * _strFileName, bool & _bTypeCanBeSaved, size_t _cLoaderIndex) {
	assert (CMetaData::s_pFirst);
	assert (_cLoaderIndex <= CMetaData::s_cMaxIndex + 2);
/*
	if (2 <= _cLoaderIndex > 0 && _cLoaderIndex <= CMetaData::s_cMaxIndex + 1) {
		// Search by index.
		size_t cIndex = 3;
		for (CMetaData * p = CMetaData::s_pFirst; p; p = p->m_pNext, ++ cIndex) {
			if (cIndex == _cLoaderIndex) {
				//return p->m_pfCreate ();
				CMapLoader * const pMapLoader = p->m_pfCreate ();
				if (pMapLoader->IsValid (_pData, _cFileSize)) {
					_bTypeCanBeSaved = p->m_bTypeCanBeSaved;
					return pMapLoader;
				}
				delete pMapLoader;
				break;
			}
		}
	}*/

	// Get file name extension.
	const char * strExt = ::strrchr (_strFileName, '.');
	if (strExt != NULL) {
		++ strExt;

		// Search for the creator associated for the extension.
		// TODO: support multiple extensions.
		for (CMetaData * p = CMetaData::s_pFirst; p; p = p->m_pNext) {
			if (::stricmp (strExt, p->m_strExt) == 0) {
				CMapLoader * const pMapLoader = p->m_pfCreate ();
				if (pMapLoader->IsValid (_pData, _cFileSize)) {
					_bTypeCanBeSaved = p->m_bTypeCanBeSaved;
					return pMapLoader;
				}
				delete pMapLoader;
			}
		}
	}

	// No association found.
	_bTypeCanBeSaved = false;
	return NULL;
}

static
DWORD EHFilter (void * _ei) {
# ifdef _DEBUG
	return EXCEPTION_CONTINUE_SEARCH;
# else
	if (::IsDebuggerPresent ())
		return EXCEPTION_CONTINUE_SEARCH;

	const EXCEPTION_POINTERS * pe = reinterpret_cast<EXCEPTION_POINTERS *> (_ei);

	const char * strType = NULL;
	switch (pe->ExceptionRecord->ExceptionCode) {
		case EXCEPTION_ACCESS_VIOLATION:    strType = "access violation"; break;
		case EXCEPTION_STACK_OVERFLOW:      strType = "stack overlow"; break;
		case EXCEPTION_FLT_DIVIDE_BY_ZERO:  strType = "divide by zero (floating-point)"; break;
		case EXCEPTION_ILLEGAL_INSTRUCTION: strType = "illegal instruction"; break;
		case EXCEPTION_INT_DIVIDE_BY_ZERO:  strType = "divide by zero (integer)"; break;
			break;

		case EXCEPTION_BREAKPOINT:
		default:
			return EXCEPTION_CONTINUE_SEARCH;
	}

	ReportWarning ("Internal error.");
	ReportText ("Please report the error log to the developer (kg@infratel.com):");
	ReportText ("------------------------------");
	ReportText ("Instruction at [0x%08x] caused %s.", pe->ExceptionRecord->ExceptionAddress, strType);

	if (
		pe->ExceptionRecord->ExceptionCode == EXCEPTION_ACCESS_VIOLATION &&
		pe->ExceptionRecord->NumberParameters >= 2
	) {
		ReportText (
			"Attempt to %s to [0x%08x] which is illegal address.",
			pe->ExceptionRecord->ExceptionInformation [0] == 0 ? "read" : "write",
			pe->ExceptionRecord->ExceptionInformation [1]
		);
	}
/*
	// Report stack.
	STACKFRAME sf;
	const size_t cStackDepth = 5;
	DWORD dwPC [cStackDepth];
	for (size_t cFrame = 0; cFrame < cStackDepth; ++ cFrame) {
		::StackWalk (IMAGE_FILE_MACHINE_I386, ::GetCurrentProcess (), ::GetCurrentThread (), & sf, NULL, NULL, NULL, NULL, NULL);
		dwPC [cFrame] = sf.AddrPC.Offset;
	}
	ReportText ("Stack: 0x%08x 0x%08x 0x%08x 0x%08x", dwPC [0], dwPC [1], dwPC [2], dwPC [3]);
*/
	// Report OS version.
	OSVERSIONINFO oi = {sizeof (OSVERSIONINFO)};
	::GetVersionEx (& oi);
	ReportText (
		"OS: %s %d.%d %s",
		oi.dwPlatformId == VER_PLATFORM_WIN32_NT ? "NT" : "9x",
		oi.dwMajorVersion, oi.dwMinorVersion, oi.szCSDVersion 
	);

	// Report program version.
	ReportText ("Version: " GME_VER_S);
	ReportText ("------------------------------");

	return EXCEPTION_EXECUTE_HANDLER;
# endif // _DEBUG
}

static
bool ExecuteLoadMapSafely (map_t * _pMap, const char * _pData, DWORD _dwFileSize, const char * _strFileName, CMapLoader * _pLoader) {
	__try {
		assert (_pMap);
		return _pMap->LoadMap (_pData, _dwFileSize, _strFileName, _pLoader);
	} __except (EHFilter (_exception_info ())) {
		return false;
	}
}

bool CMapLoader::LoadFile (const char * _strFileName, map_t * _pMap, bool & _bTypeCanBeSaved, size_t _cLoaderIndex) {
	CHandle hFile = ::CreateFile (_strFileName, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL | FILE_FLAG_SEQUENTIAL_SCAN, NULL);
	if (hFile == NULL || hFile == INVALID_HANDLE_VALUE) {
		ReportWarning ("Can't open file '%s'.", _strFileName);
		return false;
	}

	CHandle hFileMapping = ::CreateFileMapping (hFile, NULL, PAGE_READONLY, 0, 0, NULL);
	const char * const pData = reinterpret_cast<char *> (::MapViewOfFile (hFileMapping, FILE_MAP_READ, 0, 0, 0));
	if (! pData) {
		ReportWarning ("Can't read file '%s'.", _strFileName);
		return false;
	}
	const DWORD dwFileSize = ::GetFileSize (hFile, NULL);

	CMapLoader * const pLoader = CMapLoader::CreateLoader (pData, dwFileSize, _strFileName, _bTypeCanBeSaved, _cLoaderIndex);
	if (! pLoader) {
		::UnmapViewOfFile (pData);

		ReportWarning ("Unknown file type.", _strFileName);
		return false;
	}

	bool bLoadOK = false;
	try {
		bLoadOK = ExecuteLoadMapSafely (_pMap, pData, dwFileSize, _strFileName, pLoader);
	} catch (int) {
		ReportText ("Not enough memory.");
	}

	delete pLoader;

	::UnmapViewOfFile (pData);

	return bLoadOK;
}

void CMapLoader::ShowOpenFile (bool _bRemoveOldMap) {
	char strFilesName [(MAX_PATH + 1)*100 + 1];
	strFilesName [0] = 0;

	OPENFILENAME ofn;
	::ZeroMemory (& ofn, sizeof (OPENFILENAME));
	ofn.lStructSize = sizeof (OPENFILENAME);
	ofn.hwndOwner   = g_hWnd;
	ofn.hInstance   = g_hInst;
	ofn.lpstrTitle  = _bRemoveOldMap ? NULL : "Add";
	ofn.lpstrFile   = strFilesName;
	ofn.nMaxFile    = sizeof (strFilesName);
	ofn.Flags = OFN_EXPLORER | OFN_PATHMUSTEXIST | OFN_FILEMUSTEXIST | OFN_ENABLESIZING | OFN_FILEMUSTEXIST | OFN_HIDEREADONLY | OFN_ALLOWMULTISELECT;

	//
	// Prepare ofn.lpstrFilter.
	//
	CMetaData * p;

	static const char strAllFiles [] = "All files (*.*)\0*.*\0";

	// Prepare sub-filter for all supported types.
	std::string strAllSupportedFiles = "All supported files (";
	typedef std::set<std::string> exts_t;
	exts_t exts;
	for (p = CMetaData::s_pFirst; p; p = p->m_pNext)
		exts.insert (p->m_strExt);

	exts_t::const_iterator e;
	for (e = exts.begin (); e != exts.end (); ++ e) {
		if (e != exts.begin ())
			strAllSupportedFiles.append (",");
		strAllSupportedFiles.append ("*.");
		strAllSupportedFiles.append (* e);
	}
	strAllSupportedFiles.append (")\0", 2);
	for (e = exts.begin (); e != exts.end (); ++ e) {
		if (e != exts.begin ())
			strAllSupportedFiles.append (";");
		strAllSupportedFiles.append ("*.");
		strAllSupportedFiles.append (* e);
	}
	strAllSupportedFiles.append ("\0", 1);

	// Estimate the size of the filter.
	size_t cFilterLen = sizeof (strAllFiles) + strAllSupportedFiles.size ();
	for (p = CMetaData::s_pFirst; p; p = p->m_pNext)
		cFilterLen += p->m_cFilterLen;

	// Build the filter.
	char * const strFilter = reinterpret_cast<char *> (::alloca (cFilterLen));

	::memcpy (strFilter, strAllSupportedFiles.c_str (), strAllSupportedFiles.size ());
	size_t cOffset = strAllSupportedFiles.size ();

	for (p = CMetaData::s_pFirst; p; p = p->m_pNext)
		if (p->m_bDefault) {
			// Put the default filter second.
			::memcpy (strFilter + cOffset, p->m_strFilter, p->m_cFilterLen);
			cOffset += p->m_cFilterLen;
			break;
		}

	for (p = CMetaData::s_pFirst; p; p = p->m_pNext) {
		if (p->m_bDefault)
			continue;
		::memcpy (strFilter + cOffset, p->m_strFilter, p->m_cFilterLen);
		cOffset += p->m_cFilterLen;
	}
	::memcpy (strFilter + cOffset, strAllFiles, sizeof (strAllFiles));

	ofn.lpstrFilter = strFilter;

	//
	// Prepare ofn.nFilterIndex.
	//
	static size_t s_cFilterIndex = 0;
	ofn.nFilterIndex = s_cFilterIndex;
	if (ofn.nFilterIndex == 0)
		ofn.nFilterIndex = 1; // "All supported types".
	
	//
	// Show the dialog box & load.
	//
	if (::GetOpenFileName (& ofn) == TRUE) {
		// Save ofn.nFilterIndex.
		s_cFilterIndex = ofn.nFilterIndex;

		std::string strPath;
		file_names_t file_names;

		const char * p = strFilesName;
		while (* p) {
			const char * pEnd = p + ::strlen (p);

			std::string strFileName (p, pEnd);
			if (strPath.empty ())
				strPath.swap (strFileName);
			else
				file_names.push_back (strPath + "\\" + strFileName);

			p = pEnd + 1;
		}

		if (file_names.empty ())
			file_names.push_back (strPath);

		LoadMap (file_names, ofn.nFilterIndex, _bRemoveOldMap);
	}
}

///////////////////////////////////////////////

HANDLE g_hLoadThread;
volatile bool g_bLoading = false;
volatile bool g_bStop = false;

struct load_data_t {
	file_names_t file_names;
	size_t       cFileType;
	bool         bRemoveOldMap;
};

DWORD WINAPI LoadMapProc (LPVOID);

void LoadMap (const char * _strFileName, size_t _cFileType, bool _bRemoveOldMap) {
	file_names_t file_names;
	file_names.push_back (_strFileName);

	LoadMap (file_names, _cFileType, _bRemoveOldMap);
}

void LoadMap (const file_names_t & _file_names, size_t _cFileType, bool _bRemoveOldMap) {
	// Wait for previous parsing ends.
	if (g_hLoadThread)
		StopLoadThread ();

	g_bStop = false;

	g_bLoading = true;
	UpdateMenu ();

	// Remove old map.
	if (_bRemoveOldMap)
		RemoveMap ();

	// Prepare input data for the thread proc.
	load_data_t * const pParams = new load_data_t;
	pParams->cFileType = _cFileType;
	pParams->bRemoveOldMap = _bRemoveOldMap;
	std::set<std::string> names;
	for (file_names_t::const_iterator fi = _file_names.begin (); fi != _file_names.end (); ++ fi) {
		// Make absolute file path.
		char strFullPathName [MAX_PATH];
		char * p = NULL;
		::GetFullPathName (fi->c_str (), MAX_PATH, strFullPathName, & p);

		if (names.find (strFullPathName) != names.end ())
			// Omit duplicated name.
			continue;

		names.insert (strFullPathName);
		pParams->file_names.push_back (strFullPathName);
	}

	// Perform the loading in parallel thread.
	DWORD dwThreadId = 0;
	g_hLoadThread = ::CreateThread (NULL, 0, & LoadMapProc, pParams, 0, & dwThreadId);
	if (! g_hLoadThread) {
		ErrorMsgBox ("Internal error.");
		return;
	}
}

static
DWORD WINAPI LoadMapProc (LPVOID _pParams) {
	// Get input params.
	load_data_t * const pParams = reinterpret_cast<load_data_t *> (_pParams);
	load_data_t params;
	params.file_names.swap (pParams->file_names);
	params.cFileType = pParams->cFileType;
	params.bRemoveOldMap = pParams->bRemoveOldMap;
	delete pParams;

	CWaiter w;

	ClearLog ();

	const long lStartTime = ::GetTickCount ();

	for (file_names_t::const_iterator fi = params.file_names.begin (); fi != params.file_names.end (); ++ fi) {
		const char * strFileName = fi->c_str ();
		const bool bFirstFile = fi == params.file_names.begin ();

		if (! bFirstFile)
			ReportText ("");
		ReportText ("Loading map from file '%s'...", strFileName);

		map_t map;
		bool bTypeCanBeSaved = false;
		if (! CMapLoader::LoadFile (strFileName, & map, bTypeCanBeSaved, params.cFileType)) {
			ResetStatus ();
			g_bLoading = false;
			UpdateMenu ();

			::SetCursor (g_hPermanentCursor);

			if (g_bStop) {
				ReportText ("Loading was terminated by user request.");
				return 0;
			}

			ReportText ("ERROR: loading has terminated.");
			//ErrorMsgBox ("Can't load file\n%s.\n\nSee File|Error Log for details.", strFileName);
			ShowErrorLog ();

			return 1;
		}
		if (params.bRemoveOldMap)
			AddRecentFile (strFileName);

		//
		// Actualize the new map.
		//

		if (bFirstFile) {
			if (params.bRemoveOldMap) {
				CCritSec::CLock lock (g_cs);
				g_map.Swap (map);

				::strcpy (g_strFileName, strFileName);
				g_bTypeCanBeSaved = bTypeCanBeSaved;
			} else {
				if (! AddMap (& map)) {
					ResetStatus ();
					g_bLoading = false;
					UpdateMenu ();

					return 1;
				}
			}
		} else {
			CCritSec::CLock lock (g_cs);
			g_map.Join (map);
		}
	}

	//
	// Report about results.
	//

	if (GetWarningsCount ())
		ReportText ("%d warning(s).", GetWarningsCount ());

	char strTime [64];
	::sprintf (strTime, "Load time %.2f s.", .001*(::GetTickCount () - lStartTime));
	ReportText (strTime);

	bool bRectBoundInvalid;
	{
		CCritSec::CLock lock (g_cs);
		bRectBoundInvalid = g_map.rectBound.Invalid ();
	}
	if (! bRectBoundInvalid) {
		//
		// Set the view to the map's center.
		//

		SetFullMapScale ();

		//
		// Render the mini-map.
		//

		if (g_lTotalHeight > g_lTotalWidth) {
			g_lMiniW = 100;
			g_lMiniH = g_lTotalWidth ? 100*g_lTotalHeight/g_lTotalWidth : 100;
		} else {
			g_lMiniW = g_lTotalHeight ? 100*g_lTotalWidth/g_lTotalHeight : 100;
			g_lMiniH = 100;
		}

		if (g_hMiniBmp)
			::DeleteObject (g_hMiniBmp);
		if (g_hMiniMemDC)
			::DeleteDC (g_hMiniMemDC);
		const CClientDC hMapWndDC (g_hMapWnd);
		g_hMiniMemDC = ::CreateCompatibleDC (hMapWndDC);
		g_hMiniBmp = ::CreateCompatibleBitmap (hMapWndDC, g_lMiniW, g_lMiniH);
		::SelectObject (g_hMiniMemDC, g_hMiniBmp);

		long lW, lH;
		{
			CCritSec::CLock lock (g_cs);
			lW = fast_floor (.5 + g_map.rectBound.Width ()*g_kX);
			lH = fast_floor (.5 + g_map.rectBound.Height ()*g_kY);
		}
		if (lW > 0 && lH > 0) {
			::SetStretchBltMode (g_hMiniMemDC, HALFTONE);
			::StretchBlt (
				g_hMiniMemDC,
				0, 0, g_lMiniW, g_lMiniH, g_hMemDC,
				(g_lWidth - lW)/2, (g_lHeight - lH)/2, lW, lH,
				SRCCOPY
			);
		}

		// Show the mini-map.
		::SetWindowPos (g_hMiniMapWnd, NULL, 0, 0, g_lMiniW, g_lMiniH, SWP_NOMOVE | SWP_NOZORDER);
		RECT rectMini;
		::GetClientRect (g_hMiniMapWnd, & rectMini);
		::SetWindowPos (
			g_hMiniMapWnd, NULL, 0, 0,
			2*g_lMiniW - (rectMini.right - rectMini.left),
			2*g_lMiniH - (rectMini.bottom - rectMini.top),
			SWP_NOMOVE | SWP_NOZORDER
		);
	}

	//
	// Update GUI elements.
	//

	g_bLoading = false;

	UpdateTitle ();
	MakeLevelsMenu ();
	UpdateMenu ();

	// NOTE: previous Refresh() does not display ruler due to g_bLoading==false.
	// Make ruler to draw now.
	Redraw ();

	// Redraw mini-map.
	RedrawMini ();

	if (GetWarningsCount ())
		SetStatus ("Ready*");
	else
		SetStatus ("Ready");

	return 0;
}

void StopLoadThread () {
	g_bStop = true;

	if (! g_hLoadThread)
		return;

	while (true) {
		const DWORD dwRet = ::MsgWaitForMultipleObjects (1, & g_hLoadThread, FALSE, INFINITE, QS_ALLINPUT);
		if (dwRet == WAIT_OBJECT_0)
			// The thread exits.
			break;

		if (dwRet != WAIT_OBJECT_0 + 1)
			// Smth else happened.
			break; 

		// There is one or more window message available. Dispatch them.
		MSG msg;
		while (::PeekMessage (& msg, NULL, 0, 0, PM_REMOVE)) {
			if (! ::TranslateAccelerator (msg.hwnd, g_hAccelTable, & msg)) {
				::TranslateMessage (& msg);
				::DispatchMessage (& msg);
			}
			if (::WaitForSingleObject (g_hLoadThread, 0) == WAIT_OBJECT_0) {
				// The thread exits now.
				::CloseHandle (g_hLoadThread);
				g_hLoadThread = NULL;
				return;
			}
		}
	}

	::CloseHandle (g_hLoadThread);
	g_hLoadThread = NULL;
}
