# ifndef __Load_H_
# define __Load_H_

bool CheckAnsiFile (const char * _pData, size_t _cFileSize, bool _bSilent);
const char * FindNextLine (const char * _pBegin, const char * _pEnd);
bool ParseFloat (const char * _pValue, const char * _pEnd, const char ** _ppEnd, float & _fRet);

inline
bool HasPrefix (const char * _p, const char * _pEnd, const char * _pPrefix, size_t _cPrefixLen) {
	assert (_cPrefixLen == ::strlen (_pPrefix));

	return _pEnd - _p >= int (_cPrefixLen) && ::_memicmp (_p, _pPrefix, _cPrefixLen) == 0;
}

////////////////////////////////////////////

extern volatile bool g_bLoading;
extern volatile bool g_bStop;

typedef std::list<std::string> file_names_t;

void LoadMap (const char * _strFileName, size_t _cFileType, bool _bRemoveOldMap);
void LoadMap (const file_names_t & _FileNames, size_t _cFileType, bool _bRemoveOldMap);
void StopLoadThread ();

struct map_t;

class CMapLoader {
  public:
	virtual bool IsValid (const char * _pData, size_t _cFileSize) const = 0;
	// NOTE: _strFileName is supplied to get folder path, if this is required to locate some referred files.
	virtual bool Load (const char * _pData, size_t _cFileSize, const char * _strFileName, map_t * _pMap) = 0;

	//
	// The utilities for map loaders (registration, search etc).
	//

	// If _cLoaderIndex=0, then loader is searched by file extension.
	static CMapLoader * CreateLoader (const char * _pData, size_t _cFileSize, const char * _pFileName, bool & _bTypeCanBeSaved, size_t _cLoaderIndex = 0);
	static bool LoadFile (const char * _strFileName, map_t * _pMap, bool & _bTypeCanBeSaved, size_t _cLoaderIndex = 0);
	//
	static void ShowOpenFile (bool _bRemoveOldMap);

	class CMetaData {
	  public:
		typedef CMapLoader * (* pfCreate_t) ();

		CMetaData::CMetaData (pfCreate_t _pfCreate, const char * _strExt, const char * _strFilter, size_t _cFilterLen, bool _bDefault, bool _bTypeCanBeSaved) :
			m_pfCreate (_pfCreate),
			m_strExt (_strExt),
			m_strFilter (_strFilter),
			m_cFilterLen (_cFilterLen),
			m_bDefault (_bDefault),
			m_bTypeCanBeSaved (_bTypeCanBeSaved)
		{
			// Insert to the head of the list.
			m_pNext = CMetaData::s_pFirst;
			CMetaData::s_pFirst = this;

			++ CMetaData::s_cMaxIndex;
		}

		const pfCreate_t   m_pfCreate;
		const char * const m_strExt;
		const char * const m_strFilter;
		const size_t       m_cFilterLen;
		const bool         m_bDefault;
		const bool         m_bTypeCanBeSaved;

	  private:
		friend class CMapLoader;

		// The linked list of creators.
		CMetaData * m_pNext;
		static CMetaData * s_pFirst;
		static size_t s_cMaxIndex;
	};
};

# define DECLARE_MAP_LOADER_impl(_T,_strExt,_strFilter,_bDefault,_bTypeCanBeSaved) \
	CMapLoader * Create_##_T () {return new _T;} \
	CMapLoader::CMetaData Creator_##_T (& Create_##_T, _strExt, _strFilter, sizeof (_strFilter), _bDefault, _bTypeCanBeSaved);

# define DECLARE_MAP_LOADER(_T,_strExt,_strFilter,_bTypeCanBeSaved)         DECLARE_MAP_LOADER_impl(_T,_strExt,_strFilter, false, _bTypeCanBeSaved)
# define DECLARE_DEFAULT_MAP_LOADER(_T,_strExt,_strFilter,_bTypeCanBeSaved) DECLARE_MAP_LOADER_impl(_T,_strExt,_strFilter, true,  _bTypeCanBeSaved)


# endif // __Load_H_