# include "stdafx.h"
# include "resource.h"
# include "Wrappers.h"
# include "Log.h"
# include "Globals.h"

static char * g_strErrorLog;
static size_t g_cErrorLogSize;
static size_t g_cWarningCounter;

typedef std::map<size_t, size_t> wtc_t;
static wtc_t g_WarningTypesCount;

# define WARNING_PREFIX "WARNING (offset "

void ClearLog () {
	delete [] g_strErrorLog;
	g_strErrorLog = NULL;
	g_cErrorLogSize = 0;
	g_cWarningCounter = 0;
	g_WarningTypesCount.clear ();
}

bool IsLogEmpty () {
	return g_cErrorLogSize == 0;
}

size_t GetWarningsCount () {
	return g_cWarningCounter;
}

static
void ReportTextImpl (const char * _strText, ...) {
	assert (_strText);

	if (! g_strErrorLog) {
		g_cErrorLogSize = ::strlen (_strText);
		g_strErrorLog = new char [g_cErrorLogSize + 1];
		::strcpy (g_strErrorLog, _strText);
	} else {
		assert (g_cErrorLogSize == ::strlen (g_strErrorLog));

		const size_t cErrorLogSize = g_cErrorLogSize + ::strlen (_strText) + 2;
		char * strErrorLog = new char [cErrorLogSize + 1];

		::strcpy (strErrorLog, g_strErrorLog);
		::strcpy (strErrorLog + g_cErrorLogSize, "\r\n");
		::strcpy (strErrorLog + g_cErrorLogSize + 2, _strText);

		delete [] g_strErrorLog;

		g_strErrorLog = strErrorLog;
		g_cErrorLogSize = cErrorLogSize;
	}
}

static
bool CanReportWarning (const char * _strText) {
	assert (_strText);

	const size_t cKey = reinterpret_cast<size_t> (_strText);

	wtc_t::iterator i = g_WarningTypesCount.find (cKey);
	if (i == g_WarningTypesCount.end ()) {
		g_WarningTypesCount.insert (wtc_t::value_type (cKey, 1));
		return true;
	}

	++ i->second;
	return i->second <= 10;
}

void ReportWarning (const char * _strText, ...) {
	assert (_strText);

	if (! CanReportWarning (_strText))
		return;

	++ g_cWarningCounter;
	if (g_cWarningCounter == 200)
		ReportTextImpl ("ERROR: Too much warnings.");
	if (g_cWarningCounter >= 200)
		return;

	va_list vl;
	va_start (vl, _strText);

	const size_t BUF_LEN = 256;
	char buf [BUF_LEN] = "WARNING: ";
	const size_t cPrefixLen = ::strlen (buf);
	::_vsnprintf (buf + cPrefixLen, BUF_LEN - 1 - cPrefixLen, _strText, vl);

	va_end (vl);

	ReportTextImpl (buf);
}

void ReportWarning (size_t _cOffset, const char * _strText, ...) {
	assert (_strText);

	if (! CanReportWarning (_strText))
		return;

	++ g_cWarningCounter;
	if (g_cWarningCounter == 200)
		ReportTextImpl ("ERROR: Too much warnings.");
	if (g_cWarningCounter >= 200)
		return;

	va_list vl;
	va_start (vl, _strText);

	const size_t BUF_LEN = 256;
	char buf [BUF_LEN];
	::sprintf (buf, WARNING_PREFIX "%u): ", _cOffset);
	const size_t cPrefixLen = ::strlen (buf);
	::_vsnprintf (buf + cPrefixLen, BUF_LEN - 1 - cPrefixLen, _strText, vl);

	va_end (vl);

	ReportTextImpl (buf);
}

void ReportText (const char * _strText, ...) {
	assert (_strText);

	va_list vl;
	va_start (vl, _strText);

	const size_t BUF_LEN = 256;
	char buf [BUF_LEN];
	::_vsnprintf (buf, BUF_LEN - 1, _strText, vl);

	va_end (vl);

	ReportTextImpl (buf);
}

int CALLBACK ErrorLogProc (HWND _hWnd, UINT _uMsg, WPARAM _wParam, LPARAM);
int CALLBACK SourceProc (HWND _hWnd, UINT _uMsg, WPARAM _wParam, LPARAM _lParam);

void ShowErrorLog () {
	::DialogBoxParam (g_hInst, MAKEINTRESOURCE (IDD_ERROR_LOG), g_hWnd, & ErrorLogProc, 0);
}

static WNDPROC g_wpEdit = NULL;

static
LRESULT EditProc (HWND _hWnd, UINT _uMsg, WPARAM _wParam, LPARAM _lParam) {
	if (_uMsg == WM_LBUTTONDBLCLK) {
		// NOTE: _lParam contains client coordinates of click point.
		// Get index of the nearest line in the edit box.
		const DWORD dwLine = HIWORD (::SendMessage (_hWnd, EM_CHARFROMPOS, 0, _lParam));

		// Get beginning part oth the line.
		char strLine [sizeof (WARNING_PREFIX) + 10];
		* (reinterpret_cast<WORD *> (strLine)) = sizeof (strLine);
		::SendMessage (_hWnd, EM_GETLINE, dwLine, reinterpret_cast<LPARAM> (strLine));

		// Check if it warning line and get source file offset for this warning.
		if (::strncmp (strLine, WARNING_PREFIX, sizeof (WARNING_PREFIX) - 1) != 0)
			return 0;
		const size_t cFileOffset = ::atoi (strLine + sizeof (WARNING_PREFIX) - 1);

		// Show the source text.
		::DialogBoxParam (g_hInst, MAKEINTRESOURCE (IDD_MAP_HEADER), _hWnd, & SourceProc, cFileOffset);

		return 0;
	}

	return ::CallWindowProc (g_wpEdit, _hWnd, _uMsg, _wParam, _lParam);
}

static
int CALLBACK ErrorLogProc (HWND _hWnd, UINT _uMsg, WPARAM _wParam, LPARAM _lParam) {
	switch (_uMsg) {
		case WM_INITDIALOG: {
			// Subclass the edit box to hook WM_LBUTTONDBLCLK.
			const HWND hEdit = ::GetDlgItem (_hWnd, IDC_LOG);
			g_wpEdit = reinterpret_cast<WNDPROC> (::GetWindowLong (hEdit, GWL_WNDPROC));
			::SetWindowLong (hEdit, GWL_WNDPROC, reinterpret_cast<long> (EditProc));

			if (g_cErrorLogSize)
				::SetDlgItemText (_hWnd, IDC_LOG, g_strErrorLog);
			else
				::SetDlgItemText (_hWnd, IDC_LOG, "");

			return TRUE;
		}

		case WM_COMMAND:
			switch (LOWORD (_wParam)) {
				case IDOK:
				case IDCANCEL:
					// On closing dialog.
					::EndDialog (_hWnd, 0);
					return TRUE;
			}
			break;
	}

	return FALSE;
}

const size_t c_cWINDOW = 32768;

static
int CALLBACK SourceProc (HWND _hWnd, UINT _uMsg, WPARAM _wParam, LPARAM _lParam) {
	switch (_uMsg) {
		case WM_INITDIALOG: {
			::SetWindowText (_hWnd, "Source text");

			//
			// Prepare source edit box.
			//
			const CHandle hFile = ::CreateFile (g_strFileName, GENERIC_READ, 0, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
			if (hFile == NULL || hFile == INVALID_HANDLE_VALUE) {
				::SetDlgItemText (_hWnd, IDC_SOURCE, "Can't open the file.");
				return TRUE;
			}

			const DWORD dwSize = ::GetFileSize (hFile, NULL);

			const size_t cFileOffset = _lParam;
			const size_t cBeginOffset = cFileOffset > c_cWINDOW ? cFileOffset - c_cWINDOW : 0;
			const size_t cEndOffset = cFileOffset + c_cWINDOW > dwSize ? dwSize : cFileOffset + c_cWINDOW;

			char * const pSource = reinterpret_cast<char *> (alloca (sizeof (char)*c_cWINDOW*2));

			::SetFilePointer (hFile, cBeginOffset, 0, FILE_BEGIN);
			DWORD dwFileSizeRead = 0;
			::ReadFile (hFile, pSource, cEndOffset - cBeginOffset, & dwFileSizeRead, NULL);
			if (dwFileSizeRead < cEndOffset - cBeginOffset) {
				::SetDlgItemText (_hWnd, IDC_SOURCE, "Can't read the file.");
				return TRUE;
			}
			pSource [cEndOffset - cBeginOffset] = 0;

			::SetDlgItemText (_hWnd, IDC_SOURCE, pSource);

			// Scroll to the line.
			const size_t cLine = ::SendDlgItemMessage (_hWnd, IDC_SOURCE, EM_LINEFROMCHAR, cFileOffset - cBeginOffset, 0);
			::SendDlgItemMessage (_hWnd, IDC_SOURCE, EM_LINESCROLL, 0, cLine > 5 ? cLine - 5 : 0);

			// Select the line.
			const size_t cSelBegin = ::SendDlgItemMessage (_hWnd, IDC_SOURCE, EM_LINEINDEX, cLine, 0);
			const size_t cSelSize = ::SendDlgItemMessage (_hWnd, IDC_SOURCE, EM_LINELENGTH, cFileOffset - cBeginOffset, 0);
			::SendDlgItemMessage (_hWnd, IDC_SOURCE, EM_SETSEL, cSelBegin, cSelBegin + cSelSize + 1);

			return TRUE;
		}

		case WM_COMMAND:
			if (LOWORD (_wParam) == IDOK || LOWORD (_wParam) == IDCANCEL) {
				// On closing dialog.
				::EndDialog (_hWnd, 0);
				return TRUE;
			}
	}

	return FALSE;
}
