# include "stdafx.h"
# include "resource.h"
# include "Globals.h"

# include "Menu.h"
# include "Status.h"
# include "Toolbar.h"
# include "MapView.h"
# include "Settings.h"

# include "Map.h"
# include "Types.h"
# include "Scale.h"
# include "Position.h"
# include "Levels.h"
# include "MouseMode.h"
# include "Selection.h"

# include "NewMap.h"
# include "Load.h"
# include "Save.h"
# include "Log.h"
# include "RecentFiles.h"

# include "MapProperties.h"
# include "RgnProperties.h"

# include "UndoRedo.h"
# include "DeleteRegion.h"
# include "DeleteElement.h"
# include "DeleteAttachment.h"
# include "DeleteImage.h"
# include "CombinedAction.h"

# include "About.h"


char g_strFileName [MAX_PATH + 1];
bool g_bTypeCanBeSaved;

bool g_bDialogMode = false;

HINSTANCE g_hInst;
HWND g_hWnd;
HACCEL g_hAccelTable;

bool g_bNT;

//////////////////////////////////////////

HFONT g_hFont;

HBRUSH g_hNullBrush;
HBRUSH g_hTmpBrush;
HBRUSH g_hBlackBrush;
HBRUSH g_hGrayBrush;
HBRUSH g_hWhiteBrush;

HBRUSH g_hWhiteHatchedBrush;

HBRUSH g_hActiveColorBrush;
HBRUSH g_hBackColorBrush;

HPEN g_hNullPen;
HPEN g_hBlackThinPen;
HPEN g_hGrayThinPen;
HPEN g_hWhiteThickPen;

HPEN g_hGridPen;

//////////////////////////////////////////

pfAlphaBlend_t g_fpAlphaBlend;

//////////////////////////////////////////

void ErrorMsgBox (const char * _strFormat, ...) {
	va_list vl;
	va_start (vl, _strFormat);

	char strBuf [MAX_PATH*2];
	::_vsnprintf (strBuf, sizeof (strBuf), _strFormat, vl);

	::MessageBox (g_hWnd, strBuf, NULL, MB_OK | MB_ICONERROR);

	va_end (vl);
}

/////////////////////////////////////////////////////
/*
static void FindFiles (const char * _strFileName, file_names_t & _file_names) {
	WIN32_FIND_DATA fd;
	CHandle hFind = ::FindFirstFile (_strFileName, & fd);
	if (hFind == INVALID_HANDLE_VALUE)
		return;

	do {
		_file_names.push_back (fd.cFileName);
	} while (::FindNextFile (hFind, & fd));
}*/

static
void LoadFromCmdLine (const char * _strCmdLine) {
	file_names_t file_names;

	const char * p = _strCmdLine;
	while (* p) {
		if (* p == ' ') {
			++ p;
			continue;
		}

		const char * pEnd = ::strchr (p + 1, * p == '\"' ? '\"' : ' ');
		if (pEnd == NULL) {
			std::string strFileName (* p == '\"' ? p + 1 : p);
			//FindFiles (strFileName.c_str (), file_names);
			file_names.push_back (strFileName);
			break;
		} else {
			std::string strFileName (* p == '\"' ? p + 1 : p, pEnd);
			//FindFiles (strFileName.c_str (), file_names);
			file_names.push_back (strFileName);
			p = pEnd + 1;
		}
	}

	LoadMap (file_names, 0, true);
}

/////////////////////////////////////////////////////

	// ::floor() and float-to-long cast appears too slow (takes hundreds of CPU ticks).
	// NOTE: IEEE format is assumed.
long fast_floor (float _f) {
	DWORD dw = reinterpret_cast<DWORD &> (_f);
	if (long (dw) < 0) {
		// For negative values.
		dw &= 0x7FFFFFFF;
		if (dw == 0)
			return 0;

		const size_t sh = 150 - ((dw & 0x7FF80000) >> 23);
		if (sh >= 24)
			return -1;

		if (dw & 0x007FFFFF)
			return - long ((0x00800000 | (dw & 0x007FFFFF)) >> sh) - 1;
		return - long ((0x00800000 | (dw & 0x007FFFFF)) >> sh);
	} else {
		// For positive values.
		if (dw == 0)
			return 0;

		const size_t sh = 150 - ((dw & 0x7FF80000) >> 23);
		if (sh >= 24)
			return 0;

		return (0x00800000 | (dw & 0x007FFFFF)) >> sh;
	}
}

/////////////////////////////////////////////////////

COutOfMemory::COutOfMemory (size_t _size) :
	runtime_error ("Not enough memory.")
{}

int NewHandler (size_t _size) {
	throw COutOfMemory (_size);
}

/////////////////////////////////////////////////////

LRESULT CALLBACK WndProc (HWND, UINT, WPARAM, LPARAM);

int APIENTRY WinMain (
	HINSTANCE _hInstance,
    HINSTANCE _hPrevInstance,
    LPSTR     _strCmdLine,
    int       _nCmdShow
) {
	// Unit test for fast_floor().
	assert (fast_floor (0) == 0);
	assert (fast_floor (0.000001f) == 0);
	assert (fast_floor (0.1f) == 0);
	assert (fast_floor (0.4f) == 0);
	assert (fast_floor (0.5f) == 0);
	assert (fast_floor (0.6f) == 0);
	assert (fast_floor (0.75f) == 0);
	assert (fast_floor (0.76f) == 0);
	assert (fast_floor (0.9f) == 0);
	assert (fast_floor (1) == 1);
	assert (fast_floor (1.1f) == 1);
	assert (fast_floor (1.9f) == 1);
	assert (fast_floor (3.5f) == 3);
	assert (fast_floor (-1) == -1);
	assert (fast_floor (-0.000001f) == -1);
	assert (fast_floor (-.4f) == -1);
	assert (fast_floor (-.5f) == -1);
	assert (fast_floor (-.6f) == -1);
	assert (fast_floor (-.75f) == -1);
	assert (fast_floor (-.76f) == -1);
	assert (fast_floor (-.1f) == -1);
	assert (fast_floor (-1.1f) == -2);
	assert (fast_floor (-3.5f) == -4);

	////////////////////////////////////////////

	g_hInst = _hInstance;

	//
	// Determine OS facilities.
	//

	OSVERSIONINFO oi = {sizeof (OSVERSIONINFO)};
	::GetVersionEx (& oi);
	g_bNT = oi.dwPlatformId == VER_PLATFORM_WIN32_NT;

	const HMODULE hShell = ::LoadLibrary ("Shell32.dll");
	void * const pDllGetVersion = ::GetProcAddress (hShell, "DllGetVersion");
	if (! pDllGetVersion)
		ErrorMsgBox ("Internet Explorer 4.0 or higher is required.");

	//
	// Set up common out-of-memory handler.
	//

	_set_new_handler (NewHandler);
	_set_new_mode (1);

	try {
		//
		// Initialize optional graphics.
		//

		// Get ::AlphaBlend()' address, if any.
		const HMODULE hMsImg32Lib = ::LoadLibrary ("Msimg32.dll");
		if (hMsImg32Lib)
			g_fpAlphaBlend = reinterpret_cast<pfAlphaBlend_t> (::GetProcAddress (hMsImg32Lib, "AlphaBlend"));

		//
		// Initialize registry access.
		//

		InitRegistryAccess ();
		
		//
		// Initalaize shared GDI objects.
		//

		g_hFont = ::CreateFont (12, 0, 0, 0, FW_DONTCARE, FALSE, FALSE, FALSE, DEFAULT_CHARSET, OUT_DEFAULT_PRECIS, CLIP_DEFAULT_PRECIS, DEFAULT_QUALITY, DEFAULT_PITCH | FF_SWISS, NULL);

		g_hNullPen         = (HPEN) ::GetStockObject (NULL_PEN);
		g_hBlackThinPen    = ::CreatePen (PS_SOLID, 1, c_colBlack);
		g_hGrayThinPen     = ::CreatePen (PS_SOLID, 1, c_colGray);
		g_hWhiteThickPen   = ::CreatePen (PS_SOLID, 3, c_colWhite);

		g_hGridPen         = ::CreatePen (PS_SOLID, 1, c_colLtGray);

		g_hNullBrush       = (HBRUSH) ::GetStockObject (NULL_BRUSH);
		g_hBlackBrush      = ::CreateSolidBrush (c_colBlack);
		g_hGrayBrush       = ::CreateSolidBrush (c_colGray);
		g_hWhiteBrush      = ::CreateSolidBrush (c_colWhite);

		g_hWhiteHatchedBrush = ::CreateHatchBrush (HS_DIAGCROSS, c_colWhite);
		g_hTmpBrush          = ::CreateHatchBrush (HS_DIAGCROSS, c_colBlack);

		g_hActiveColorBrush = ::GetSysColorBrush (COLOR_ACTIVECAPTION);
		g_hBackColorBrush   = ::GetSysColorBrush (COLOR_WINDOW);

		InitTypes ();

		//
		// Initialize cursors.
		//

		g_hArrowCursor     = ::LoadCursor (NULL, IDC_ARROW);
		g_hWaitCursor      = ::LoadCursor (NULL, IDC_WAIT);

		//
		// Create the main window.
		//

		WNDCLASSEX wcex;
		::memset (& wcex, 0, sizeof (wcex));
		wcex.cbSize = sizeof (WNDCLASSEX);
		wcex.style			= CS_HREDRAW | CS_VREDRAW;
		wcex.lpfnWndProc	= (WNDPROC) WndProc;
		wcex.hInstance		= _hInstance;
		wcex.hIcon			= ::LoadIcon (_hInstance, (LPCTSTR) IDI_MAIN);
		wcex.hCursor		= g_hArrowCursor;
		wcex.lpszMenuName	= (LPCSTR) IDC_MAIN;
		wcex.lpszClassName	= "mapview";
		::RegisterClassEx (& wcex);

		g_hWnd = ::CreateWindow (
			"mapview", MAIN_TITLE,
			WS_OVERLAPPEDWINDOW | WS_CLIPCHILDREN,
			CW_USEDEFAULT, 0, GetSetting ("MainWidth", DWORD (800)), GetSetting ("MainHeight", DWORD (600)),
			NULL, NULL, _hInstance, NULL
		);
		if (! g_hWnd)
			return TRUE;

		//
		// Create the status window & toolbar.
		//

		INITCOMMONCONTROLSEX icc;
		icc.dwSize = sizeof (icc);
		icc.dwICC = ICC_WIN95_CLASSES | ICC_USEREX_CLASSES;
		::InitCommonControlsEx (& icc);

		if (! InitStatusWindow ())
			return TRUE;
		if (! InitToolbar ())
			return TRUE;

		//
		// Create the map window.
		//

		if (! InitMapWindow ())
			return TRUE;

		//
		// Prepare data.
		//

		InitRecentFiles ();
		CMouseMode::_Init ();
		RemoveMap ();

		//
		// Show the main window.
		//

		::ShowWindow (g_hWnd, GetSetting ("MainMaximized", false) ? SW_SHOWMAXIMIZED : _nCmdShow);
		::UpdateWindow (g_hWnd);
		::SetFocus (g_hWnd);

		::DragAcceptFiles (g_hWnd, TRUE);

		if (_strCmdLine && _strCmdLine [0])
			LoadFromCmdLine (_strCmdLine);

		//
		// Main message loop.
		//

		g_hAccelTable = ::LoadAccelerators (_hInstance, (LPCTSTR) IDC_MAIN);
		MSG msg;
		while (::GetMessage (& msg, NULL, 0, 0)) {
			if (g_bDialogMode) {
				if (
					(msg.message == WM_KEYDOWN || msg.message == WM_KEYUP) &&
					(msg.wParam == VK_RETURN || msg.wParam == VK_ESCAPE)
				) {
					::TranslateMessage (& msg);
					::DispatchMessage (& msg);
					continue;
				}
				if (::IsDialogMessage (g_hWnd, & msg))
					continue;
			}

			if (::TranslateAccelerator (msg.hwnd, g_hAccelTable, & msg))
				continue;

			::TranslateMessage (& msg);
			::DispatchMessage (& msg);
		}

		return msg.wParam;
	} catch (const std::exception & _e) {
		ErrorMsgBox ("%s\nThe application will close.", _e.what ());
		return -1;
	}
}

void UpdateTitle () {
	if (g_strFileName [0] == '\0')
		::SetWindowText (g_hWnd, MAIN_TITLE);
	else {
		char strTitle [MAX_PATH + sizeof (MAIN_TITLE) + 30];
		::sprintf (strTitle, "%s - " MAIN_TITLE, g_strFileName);
		::SetWindowText (g_hWnd, strTitle);
	}
}

///////////////////////////

static
BOOL CALLBACK EnumChildProc (HWND _hChildWnd, LPARAM _lParam) {
	std::vector<HWND> & vecChildren = * reinterpret_cast<std::vector<HWND> *> (_lParam);
	vecChildren.push_back (_hChildWnd);
	return TRUE;
}

static
void EnumChildWindows (HWND _hParentWnd, std::vector<HWND> & _vecChildren) {
	::EnumChildWindows (_hParentWnd, EnumChildProc, reinterpret_cast<LPARAM> (& _vecChildren));
}

///////////////////////////

static
LRESULT CALLBACK WndProc (HWND _hWnd, UINT _uMsg, WPARAM _wParam, LPARAM _lParam) {
	switch (_uMsg) {
		case WM_SIZING:
			return TRUE;

		case WM_SIZE: {
			// Save size and maximize status
			if (_wParam == SIZE_RESTORED) {
				RECT rect;
				::GetWindowRect (_hWnd, & rect);
				SaveSetting ("MainWidth", DWORD (rect.right  - rect.left));
				SaveSetting ("MainHeight", DWORD (rect.bottom - rect.top));
				SaveSetting ("MainMaximized", false);
			} else if (_wParam == SIZE_MAXIMIZED)
				SaveSetting ("MainMaximized", true);

			// Calculate client area size.
			RECT rectClient;
			::GetClientRect (_hWnd, & rectClient);
			const long lMainW = rectClient.right - rectClient.left;
			const long lMainH = rectClient.bottom - rectClient.top;

			RECT rectStatus;
			::GetWindowRect (g_hStatusWnd, & rectStatus);
			const long lStatusH = rectStatus.bottom - rectStatus.top;

			RECT rectToolbar;
			// NOTE: GetWindowRect() returns incorrect size for toolbar.
			::GetClientRect (g_hToolbarWnd, & rectToolbar);
			const long lToolbarH = rectToolbar.bottom - rectToolbar.top - 1;

			// Set the map window size.
			long lW = lMainW;
			if (lW < 100)
				lW = 100;
			long lH = lMainH - lStatusH - lToolbarH;
			if (lH < 100)
				lH = 100;
			::SetWindowPos (g_hMapWnd, NULL, 0, 0, lW, lH, SWP_NOZORDER | SWP_NOMOVE | SWP_DEFERERASE);

			std::vector<HWND> vecChildren;
			EnumChildWindows (g_hMapWnd, vecChildren);

			// Set map view's child windows position to ensure them being visible.
			for (size_t cWnd = 0; cWnd < vecChildren.size (); ++ cWnd) {
				const HWND hChild = vecChildren [cWnd];
				const HWND hParent = ::GetParent (hChild);
				if (hParent != g_hMapWnd)
					continue;

				RECT rectChild;
				::GetWindowRect (hChild, & rectChild);
				POINT p0 = {rectChild.left, rectChild.top};
				::ScreenToClient (g_hMapWnd, & p0);
				POINT p1 = {rectChild.right, rectChild.bottom};
				::ScreenToClient (g_hMapWnd, & p1);
				RECT rectMap;
				::GetClientRect (g_hMapWnd, & rectMap);
				if (p1.x > rectMap.right) {
					p0.x += rectMap.right - p1.x;
					if (p0.x < 0)
						p0.x = 0;
				}
				if (p1.y > rectMap.bottom) {
					p0.y += rectMap.bottom - p1.y;
					if (p0.y < 0)
						p0.y = 0;
				}
				::MoveWindow (
					hChild,
					p0.x, p0.y,	rectChild.right - rectChild.left, rectChild.bottom - rectChild.top,
					TRUE
				);
			}

			// NOTE: toolbar & status bar set size & position automaticaly.
			::SetWindowPos (g_hStatusWnd, NULL, 0, 0, 0, 0,	SWP_NOZORDER | SWP_DEFERERASE);
			::SetWindowPos (g_hToolbarWnd, NULL, 0, 0, 0, 0, SWP_NOZORDER | SWP_DEFERERASE);

			break;
		}

		case WM_NOTIFY: {
			const NMHDR * const pHdr = reinterpret_cast<NMHDR *> (_lParam);
			switch (pHdr->code) { 
				case TTN_GETDISPINFO: {
					TOOLTIPTEXT * const pttt = reinterpret_cast <TOOLTIPTEXT *> (_lParam);
					pttt->hinst = NULL;
					if (pHdr->hwndFrom == g_hToolbarTipWnd)
						//
						// Provide tooltip text for toolbar buttons.
						//
						::GetMenuString (::GetMenu (g_hWnd), pttt->hdr.idFrom, pttt->szText, sizeof (pttt->szText), MF_BYCOMMAND);

					break;
				}

				case NM_RCLICK:
				case NM_LDOWN:
					// Return keyboard input from the label edit box.
					if (g_bDialogMode)
						::SetFocus (g_hWnd);
					break;

				case TBN_DROPDOWN: {
					//
					// Provide dropdown lists for Undo/Redo.
					//
					NMTOOLBAR * const pNMT = reinterpret_cast<NMTOOLBAR *> (_lParam);
					if (pHdr->hwndFrom == g_hToolbarWnd) {
						RECT rect;
						::SendMessage (g_hToolbarWnd, TB_GETRECT, (WPARAM) pNMT->iItem, (LPARAM) & rect);
						::MapWindowPoints (g_hToolbarWnd, HWND_DESKTOP, (LPPOINT) & rect, 2);
						TPMPARAMS tpm;
						tpm.cbSize = sizeof (tpm);
						tpm.rcExclude = rect;
   
						const HMENU hPopupMenu = ::CreatePopupMenu ();
						if (pNMT->iItem == ID_EDIT_UNDO) {
							size_t c = 0;
							for (actions_t::const_iterator i = g_Undo.end (); i != g_Undo.begin (); -- i, ++ c) {
								actions_t::const_iterator ic = i;
								-- ic;
								std::string strName;
								(* ic)->GetName (strName);
								::AppendMenu (hPopupMenu, MF_STRING, 2100 + c, strName.c_str ());

								if (c > 20)
									break;
							}
						} else if (pNMT->iItem == ID_EDIT_REDO) {
							size_t c = 0;
							for (actions_t::const_iterator i = g_Redo.end (); i != g_Redo.begin (); -- i, ++ c) {
								actions_t::const_iterator ic = i;
								-- ic;
								std::string strName;
								(* ic)->GetName (strName);
								::AppendMenu (hPopupMenu, MF_STRING, 2200 + c, strName.c_str ());

								if (c > 20)
									break;
							}
						}
						::TrackPopupMenuEx (
							hPopupMenu,
							TPM_RIGHTALIGN | TPM_LEFTBUTTON | TPM_VERTICAL,               
							rect.right, rect.bottom, _hWnd, & tpm
						);
						::DestroyMenu (hPopupMenu);
						return TBDDRET_DEFAULT;
					}
				}
			}
			break;
		}

		case WM_COMMAND: {
			const WORD wCmd = LOWORD (_wParam);

			// First, get chance for current mouse mode to process popup menu commands.
			if (! g_bLoading)
				if (CMouseMode::_OnCommand (wCmd))
					break;

			switch (wCmd) {
				//
				// File menu.
				//
//				case IDM_EXIT:          ::DestroyWindow (_hWnd); break;
				case ID_FILE_NEW:       if (! g_bLoading) RemoveMap (); break;

				case ID_FILE_ADD:
				case ID_FILE_OPEN:      if (! g_bLoading) CMapLoader::ShowOpenFile (wCmd == ID_FILE_OPEN); break;

				case ID_FILE_SAVE:      if (! g_bLoading) SaveMap (g_strFileName); break;
				case ID_FILE_SAVE_AS:   if (! g_bLoading) SaveMapAs (); break;

				case ID_MAP_PROPERTIES:	if (! g_bLoading) ShowMapProperties (_hWnd); break;
				case ID_RGN_PROPERTIES: if (! g_bLoading && g_Selection.size () == 1) ShowRgnProperties (g_Selection [0]); break;

				case ID_ERROR_LOG:      if (! g_bLoading) ShowErrorLog (); break;

				case ID_CANCEL_ACTION:
					if (g_bLoading)
						StopLoadThread ();
					else
						UnselectObjects ();

					CMouseMode::_OnEsc ();
					break;

				//
				// 'Edit' menu.
				//
				case ID_EDIT_UNDO: if (! g_bLoading) {UndoEditAction (); Refresh (); UpdateMenu ();} break;
				case ID_EDIT_REDO: if (! g_bLoading) {RedoEditAction (); Refresh (); UpdateMenu ();} break;

				case ID_EDIT_CUT:
				case ID_EDIT_COPY:
				case ID_EDIT_PASTE:
					// TODO: implement.
					break;

				case ID_EDIT_DELETE:
					if (! g_Selection.empty ()) {
						const size_t cSelItems = g_Selection.size ();

						int iDepth = 0;
						if (cSelItems > 1)
							iDepth = BeginCombinedAction ();

							//pCombined = new CCombinedAction ("delete multiple selection");
						for (size_t cSelItem = 0; cSelItem < cSelItems; ++ cSelItem) {
							find_info_t & SelItem = g_Selection [cSelItem];

							if (SelItem.pLayer) {
								if (SelItem.iRgn->IsDeleted ())
									continue;

								if (SelItem.iRgn->dwFlags & rgn_t::fWaypointOrTrack)
									DeleteAttachment (SelItem.iRgn->pFileRef);
								else if (SelItem.cElement != -1) {
									if (SelItem.iRgn->elements [SelItem.cElement].bDeleted)
										continue;
									DeleteElement (SelItem.pLayer, SelItem.iRgn, SelItem.cElement);
								} else if (SelItem.cOrigin != -1) {
									if (SelItem.iRgn->origins [SelItem.cOrigin].bDeleted)
										continue;
									DeleteElement (SelItem.pLayer, SelItem.iRgn, SelItem.cOrigin);
								} else
									DeleteRegion (SelItem.pLayer, SelItem.iRgn);
							} else if (SelItem.pImage) {
								if (SelItem.pImage->bDeleted)
									continue;
								DeleteImage (SelItem.pImage);
							}
						}

						if (cSelItems > 1)
							EndCombinedAction (iDepth, "delete [multiple objects]");

						g_Selection.clear ();
						Redraw ();
					}
					break;

				case ID_UNSELECT:    if (! g_bLoading) UnselectObjects (); break;

				case ID_MODE_DRAG:        CMouseMode::Set (0); break;
				case ID_MODE_ZOOM:        CMouseMode::Set (1); break;
				case ID_MODE_SELECT:      CMouseMode::Set (2); break;
				case ID_MODE_MOVE_POINTS: CMouseMode::Set (3); break;
				case ID_MODE_AT:          CMouseMode::Set (4); break;

				//
				// View menu.
				//
				case ID_REFRESH:     if (! g_bLoading) Refresh (); UpdateMenu (); break;

				case ID_ZOOM_IN:     if (! g_bLoading) SetScale (g_fScale/c_Sqrt4); break;
				case ID_ZOOM_OUT:    if (! g_bLoading) SetScale (g_fScale*c_Sqrt4); break;

				case ID_ZOOM_10M:    if (! g_bLoading) SetScale (c_Scale_10m); break;
				case ID_ZOOM_15M:    if (! g_bLoading) SetScale (c_Scale_15m); break;
				case ID_ZOOM_20M:    if (! g_bLoading) SetScale (c_Scale_20m); break;
				case ID_ZOOM_30M:    if (! g_bLoading) SetScale (c_Scale_30m); break;
				case ID_ZOOM_50M:    if (! g_bLoading) SetScale (c_Scale_50m); break;
				case ID_ZOOM_70M:    if (! g_bLoading) SetScale (c_Scale_70m); break;
				case ID_ZOOM_100M:   if (! g_bLoading) SetScale (c_Scale_100m); break;
				case ID_ZOOM_150M:   if (! g_bLoading) SetScale (c_Scale_150m); break;
				case ID_ZOOM_200M:   if (! g_bLoading) SetScale (c_Scale_200m); break;
				case ID_ZOOM_300M:   if (! g_bLoading) SetScale (c_Scale_300m); break;
				case ID_ZOOM_500M:   if (! g_bLoading) SetScale (c_Scale_500m); break;
				case ID_ZOOM_700M:   if (! g_bLoading) SetScale (c_Scale_700m); break;
				case ID_ZOOM_1KM:    if (! g_bLoading) SetScale (c_Scale_1km); break;
				case ID_ZOOM_1_5KM:  if (! g_bLoading) SetScale (c_Scale_1_5km); break;
				case ID_ZOOM_2KM:    if (! g_bLoading) SetScale (c_Scale_2km); break;
				case ID_ZOOM_3KM:    if (! g_bLoading) SetScale (c_Scale_3km); break;
				case ID_ZOOM_5KM:    if (! g_bLoading) SetScale (c_Scale_5km); break;
				case ID_ZOOM_7KM:    if (! g_bLoading) SetScale (c_Scale_7km); break;
				case ID_ZOOM_10KM:   if (! g_bLoading) SetScale (c_Scale_10km); break;
				case ID_ZOOM_15KM:   if (! g_bLoading) SetScale (c_Scale_15km); break;
				case ID_ZOOM_20KM:   if (! g_bLoading) SetScale (c_Scale_20km); break;
				case ID_ZOOM_30KM:   if (! g_bLoading) SetScale (c_Scale_30km); break;
				case ID_ZOOM_50KM:   if (! g_bLoading) SetScale (c_Scale_50km); break;
				case ID_ZOOM_70KM:   if (! g_bLoading) SetScale (c_Scale_70km); break;
				case ID_ZOOM_100KM:  if (! g_bLoading) SetScale (c_Scale_100km); break;
				case ID_ZOOM_150KM:  if (! g_bLoading) SetScale (c_Scale_150km); break;
				case ID_ZOOM_200KM:  if (! g_bLoading) SetScale (c_Scale_200km); break;
				case ID_ZOOM_300KM:  if (! g_bLoading) SetScale (c_Scale_300km); break;
				case ID_ZOOM_500KM:  if (! g_bLoading) SetScale (c_Scale_500km); break;
				case ID_ZOOM_700KM:  if (! g_bLoading) SetScale (c_Scale_700km); break;
				case ID_ZOOM_1000KM: if (! g_bLoading) SetScale (c_Scale_1000km); break;
				case ID_ZOOM_1200KM: if (! g_bLoading) SetScale (c_Scale_1200km); break;
				case ID_ZOOM_1500KM: if (! g_bLoading) SetScale (c_Scale_1500km); break;
				case ID_ZOOM_2000KM: if (! g_bLoading) SetScale (c_Scale_2000km); break;
				case ID_ZOOM_3000KM: if (! g_bLoading) SetScale (c_Scale_3000km); break;

				case ID_ZOOM_FULL: {
					if (g_bLoading)
						break;

					bool bRectBoundInvalid;
					{
						CCritSec::CLock lock (g_cs);
						bRectBoundInvalid = g_map.rectBound.Invalid ();
					}
					if (! bRectBoundInvalid)
						SetFullMapScale ();
					break;
				}

				case ID_VIEW_ALL_LEVELS: if (! g_bLoading) SetAllLevelsVisible (); break;

				case ID_SHOW_MINI:
					if (::IsWindowVisible (g_hMiniMapWnd)) {
						::ShowWindow (g_hMiniMapWnd, SW_HIDE);
						SaveSetting ("ShowMiniMap", false);
					} else {
						::ShowWindow (g_hMiniMapWnd, SW_SHOW);
						SaveSetting ("ShowMiniMap", true);
					}
					UpdateMenu ();
					break;

				case ID_SHOW_RULER:
					if (::GetMenuState (::GetMenu (g_hWnd), ID_SHOW_RULER, MF_BYCOMMAND) & MF_CHECKED)
						SaveSetting ("ShowRuler", false);
					else
						SaveSetting ("ShowRuler", true);
					UpdateMenu ();
					Redraw ();
					break;

				case ID_SHOW_GRID:
					if (::GetMenuState (::GetMenu (g_hWnd), ID_SHOW_GRID, MF_BYCOMMAND) & MF_CHECKED)
						SaveSetting ("ShowGrid", false);
					else
						SaveSetting ("ShowGrid", true);
					UpdateMenu ();
					Redraw ();
					break;

				case ID_SHOW_LABELS:
					if (::GetMenuState (::GetMenu (g_hWnd), ID_SHOW_LABELS, MF_BYCOMMAND) & MF_CHECKED)
						SaveSetting ("ShowLabels", false);
					else
						SaveSetting ("ShowLabels", true);
					UpdateMenu ();
					Refresh ();
					break;

				case ID_TRANSPARENT_POLYGONS:
					if (::GetMenuState (::GetMenu (g_hWnd), ID_TRANSPARENT_POLYGONS, MF_BYCOMMAND) & MF_CHECKED)
						SaveSetting ("TransparentPolygons", false);
					else
						SaveSetting ("TransparentPolygons", true);
					UpdateMenu ();
					Refresh ();
					break;

				//
				// 'Help' menu.
				//
				case ID_GO_WEB:
					::ShellExecute (g_hWnd, NULL, "http://kgy.narod.ru/util/mapedit/mapedit_e.htm", NULL, NULL, SW_SHOWNORMAL);
					break;

				case IDM_ABOUT:
					ShowAboutBox (_hWnd);
					break;

				default:
					if (2000 <= wCmd && wCmd < 2100) {
						if (! g_bLoading)
							LoadRecentFile (wCmd - 2000);
					} else if (2100 <= wCmd && wCmd < 2200) {
						const size_t cTimes = wCmd - 2100 + 1;
						for (size_t c = 0; c < cTimes; ++ c)
							UndoEditAction ();
						Refresh ();
						UpdateMenu ();
					} else if (2200 <= wCmd && wCmd < 2300) {
						const size_t cTimes = wCmd - 2200 + 1;
						for (size_t c = 0; c < cTimes; ++ c)
							RedoEditAction ();
						Refresh ();
						UpdateMenu ();
					} else if (2300 <= wCmd && wCmd < 2400)
						SetVisibleLevel (wCmd - 2300);

					break;

				case ID_SCALES_CB: {
					switch (HIWORD (_wParam)) {
						case CBN_SELCHANGE: {
							if (! g_hScaleCB)
								break;
							const size_t cIndex = ::SendMessage (g_hScaleCB, CB_GETCURSEL, 0, 0);
							const size_t cCmd = ::SendMessage (g_hScaleCB, CB_GETITEMDATA, cIndex, 0);
							::SetFocus (g_hWnd);
							::SendMessage (g_hWnd, WM_COMMAND, MAKELONG (cCmd, 0), 0);
							break;
						}

						case CBN_SETFOCUS:
							g_bDialogMode = true;
							break;	

						case CBN_KILLFOCUS:
							g_bDialogMode = false;
							break;	
					}
					break;
				}
			}
			break;
		}

		case WM_DROPFILES: {
			const HDROP hDrop = reinterpret_cast <HDROP> (_wParam);

			file_names_t file_names;
			size_t c = 0;
			while (true) {
				char strFileName [MAX_PATH + 1];
				if (::DragQueryFile (hDrop, c ++, strFileName, MAX_PATH) == 0)
					break;

				file_names.push_back (strFileName);
			}
			::DragFinish (hDrop);

			LoadMap (file_names, 0, false);

			break;
		}

			// Emulate scrolling by keyboard.
		case WM_KEYDOWN:
			switch (_wParam) { 
				case VK_LEFT:  KbHScroll (SB_LINELEFT, 0);  break;
				case VK_RIGHT: KbHScroll (SB_LINERIGHT, 0); break;
				case VK_UP:    KbVScroll (SB_LINEUP, 0);    break;
				case VK_DOWN:  KbVScroll (SB_LINEDOWN, 0);  break;

				// TODO: optimize
 				case VK_PRIOR: KbHScroll (SB_LINERIGHT, 0); KbVScroll (SB_LINEUP, 0);   break;
 				case VK_NEXT:  KbHScroll (SB_LINERIGHT, 0); KbVScroll (SB_LINEDOWN, 0);	break;
 				case VK_HOME:  KbHScroll (SB_LINELEFT, 0);  KbVScroll (SB_LINEUP, 0);   break;
 				case VK_END:   KbHScroll (SB_LINELEFT, 0);  KbVScroll (SB_LINEDOWN, 0); break;
			} 
 			break; 

		case WM_KEYUP:
			StopKbScroll ();
			break;

		case WM_LBUTTONDOWN:
		case WM_RBUTTONDOWN:
			// Return keyboard input from label edit box.
			//if (g_bDialogMode)
				::SetFocus (_hWnd);
			break;

		case WM_CLOSE:
			if (g_bLoading)
				StopLoadThread ();

			if (! RequestSave ())
				return 0;

			::DestroyWindow (_hWnd);  
			return 0;

		case WM_DESTROY:
			::PostQuitMessage (0);
			::TerminateProcess (::GetCurrentProcess (), 0);
			break;

			// Draw progress bar (requested by status bar).
		case WM_DRAWITEM: {
			const DRAWITEMSTRUCT * const pDI = reinterpret_cast<DRAWITEMSTRUCT *> (_lParam);

			const x0 = pDI->rcItem.left + 2;

			RECT rect;
			rect.left = x0;
			rect.top = pDI->rcItem.top + 2;
			rect.right = x0 + pDI->itemData;
			rect.bottom = pDI->rcItem.bottom - 2;
			::FillRect (pDI->hDC, & rect, g_hActiveColorBrush);

			rect.left = x0 + pDI->itemData;
			rect.right = x0 + 100;
			::FillRect (pDI->hDC, & rect, g_hBackColorBrush);
		}

		default:
			return ::DefWindowProc (_hWnd, _uMsg, _wParam, _lParam);
   }

   return FALSE;
}

