# include "stdafx.h"
# include "Resource.h"
# include "Map.h"
# include "Types.h"
# include "Load.h"
# include "MapProperties.h"
# include "Globals.h"

int CALLBACK MapPropertiesProc (HWND _hWnd, UINT _uMsg, WPARAM _wParam, LPARAM);
int CALLBACK MapHeaderProc (HWND _hWnd, UINT _uMsg, WPARAM _wParam, LPARAM);

void ShowMapProperties (HWND _hWnd) {
	if (g_bLoading)
		return;

	{
		CCritSec::CLock lock (g_cs);
		if (g_map.empty ())
			return;
	}

	::DialogBox (g_hInst, MAKEINTRESOURCE (IDD_MAP_PROPERTIES), _hWnd, & MapPropertiesProc);
}

static
void AddColumn (HWND _hTypes, int _i, const char * _strText, int _w) {
	LVCOLUMN clm;
	clm.mask = LVCF_FMT | LVCF_TEXT | LVCF_WIDTH;
	clm.fmt = LVCFMT_LEFT;
	clm.cx = _w;
	clm.pszText = const_cast<char *> (_strText);
	clm.cchTextMax = strlen (_strText);
	ListView_InsertColumn (_hTypes, _i, & clm);
}

static HWND g_hTypes;
static int g_iSortColumn = -1;
static bool g_bAccent = false;

static
int CALLBACK CompareItems (LPARAM _lParam1, LPARAM _lParam2, LPARAM) {
	char str1 [64], str2 [64];
	ListView_GetItemText (g_hTypes, _lParam1, g_iSortColumn, str1, 64); 
	ListView_GetItemText (g_hTypes, _lParam2, g_iSortColumn, str2, 64);

	if (g_iSortColumn == 4 || g_iSortColumn == 5) {
		// Compare as integers.
		return g_bAccent ? atoi (str1) - atoi (str2) : atoi (str2) - atoi (str1);
	} else
		// Compare as strings.
		return g_bAccent ? strcmp (str1, str2) : strcmp (str2, str1);
} 

int CALLBACK MapPropertiesProc (HWND _hWnd, UINT _uMsg, WPARAM _wParam, LPARAM _lParam) {
	switch (_uMsg) {
		case WM_INITDIALOG: {
			CCritSec::CLock lock (g_cs);

			char strBuf [64];

			// Fill info about the map rectangle.
			::sprintf (strBuf, "%s %.6f", g_map.rectBound.x0 > 0 ? "E" : "W", ::fabs (g_map.rectBound.x0));
			::SetDlgItemText (_hWnd, IDC_LEFT, strBuf);
			::sprintf (strBuf, "%s %.6f", g_map.rectBound.x1 > 0 ? "E" : "W", ::fabs (g_map.rectBound.x1));
			::SetDlgItemText (_hWnd, IDC_RIGHT, strBuf);
			::sprintf (strBuf, "%s %.6f", g_map.rectBound.y0 > 0 ? "N" : "S", ::fabs (g_map.rectBound.y0));
			::SetDlgItemText (_hWnd, IDC_BOTTOM, strBuf);
			::sprintf (strBuf, "%s %.6f", g_map.rectBound.y1 > 0 ? "N" : "S", ::fabs (g_map.rectBound.y1));
			::SetDlgItemText (_hWnd, IDC_TOP, strBuf);

			size_t cPoints = 0, cPolylines = 0, cPolygons = 0;
			size_t cMaxVertices = 0, cMaxElements = 0, cTotalPoints = 0;
			
			// Prepare list view control for types.			
			g_hTypes = ::GetDlgItem (_hWnd, IDC_TYPES);

			ListView_SetExtendedListViewStyleEx (g_hTypes, LVS_EX_FULLROWSELECT, LVS_EX_FULLROWSELECT);

			AddColumn (g_hTypes, 0, "Code", 50);
			AddColumn (g_hTypes, 1, "Kind", 50);
			AddColumn (g_hTypes, 2, "Description", 200);
			AddColumn (g_hTypes, 3, "Count", 50);
			AddColumn (g_hTypes, 4, "Size", 60);

			LVITEM lvi;
			lvi.mask = LVIF_TEXT | LVIF_STATE;
			lvi.state = 0;
			lvi.stateMask = 0;
			lvi.iItem = 0;
		    lvi.pszText = strBuf;
			lvi.cchTextMax = sizeof (strBuf)/sizeof (strBuf [0]);

			// The array of counters being parallel to g_types.
			size_t * const pCounters = reinterpret_cast<size_t *> (alloca (sizeof (size_t)*g_cTypes));
			::memset (pCounters, 0, sizeof (size_t)*g_cTypes);
			size_t * const pSizes = reinterpret_cast<size_t *> (alloca (sizeof (size_t)*g_cTypes));
			::memset (pSizes, 0, sizeof (size_t)*g_cTypes);

			for (int cLayer = 0; cLayer < 4; ++ cLayer) {
				const rgns_t & rgns = g_map.layers [cLayer].rgns;

				// Count types for the given layer.
				for (rgns_t::const_iterator iRgn = rgns.begin (); iRgn != rgns.end (); ++ iRgn) {
					const rgn_t & rgn = * iRgn;
					if (rgn.IsDeleted () || (rgn.dwFlags & rgn_t::fWaypointOrTrack))
						continue;

					if (rgn.IsPolyline ())
						++ cPolylines;
					else if (rgn.IsPolygon ())
						++ cPolygons;
					else
						++ cPoints;

					const size_t cElements = rgn.elements.size ();
					if (cElements > cMaxElements)
						cMaxElements = cElements;

					for (size_t cElement = 0; cElement < cElements; ++ cElement) {
						const rgn_element_t & element = rgn.elements [cElement];
						if (element.bDeleted)
							continue;

						const size_t cPoints = element.points.size ();
						cTotalPoints += cPoints;

						if (rgn.elements [cElement].points.size () > cMaxVertices)
							cMaxVertices = cPoints;
					}

					const size_t cOrigins = rgn.origins.size ();
					for (size_t cOrigin = 0; cOrigin < cOrigins; ++ cOrigin) {
						if (rgn.origins [cOrigin].bDeleted)
							continue;
						++ cTotalPoints;
					}

					++ (pCounters [rgn.pTypeDesc - g_types]);
					if (rgn.cEndPos > rgn.cBeginPos)
						// NOTE: rgn.cEndPos may be incorrect for damaged files (w/o correct end).
						pSizes [rgn.pTypeDesc - g_types] += rgn.cEndPos - rgn.cBeginPos;
				}
			}

			// For each type.
			// NOTE: omit 0-th type ("unknown").
			for (size_t cType = 1; cType < g_cTypes; ++ cType) {
				if (pCounters [cType] == 0)
					continue;

				// Code
				::sprintf (strBuf, "0x%04x", g_types [cType].ulType & 0xFFFF);
				lvi.iSubItem = 0;
				ListView_InsertItem (g_hTypes, & lvi);

				// Kind
				assert (g_types [cType].ushKind < 4);
				::strcpy (strBuf, g_strKinds [g_types [cType].ushKind]);
				lvi.iSubItem = 1;
				ListView_SetItem (g_hTypes, & lvi);

				// Description
				::sprintf (strBuf, "%s", g_types [cType].strTypeName);
				lvi.iSubItem = 2;
				ListView_SetItem (g_hTypes, & lvi);

				// Count
				::sprintf (strBuf, "%d", pCounters [cType]);
				lvi.iSubItem = 3;
				ListView_SetItem (g_hTypes, & lvi);

				// Size
				::sprintf (strBuf, "%d", pSizes [cType]);
				lvi.iSubItem = 4;
				ListView_SetItem (g_hTypes, & lvi);

				++ lvi.iItem;
			}

			::sprintf (strBuf, "%d", cPoints);
			::SetDlgItemText (_hWnd, IDC_POINTS, strBuf);
			::sprintf (strBuf, "%d", cPolygons);
			::SetDlgItemText (_hWnd, IDC_POLYGONS, strBuf);
			::sprintf (strBuf, "%d", cPolylines);
			::SetDlgItemText (_hWnd, IDC_POLYLINES, strBuf);
			::sprintf (strBuf, "%d", cPoints + cPolygons + cPolylines);
			::SetDlgItemText (_hWnd, IDC_TOTAL, strBuf);

			::sprintf (strBuf, "%d", cMaxVertices);
			::SetDlgItemText (_hWnd, IDC_MAX_VERTICES, strBuf);
			::sprintf (strBuf, "%d", cMaxElements);
			::SetDlgItemText (_hWnd, IDC_MAX_ELEMENTS, strBuf);
			::sprintf (strBuf, "%d", cTotalPoints);
			::SetDlgItemText (_hWnd, IDC_TOTAL_POINTS, strBuf);

			return TRUE;
		}

		case WM_NOTIFY: {
			const NMHDR * const pInfo = reinterpret_cast<const NMHDR *> (_lParam);
			switch (pInfo->code) {
				case LVN_COLUMNCLICK:
					const NMLISTVIEW * const pInfo2 = reinterpret_cast<const NMLISTVIEW *> (_lParam);

					if (g_iSortColumn != pInfo2->iSubItem)
						g_iSortColumn = pInfo2->iSubItem;
					else
						g_bAccent = ! g_bAccent;
					
					ListView_SortItemsEx (g_hTypes, CompareItems, 0);

					break;
			}
			break;
		}

		case WM_COMMAND:
			switch (LOWORD (_wParam)) {
				case IDOK:
				case IDCANCEL:
					// On closing dialog.
					::EndDialog (_hWnd, 0);
					return TRUE;

				case IDC_SHOW_HEADER:
					::DialogBox (g_hInst, MAKEINTRESOURCE (IDD_MAP_HEADER), _hWnd, & MapHeaderProc);
					return TRUE;
			}
	}

	return FALSE;
}

static
int CALLBACK MapHeaderProc (HWND _hWnd, UINT _uMsg, WPARAM _wParam, LPARAM _lParam) {
	switch (_uMsg) {
		case WM_INITDIALOG: {
			//
			// Prepare source edit box.
			//
			const CHandle hFile = ::CreateFile (g_strFileName, GENERIC_READ, 0, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
			if (hFile == NULL || hFile == INVALID_HANDLE_VALUE) {
				::SetDlgItemText (_hWnd, IDC_SOURCE, "Can't open the file.");
				return TRUE;
			}

			char * const pSource = reinterpret_cast<char *> (alloca (sizeof (char)*g_map.cHeaderLen + 1));

			::SetFilePointer (hFile, 0, 0, FILE_BEGIN);
			DWORD dwFileSizeRead = 0;
			::ReadFile (hFile, pSource, g_map.cHeaderLen, & dwFileSizeRead, NULL);
			if (dwFileSizeRead < g_map.cHeaderLen) {
				::SetDlgItemText (_hWnd, IDC_SOURCE, "Can't read the file.");
				return TRUE;
			}
			pSource [g_map.cHeaderLen] = 0;

			::SetDlgItemText (_hWnd, IDC_SOURCE, pSource);

			return TRUE;
		}

		case WM_COMMAND:
			if (LOWORD (_wParam) == IDOK || LOWORD (_wParam) == IDCANCEL) {
				// On closing dialog.
				::EndDialog (_hWnd, 0);
				return TRUE;
			}
	}

	return FALSE;
}
