# include "stdafx.h"
# include "Resource.h"
# include "MapView.h"

# include "Menu.h"
# include "Status.h"
# include "Toolbar.h"
# include "Settings.h"

# include "Map.h"
# include "Load.h"
# include "RgnProperties.h"
# include "Rename.h"
# include "AffTransform.h"
# include "Label.h"

# include "Ruler.h"
# include "Grid.h"

# include "Scale.h"
# include "Position.h"

# include "Selection.h"
# include "MouseMode.h"
# include "Globals.h"

LRESULT CALLBACK MapWndProc (HWND, UINT, WPARAM, LPARAM);
LRESULT CALLBACK MiniMapWndProc (HWND, UINT, WPARAM, LPARAM);

HWND g_hMapWnd;
HWND g_hMapTipWnd;
HWND g_hMiniMapWnd;

HBITMAP g_hBmp;
HDC g_hMemDC;

HBITMAP g_hMiniBmp;
HDC g_hMiniMemDC;
long g_lMiniW, g_lMiniH;

bool InitMapWindow () {
	//
	// Estimate size of map window based on sizes of other windows.
	//

	assert (g_hWnd);
	RECT rectClient;
	::GetClientRect (g_hWnd, & rectClient);
	const long lMainW = rectClient.right - rectClient.left;
	const long lMainH = rectClient.bottom - rectClient.top;

	assert (g_hStatusWnd);
	RECT rectStatus;
	::GetClientRect (g_hStatusWnd, & rectStatus);
	const long lStatusH = rectStatus.bottom - rectStatus.top;

	assert (g_hToolbarWnd);
	RECT rectToolbar;
	// NOTE: GetWindowRect() returns incorrect size for toolbar.
	::GetClientRect (g_hToolbarWnd, & rectToolbar);
	const long lToolbarH = rectToolbar.bottom - rectToolbar.top - 1;

	//
	// Register the window class.
	//

	WNDCLASSEX wcex_map;
	::memset (& wcex_map, 0, sizeof (wcex_map));
	wcex_map.cbSize = sizeof (WNDCLASSEX); 
	wcex_map.style			= CS_HREDRAW | CS_VREDRAW | CS_OWNDC | CS_DBLCLKS;
	wcex_map.lpfnWndProc	= (WNDPROC) MapWndProc;
	wcex_map.hInstance		= g_hInst;
	wcex_map.hCursor		= g_hPermanentCursor;
	wcex_map.hbrBackground	= g_hTmpBrush;
	wcex_map.lpszClassName	= "mapview_map";
	::RegisterClassEx (& wcex_map);

	//
	// Create the map window.
	//

	g_hMapWnd = ::CreateWindowEx (
		WS_EX_CLIENTEDGE,
		"mapview_map", NULL,
		WS_CHILD | WS_HSCROLL | WS_VSCROLL | WS_VISIBLE | WS_CLIPSIBLINGS | WS_CLIPCHILDREN,
		0, lToolbarH, lMainW, lMainH - lStatusH - lToolbarH,
		g_hWnd, NULL, g_hInst, NULL
	);
	if (! g_hMapWnd)
		return false;

	// Set above toolbar.
	// NOTE: toolbar auto-sizes its window rect incorrectly while client one is OK.
	::SetWindowPos (g_hMapWnd, HWND_TOP, 0, 0, 0, 0, SWP_NOMOVE | SWP_NOSIZE);

	//
	// Cteate scrollers.
	//

	SCROLLINFO si;
	si.cbSize = sizeof (si);
	si.fMask = SIF_ALL | SIF_DISABLENOSCROLL;
	si.nMin  = 0;
	si.nMax  = 1;
	si.nPage = 2;
	si.nPos  = 0;
	::SetScrollInfo (g_hMapWnd, SB_HORZ, & si, TRUE);
	::SetScrollInfo (g_hMapWnd, SB_VERT, & si, TRUE);

	//
	// Create tooltip for the map window.
	//

	g_hMapTipWnd = ::CreateWindowEx (
		NULL, TOOLTIPS_CLASS, NULL, WS_POPUP | TTS_NOPREFIX | TTS_ALWAYSTIP,
		CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT,
		g_hMapWnd, NULL, g_hInst, NULL
	);
	if (! g_hMapTipWnd)
		return false;

	TOOLINFO ti;
	::memset (& ti, 0, sizeof (ti));
	ti.cbSize = sizeof (ti);
	ti.uFlags = TTF_IDISHWND | TTF_TRACK | TTF_ABSOLUTE | TTF_TRANSPARENT;
	ti.hwnd = g_hMapWnd;
	ti.uId = reinterpret_cast<DWORD> (g_hMapWnd);
	// NOTE: this dummy text is required to make tooltip work.
	ti.lpszText = "map";
	::SendMessage (g_hMapTipWnd, TTM_ADDTOOL, 0, reinterpret_cast<LPARAM> (& ti));

	//
	// Create edit box for labels.
	//
	if (! InitLabelEdit ())
		return false;

	//
	// Create the mini-map window.
	//

	WNDCLASSEX wcex_mm;
	::memset (& wcex_mm, 0, sizeof (wcex_map));
	wcex_mm.cbSize		= sizeof (WNDCLASSEX); 
	wcex_mm.style		= CS_HREDRAW | CS_VREDRAW | CS_OWNDC;
	wcex_mm.lpfnWndProc	= (WNDPROC) MiniMapWndProc;
	wcex_mm.hInstance	= g_hInst;
	wcex_mm.hCursor		= g_hArrowCursor;
	wcex_mm.lpszClassName = "mapview_mm";
	::RegisterClassEx (& wcex_mm);

	g_hMiniMapWnd = ::CreateWindowEx (
		WS_EX_WINDOWEDGE | WS_EX_TOOLWINDOW | WS_EX_TOPMOST,
		"mapview_mm", "Mini map",
		WS_CHILDWINDOW | WS_CAPTION | WS_CLIPSIBLINGS | WS_CLIPCHILDREN | WS_SYSMENU |
		(GetSetting ("ShowMiniMap", true) ? WS_VISIBLE : 0),
		GetSetting ("MiniMapX", DWORD (0)), GetSetting ("MiniMapY", DWORD (0)), 1, 1,
		g_hMapWnd, NULL, g_hInst, NULL
	);
	if (! g_hMiniMapWnd)
		return false;

	return true;
}

////////////////////////////////////////////////

static
long g_lPrevWidth, g_lPrevHeight;

void Refresh () {
	// Update scrollers.
	SCROLLINFO si;
	si.cbSize = sizeof (si);
	si.fMask = SIF_ALL | SIF_DISABLENOSCROLL;
	si.nMin  = 0;
	si.nMax  = g_lTotalWidth - 1;
	si.nPage = g_lWidth;
	si.nPos  = g_lX;
	::SetScrollInfo (g_hMapWnd, SB_HORZ, & si, TRUE);
	si.nMin  = 0;
	si.nMax  = g_lTotalHeight - 1;
	si.nPage = g_lHeight;
	si.nPos  = g_lTotalHeight - g_lHeight - g_lY;
	::SetScrollInfo (g_hMapWnd, SB_VERT, & si, TRUE);

	if (g_lWidth != g_lPrevWidth || g_lHeight != g_lPrevHeight || ! g_hBmp || ! g_hMemDC) {
		// Destroy old resources.
		if (g_hBmp)
			::DeleteObject (g_hBmp);
		if (g_hMemDC)
			::DeleteDC (g_hMemDC);

		// Prepare memory bitmap.
		const CClientDC hClientDC (g_hMapWnd);
		g_hBmp = ::CreateCompatibleBitmap (hClientDC, g_lWidth, g_lHeight);
		g_hMemDC = ::CreateCompatibleDC (hClientDC);
		::SelectObject (g_hMemDC, g_hBmp);

		g_lPrevWidth  = g_lWidth;
		g_lPrevHeight = g_lHeight;
	}

	// Render map to the bitmap.
	{
		CWaiter w;
		CCritSec::CLock lock (g_cs);
		g_map.Draw (g_hMemDC);
	}

	// Copy bitmap to the window.
	Redraw ();

	// Update selected rect in mini-map.
	RedrawMini ();

	// Update coordinates.
	POINT pt;
	::GetCursorPos (& pt);
	::ScreenToClient (g_hMapWnd, & pt);
	point_t point;
	{
		CCritSec::CLock lock (g_cs);
		if (g_map.rectBound.Invalid ())
			return;
		ScreenToPt (pt.x, pt.y, point);
	}
	SetCoordinatesStatus (point.x, point.y);
}

void Redraw (RECT * _pRect) {
	::InvalidateRect (g_hMapWnd, _pRect, FALSE);
	::UpdateWindow (g_hMapWnd);
}

////////////////////////////////////////////////

	// Shift in pixels during scrolling.
static long g_lShiftX, g_lShiftY;
	// Mode of keyboard scrolling.
static bool g_bHScroll, g_bVScroll;

void KbHScroll (long _Type, long _lX) {
	g_bHScroll = true;
	::SendMessage (g_hMapWnd, WM_HSCROLL, MAKELONG (_Type, 0), 0);
}

void KbVScroll (long _Type, long _lY) {
	g_bVScroll = true;
	::SendMessage (g_hMapWnd, WM_VSCROLL, MAKELONG (_Type, 0), 0);
}

void StopKbScroll () {
	if (g_bHScroll)
		::PostMessage (g_hMapWnd, WM_HSCROLL, MAKELONG (SB_ENDSCROLL, 0), 0);
	if (g_bVScroll)
		::PostMessage (g_hMapWnd, WM_VSCROLL, MAKELONG (SB_ENDSCROLL, 0), 0);
	g_bHScroll = g_bVScroll = false;
}

void ShiftMap (long _lShiftX, long _lShiftY) {
	const CClientDC hDC (g_hMapWnd);

	RECT rect;
	rect.top = 0;
	rect.bottom = g_lHeight;
	if (_lShiftX > 0) {
		rect.left = g_lWidth - _lShiftX;
		rect.right = g_lWidth;
	} else {
		rect.left = 0;
		rect.right = -_lShiftX;
	}
	::FillRect (hDC, & rect, g_hTmpBrush);

	rect.left = 0;
	rect.right = g_lWidth;
	if (_lShiftY > 0) {
		rect.top = 0;
		rect.bottom = _lShiftY;
	} else {
		rect.top = g_lHeight + _lShiftY;
		rect.bottom = g_lHeight;
	}
	::FillRect (hDC, & rect, g_hTmpBrush);

	::BitBlt (
		hDC,
		-_lShiftX, _lShiftY, g_lWidth, g_lHeight,
		g_hMemDC, 0, 0, SRCCOPY
	);
}

////////////////////////////////////////////////

const c_cTipTimerID = 10;
const c_cTipTimerTimeout = 500;

static bool g_bMapToolipVisible = false;

static
void ActivateMapTooltip (bool _bActivate) {
	if (_bActivate == g_bMapToolipVisible)
		return;
	g_bMapToolipVisible = _bActivate;

	assert (g_hMapWnd);

	TOOLINFO ti;
	::memset (& ti, 0, sizeof (ti));
	ti.cbSize = sizeof (ti);
	ti.uFlags = TTF_IDISHWND | TTF_TRACK | TTF_ABSOLUTE | TTF_TRANSPARENT;
	ti.hwnd = g_hMapWnd;
	ti.uId = reinterpret_cast<DWORD> (g_hMapWnd);
	ti.lpszText = "map";
	::SendMessage (g_hMapTipWnd, TTM_TRACKACTIVATE, _bActivate ? TRUE : FALSE, reinterpret_cast<LPARAM> (& ti));
}

////////////////////////////////////////////////

long g_lRClickX;
long g_lRClickY;

////////////////////////////////////////////////

const c_cSizeTimerID = 11;
const c_cSizeTimerTimeout = 100;

LRESULT CALLBACK MapWndProc (HWND _hWnd, UINT _uMsg, WPARAM _wParam, LPARAM _lParam) {
	switch (_uMsg) {
		case WM_PAINT: {
			PAINTSTRUCT ps;
			const HDC hDC = ::BeginPaint (_hWnd, & ps);

			// Draw cached map image.
			::BitBlt (hDC, 0, 0, g_lWidth, g_lHeight, g_hMemDC, 0, 0, SRCCOPY);
			DrawSelection (hDC, g_Selection);

			CMouseMode::_OnPaint (hDC);

			// Show ruler.
			if (GetSetting ("ShowRuler", true)) {
				RECT rect;
				::GetClientRect (_hWnd, & rect);
				ShowRuler (hDC, rect.left + 20, rect.bottom - 40);
			}

			// Show grid.
			if (GetSetting ("ShowGrid", true)) {
				RECT rect;
				::GetClientRect (_hWnd, & rect);
				ShowGrid (hDC, rect);
			}

			::EndPaint (_hWnd, & ps);

			// Hide map's tooltip for time-out.
			ActivateMapTooltip (false);
			::SetTimer (_hWnd, c_cTipTimerID, c_cTipTimerTimeout, NULL);

			break;
		}

			// Re-create the buffer bitmap for the changed color depth.
		case WM_DISPLAYCHANGE: Refresh (); break;

		case WM_SIZING:
			return TRUE;

		case WM_SIZE:
			::KillTimer (_hWnd, c_cSizeTimerID);
			::SetTimer (_hWnd, c_cSizeTimerID, c_cSizeTimerTimeout, NULL);

			g_lWidth  = LOWORD (_lParam);
			g_lHeight = HIWORD (_lParam);

			break;

		case WM_SETCURSOR:
			if (reinterpret_cast<HWND> (_wParam) == _hWnd && LOWORD (_lParam) == HTCLIENT)
				::SetCursor (g_hCursor);
			else
				::SetCursor (g_hArrowCursor);
			return TRUE;

		case WM_HSCROLL: {
			if (g_lTotalWidth <= g_lWidth)
				break;

			long lStep = 0;
			switch (LOWORD (_wParam)) {
				case SB_ENDSCROLL:
					if (! g_lShiftX)
						break;

					SetX (g_lX + g_lShiftX);
					g_lShiftX = 0;
					Refresh ();
					break;

				case SB_RIGHT:      lStep = (g_lTotalWidth - g_lWidth - 1) - (g_lX + g_lShiftX); break;
				case SB_LEFT:       lStep = -(g_lX + g_lShiftX); break;
				case SB_LINERIGHT:  lStep = g_lWidth/20; break;
				case SB_LINELEFT:   lStep = -g_lWidth/20; break;
				case SB_PAGERIGHT:  lStep = g_lWidth/2; break;
				case SB_PAGELEFT:   lStep = -g_lWidth/2; break;
				case SB_THUMBTRACK:	lStep = HIWORD (_wParam) - (g_lX + g_lShiftX); break;
			}
			if (g_lX + g_lShiftX + lStep < 0)
				lStep = -(g_lX + g_lShiftX);
			else if (g_lX + g_lShiftX + lStep >= g_lTotalWidth - g_lWidth)
				lStep = (g_lTotalWidth - g_lWidth - 1) - (g_lX + g_lShiftX);
			if (lStep == 0)
				break;

			g_lShiftX += lStep;

			// Update h-scroll bar.
			SCROLLINFO si;
			si.cbSize = sizeof (si);
			si.fMask = SIF_POS | SIF_DISABLENOSCROLL;
			si.nPos  = g_lX + g_lShiftX;
			::SetScrollInfo (g_hMapWnd, SB_HORZ, & si, TRUE);

			// Shift the map image.
			ShiftMap (g_lShiftX, 0);
			break;
		}

		case WM_VSCROLL: {
			if (g_lTotalHeight <= g_lHeight)
				break;

			long lStep = 0;
			switch (LOWORD (_wParam)) {
				case SB_ENDSCROLL:
					if (! g_lShiftY)
						break;

					SetY (g_lY + g_lShiftY);
					g_lShiftY = 0;
					Refresh ();
					break;

				case SB_TOP:           lStep = (g_lTotalHeight - g_lHeight - 1) - (g_lY + g_lShiftY); break;
				case SB_BOTTOM:        lStep = -(g_lY + g_lShiftY); break;
				case SB_LINEUP:        lStep = g_lHeight/20; break;
				case SB_LINEDOWN:      lStep = -g_lHeight/20; break;
				case SB_PAGEUP:        lStep = g_lHeight/2; break;
				case SB_PAGEDOWN:      lStep = -g_lHeight/2; break;
				case SB_THUMBTRACK:    lStep = (g_lTotalHeight - g_lHeight - 1 - HIWORD (_wParam)) - (g_lY + g_lShiftY); break;
			}
			if (g_lY + g_lShiftY + lStep < 0)
				lStep = -(g_lY + g_lShiftY);
			else if (g_lY + g_lShiftY + lStep >= g_lTotalHeight - g_lHeight)
				lStep = (g_lTotalHeight - g_lHeight - 1) - (g_lY + g_lShiftY);
			if (lStep == 0)
				break;

			g_lShiftY += lStep;

			// Update v-scroll bar.
			SCROLLINFO si;
			si.cbSize = sizeof (si);
			si.fMask = SIF_POS | SIF_DISABLENOSCROLL;
			si.nPos  = (g_lTotalHeight - g_lHeight - 1) - (g_lY + g_lShiftY);
			::SetScrollInfo (g_hMapWnd, SB_VERT, & si, TRUE);

			// Move the map image.
			ShiftMap (0, g_lShiftY);
			break;
		}

			// Begin of drag-and-drop.
		case WM_LBUTTONDOWN:
			// Return keyboard input from label edit box.
			//if (g_bDialogMode)
				::SetFocus (g_hWnd);

			// Disable tooltip timer.
			::KillTimer (_hWnd, c_cTipTimerID);

			if (g_bLoading)
				break;

			CMouseMode::_OnLDown (GET_X_LPARAM (_lParam), GET_Y_LPARAM (_lParam), _wParam);
			break;

		case WM_LBUTTONUP:
			CMouseMode::_OnLUp (GET_X_LPARAM (_lParam), GET_Y_LPARAM (_lParam), _wParam);
			break;

		case WM_MOUSEMOVE: {
			if (g_bLoading)
				break;

			// NOTE: Tooltip activation has side effect: the window gets WM_MOUSEMOVE.
			static LPARAM lOldXY = 0;
			if (lOldXY != _lParam)
				// Hide map's tooltip if cursor moves indeed.
				ActivateMapTooltip (false);
			lOldXY = _lParam;

			// Restart tooltip timer.
			if (! CMouseMode::PendingDND ())
				::SetTimer (_hWnd, c_cTipTimerID, c_cTipTimerTimeout, NULL);

			if (GetSetting ("SelectMode", (DWORD) 0) != 0 || ! CMouseMode::PendingDND ()) {
				point_t point;
				{
					CCritSec::CLock lock (g_cs);
					if (g_map.rectBound.Invalid ())
						break;
					ScreenToPt (GET_X_LPARAM (_lParam), GET_Y_LPARAM (_lParam), point);
				}
				SetCoordinatesStatus (point.x, point.y);
			}

			CMouseMode::_OnMove (GET_X_LPARAM (_lParam), GET_Y_LPARAM (_lParam), _wParam);
			break;
		}

		case WM_LBUTTONDBLCLK: {
			if (g_bLoading)
				break;

			CMouseMode::_OnLDblClk (GET_X_LPARAM (_lParam), GET_Y_LPARAM (_lParam), _wParam);
			break;
		}

		case WM_RBUTTONDOWN:
			// Return keyboard input from label edit box.
			//if (g_bDialogMode)
				::SetFocus (g_hWnd);

			CMouseMode::_OnRDown (GET_X_LPARAM (_lParam), GET_Y_LPARAM (_lParam), _wParam);
			break;

		case WM_RBUTTONUP: {
			if (CMouseMode::PendingDND () || g_bLoading)
				break;

			CMouseMode::_OnRUp (GET_X_LPARAM (_lParam), GET_Y_LPARAM (_lParam), _wParam);
			break;
		}

		case WM_TIMER:
			switch (_wParam) {
				case c_cSizeTimerID:
					::KillTimer (_hWnd, c_cSizeTimerID);

					if (g_bLoading)
						break;

					RefreshScales ();
					Refresh ();

					break;

				case c_cTipTimerID: {
					::KillTimer (_hWnd, c_cTipTimerID);

					// Get region under cursor.
					POINT ptClient;
					::GetCursorPos (& ptClient);
					::ScreenToClient (g_hMapWnd, & ptClient);
					if (
						0 > ptClient.x || ptClient.x > g_lWidth ||
						0 > ptClient.y || ptClient.y > g_lHeight
					)
						break;
					const find_info_t fi = FindRgn (ptClient.x, ptClient.y);
					if (fi.empty ()) {
						// Hide map's tooltip.
						ActivateMapTooltip (false);
						break;
					}

					// Prepare text for the tooltip.
					static char strTipText [128];
					if (fi.pLayer) {
						std::string strHintText;
						GetHintText (* fi.iRgn, strHintText);
						::strncpy (strTipText, strHintText.c_str (), sizeof (strTipText));
					} else if (fi.pImage) {
						::strncpy (strTipText, fi.pImage->strFileName.c_str (), sizeof (strTipText));
						strTipText [sizeof (strTipText) - 1];
					} else {
						assert (false);
						break;
					}

					// Set map's tooltip position.
					RECT rectTip;
					::GetClientRect (g_hMapTipWnd, & rectTip);
					const size_t cTipH = rectTip.bottom - rectTip.top;
					POINT pt;
					::GetCursorPos (& pt);
					::SendMessage (g_hMapTipWnd, TTM_TRACKPOSITION, 0, MAKELONG (pt.x + 3, pt.y - cTipH - 3));

					// Set the text.
					TOOLINFO ti;
					::memset (& ti, 0, sizeof (ti));
					ti.cbSize = sizeof (ti);
					ti.uFlags = TTF_IDISHWND | TTF_TRACK | TTF_ABSOLUTE | TTF_TRANSPARENT;
					ti.hwnd = g_hMapWnd;
					ti.uId = reinterpret_cast<DWORD> (g_hMapWnd);
					ti.lpszText = strTipText;
					::SendMessage (g_hMapTipWnd, TTM_UPDATETIPTEXT, 0, reinterpret_cast<LPARAM> (& ti));

					ActivateMapTooltip (true);
					break;
				}
			}
			break;

		default:
			return ::DefWindowProc (_hWnd, _uMsg, _wParam, _lParam);
   }

   return FALSE;
}

/////////////////////////////////////////////////////

void RedrawMini () {
	::InvalidateRect (g_hMiniMapWnd, NULL, FALSE);
	::UpdateWindow (g_hMiniMapWnd);
}

static
LRESULT CALLBACK MiniMapWndProc (HWND _hWnd, UINT _uMsg, WPARAM _wParam, LPARAM _lParam) {
	switch (_uMsg) {
		case WM_PAINT: {
			PAINTSTRUCT ps;
			const HDC hDC = ::BeginPaint (_hWnd, & ps);

			// Draw the frame.
			RECT rectFrame;
			::GetClientRect (_hWnd, & rectFrame);
			::FrameRect (hDC, & rectFrame, g_hBlackBrush);

			// Draw the mini-map.
			::BitBlt (hDC, 0, 0, g_lMiniW, g_lMiniH, g_hMiniMemDC, 0, 0, SRCCOPY);

			// Highlight the map area.
			if (g_lTotalWidth > 0 && g_lTotalHeight > 0) {
				RECT rect;
				rect.left = g_lX*g_lMiniW/g_lTotalWidth;
				rect.bottom = g_lMiniH - g_lY*g_lMiniH/g_lTotalHeight;
				rect.right = rect.left + g_lWidth*g_lMiniW/g_lTotalWidth;
				rect.top = rect.bottom - g_lHeight*g_lMiniH/g_lTotalHeight;
				::SetROP2 (hDC, R2_XORPEN);
				::FrameRect (hDC, & rect, g_hBlackBrush);
			}

			::EndPaint (_hWnd, & ps);
			break;
		}

		case WM_LBUTTONDOWN:
		case WM_RBUTTONDOWN:
			// Return keyboard input from label edit box.
			//if (g_bDialogMode)
				::SetFocus (g_hWnd);
			break;

		case WM_LBUTTONUP: {
			bool bRedraw = false;
			if (g_lTotalWidth > g_lWidth) {
				SetX ((GET_X_LPARAM (_lParam) - 1)*g_lTotalWidth/g_lMiniW - g_lWidth/2);
				bRedraw = true;
			}
			if (g_lTotalHeight > g_lHeight) {
				SetY ((g_lMiniH - (GET_Y_LPARAM (_lParam) - 1))*g_lTotalHeight/g_lMiniH - g_lHeight/2);
				bRedraw = true;
			}
			if (bRedraw)
				Refresh ();
			break;
		}

		case WM_MOVE: {
			RECT rect;
			::GetWindowRect (_hWnd, & rect);
			POINT p = {rect.left, rect.top};
			::ScreenToClient (g_hMapWnd, & p);
			if (p.x < 0)
				p.x = 0;
			if (p.y < 0)
				p.y = 0;

			SaveSetting ("MiniMapX", DWORD (p.x));
			SaveSetting ("MiniMapY", DWORD (p.y));
			break;
		}

		case WM_CLOSE:
			::ShowWindow (g_hMiniMapWnd, SW_HIDE);
			SaveSetting ("ShowMiniMap", false);
			UpdateMenu ();
			break;

		default:
			return ::DefWindowProc (_hWnd, _uMsg, _wParam, _lParam);
   }

   return FALSE;
}
