# include "stdafx.h"
# include "MathAlgo.h"
# include "PtAlgo.h"

	// Solve system of 3 linear equations:  B + Ax*x[i] + Ay*y[i] = screen[i]
	// to get values of B, Ax and Ay from known (x,y)->screen.
bool Solve3x3 (const point_t _points [3], const float _lScreen [3], float & _Ax, float & _Ay, float & _B) {
	const float K = (_points [2].y - _points [0].y)*(_points [1].x - _points [0].x) - (_points [1].y - _points [0].y)*(_points [2].x - _points [0].x);
	if (K == 0)
		// All 3 points belong to a line.
		return false;

	_Ay = ((_lScreen [2] - _lScreen [0])*(_points [1].x - _points [0].x) - (_lScreen [1] - _lScreen [0])*(_points [2].x - _points [0].x))/K;

	if (_points [1].x != _points [0].x)
		_Ax = (_lScreen [1] - _lScreen [0] - _Ay*(_points [1].y - _points [0].y))/(_points [1].x - _points [0].x);
	else if (_points [2].x != _points [0].x)
		_Ax = (_lScreen [2] - _lScreen [0] - _Ay*(_points [2].y - _points [0].y))/(_points [2].x - _points [0].x);
	else
		return false;

	_B = _lScreen [0] - _Ax*_points [0].x - _Ay*_points [0].y;

	return true;
}

	// Solve system of 4 linear equations:  B + Ax*x[i] + Ay*y[i] + D*x[i]*y[i] = screen[i]
	// to get values of B, Ax and Ay, D from known (x,y)->screen.
bool Solve4x4_XY (const point_t _points [4], const float _lScreen [4], float & _Ax, float & _Ay, float & _B, float & _D) {
	const float & x1 = _points [0].x, & y1 = _points [0].y;
	const float & x2 = _points [1].x, & y2 = _points [1].y;
	const float & x3 = _points [2].x, & y3 = _points [2].y;
	const float & x4 = _points [3].x, & y4 = _points [3].y;

	const float l1 = _lScreen [0];
	const float l2 = _lScreen [1];
	const float l3 = _lScreen [2];
	const float l4 = _lScreen [3];

	const float K3 = (x2 - x1)*(y3 - y1) - (x3 - x1)*(y2 - y1);
	const float K4 = (x2 - x1)*(y4 - y1) - (x4 - x1)*(y2 - y1);
	const float M3 = (x2 - x1)*(x3*y3 - x1*y1) - (x3 - x1)*(x2*y2 - x1*y1);
	const float M4 = (x2 - x1)*(x4*y4 - x1*y1) - (x4 - x1)*(x2*y2 - x1*y1);
	const float N3 = (x2 - x1)*(l3 - l1) - (x3 - x1)*(l2 - l1);
	const float N4 = (x2 - x1)*(l4 - l1) - (x4 - x1)*(l2 - l1);

	// D.
	if (K3*M4 == K4*M3)
		return false;
	_D = (K3*N4 - N3*K4)/(K3*M4 - K4*M3);

	// Ay.
	if (K3 != 0)
		_Ay = (N3 - _D*M3)/K3;
	else
		_Ay = (N4 - _D*M4)/K4;

	// Ax.
	if (x2 != x1)
		_Ax = ((l2 - l1) - _Ay*(y2 - y1) - _D*(x2*y2 - x1*y1))/(x2 - x1);
	else if (x3 != x1)
		_Ax = ((l3 - l1) - _Ay*(y3 - y1) - _D*(x3*y3 - x1*y1))/(x3 - x1);
	else if (x4 != x1)
		_Ax = ((l4 - l1) - _Ay*(y4 - y1) - _D*(x4*y4 - x1*y1))/(x4 - x1);
	else
		return false;

	// B.
	_B = l1 - _Ax*x1 - _Ay*y1 - _D*x1*y1;

	return true;
}
/*
	// Solve system of 4 linear equations:  B + Ax*x[i] + Ay*y[i] + D*x[i]*x[i] = screen[i]
	// to get values of B, Ax and Ay, D from known (x,y)->screen.
static
bool Solve4x4_XX (const point_t _points [4], const float _lScreen [4], float & _Ax, float & _Ay, float & _B, float & _D) {
	const float & x1 = _points [0].x, & y1 = _points [0].y;
	const float & x2 = _points [1].x, & y2 = _points [1].y;
	const float & x3 = _points [2].x, & y3 = _points [2].y;
	const float & x4 = _points [3].x, & y4 = _points [3].y;

	const float & l1 = _lScreen [0];
	const float & l2 = _lScreen [1];
	const float & l3 = _lScreen [2];
	const float & l4 = _lScreen [3];

	const float K3 = (x2 - x1)*(y3 - y1) - (x3 - x1)*(y2 - y1);
	const float K4 = (x2 - x1)*(y4 - y1) - (x4 - x1)*(y2 - y1);
	const float M3 = (x2 - x1)*(x3*x3 - x1*x1) - (x3 - x1)*(x2*x2 - x1*x1);
	const float M4 = (x2 - x1)*(x4*x4 - x1*x1) - (x4 - x1)*(x2*x2 - x1*x1);
	const float N3 = (x2 - x1)*(l3 - l1) - (x3 - x1)*(l2 - l1);
	const float N4 = (x2 - x1)*(l4 - l1) - (x4 - x1)*(l2 - l1);

	// D.
	if (K3*M4 == K4*M3)
		return false;
	_D = (K3*N4 - N3*K4)/(K3*M4 - K4*M3);

	// Ay.
	if (K3 != 0)
		_Ay = (N3 - _D*M3)/K3;
	else
		_Ay = (N4 - _D*M4)/K4;

	// Ax.
	if (x2 != x1)
		_Ax = ((l2 - l1) - _Ay*(y2 - y1) - _D*(x2*x2 - x1*x1))/(x2 - x1);
	else if (x3 != x1)
		_Ax = ((l3 - l1) - _Ay*(y3 - y1) - _D*(x3*x3 - x1*x1))/(x3 - x1);
	else
		_Ax = ((l4 - l1) - _Ay*(y4 - y1) - _D*(x4*x4 - x1*x1))/(x4 - x1);

	// B.
	_B = l1 - _Ax*x1 - _Ay*y1 - _D*x1*x1;

	return true;
}*/

