# include "stdafx.h"
# include "Resource.h"
# include "Menu.h"
# include "Settings.h"
# include "Toolbar.h"
# include "Map.h"
# include "Selection.h"
# include "Levels.h"

# include "UndoRedo.h"
# include "Log.h"
# include "RecentFiles.h"
# include "Load.h"
# include "Globals.h"

void UpdateMenu () {
	const HMENU hMenu = ::GetMenu (g_hWnd);
	const HMENU hViewMenu = ::GetSubMenu (hMenu, 2);

	bool bEmptyMap, bNoMap;
	size_t cLevels;
	{
		CCritSec::CLock lock (g_cs);
		bEmptyMap = g_map.rectBound.Invalid ();
		bNoMap = g_map.empty ();
		cLevels = g_map.Levels.size ();
	}

	// 'File' menu.
	::EnableMenuItem (hMenu, ID_FILE_NEW,       MF_BYCOMMAND | (bNoMap           || g_bLoading ? MF_GRAYED : MF_ENABLED));
	::EnableMenuItem (hMenu, ID_FILE_OPEN,      MF_BYCOMMAND | (                    g_bLoading ? MF_GRAYED : MF_ENABLED));
	::EnableMenuItem (hMenu, ID_FILE_ADD,       MF_BYCOMMAND | (                    g_bLoading ? MF_GRAYED : MF_ENABLED));
	::EnableMenuItem (hMenu, ID_FILE_SAVE,      MF_BYCOMMAND | (! CanSave ()     || g_bLoading ? MF_GRAYED : MF_ENABLED));
	::EnableMenuItem (hMenu, ID_FILE_SAVE_AS,   MF_BYCOMMAND | (bNoMap           || g_bLoading ? MF_GRAYED : MF_ENABLED));
	::EnableMenuItem (hMenu, ID_MAP_PROPERTIES, MF_BYCOMMAND | (bEmptyMap        || g_bLoading ? MF_GRAYED : MF_ENABLED));
	::EnableMenuItem (hMenu, ID_ERROR_LOG,      MF_BYCOMMAND | (IsLogEmpty ()    || g_bLoading ? MF_GRAYED : MF_ENABLED));
	// Recent files.
	for (size_t cID = 2000; cID <= 2000 + 9; ++ cID) {
		const bool bIsFileAccessible = ::IsFileAccessible (cID - 2000);
		::EnableMenuItem (hMenu, cID,           MF_BYCOMMAND | (! bIsFileAccessible || g_bLoading ? MF_GRAYED : MF_ENABLED));
	}

	// 'Edit' menu.
	if (! CanUndo () || g_bLoading)
		::ModifyMenu (hMenu, ID_EDIT_UNDO,      MF_BYCOMMAND | MF_STRING | MF_GRAYED,  ID_EDIT_UNDO, "Undo\tCtrl+Z");
	else {
		std::string strText;
		strText = "Undo ";
		g_Undo.back ()->GetName (strText);
		strText.append ("\tCtrl+Z");
		::ModifyMenu (hMenu, ID_EDIT_UNDO, MF_BYCOMMAND | MF_STRING | MF_ENABLED, ID_EDIT_UNDO, strText.c_str ());
	}

	if (! CanRedo () || g_bLoading)
		::ModifyMenu (hMenu, ID_EDIT_REDO, MF_BYCOMMAND | MF_STRING | MF_GRAYED,  ID_EDIT_REDO, "Redo\tCtrl+Y");
	else {
		std::string strText;
		strText = "Redo ";
		g_Redo.back ()->GetName (strText);
		strText.append ("\tCtrl+Z");
		::ModifyMenu (hMenu, ID_EDIT_REDO, MF_BYCOMMAND | MF_STRING | MF_ENABLED, ID_EDIT_REDO, strText.c_str ());
	}

	::EnableMenuItem (hMenu, ID_EDIT_CUT,     MF_BYCOMMAND | (true            || g_bLoading ? MF_GRAYED : MF_ENABLED));
	::EnableMenuItem (hMenu, ID_EDIT_COPY,    MF_BYCOMMAND | (true            || g_bLoading ? MF_GRAYED : MF_ENABLED));
	::EnableMenuItem (hMenu, ID_EDIT_PASTE,   MF_BYCOMMAND | (true            || g_bLoading ? MF_GRAYED : MF_ENABLED));
	::EnableMenuItem (hMenu, ID_EDIT_DELETE,  MF_BYCOMMAND | (g_Selection.empty () || g_bLoading ? MF_GRAYED : MF_ENABLED));

	::EnableMenuItem (hMenu, ID_MODE_DRAG,    MF_BYCOMMAND | (                   g_bLoading ? MF_GRAYED : MF_ENABLED));
	::EnableMenuItem (hMenu, ID_MODE_ZOOM,    MF_BYCOMMAND | (                   g_bLoading ? MF_GRAYED : MF_ENABLED));
	::EnableMenuItem (hMenu, ID_MODE_SELECT,  MF_BYCOMMAND | (                   g_bLoading ? MF_GRAYED : MF_ENABLED));
	::EnableMenuItem (hMenu, ID_MODE_MOVE_POINTS, MF_BYCOMMAND | (               g_bLoading ? MF_GRAYED : MF_ENABLED));
	::EnableMenuItem (hMenu, ID_MODE_AT,      MF_BYCOMMAND | (                   g_bLoading ? MF_GRAYED : MF_ENABLED));
	static const int iaModes [] = {ID_MODE_DRAG, ID_MODE_ZOOM, ID_MODE_SELECT, ID_MODE_MOVE_POINTS, ID_MODE_AT};
	int iMode = GetSetting ("SelectMode", (DWORD) 0);
	if (iMode >= sizeof (iaModes)/sizeof (iaModes [0]))
		iMode = 0;
	::CheckMenuRadioItem (hMenu, ID_MODE_DRAG, ID_MODE_AT, iaModes [iMode], MF_BYCOMMAND);

	::EnableMenuItem (hMenu, ID_UNSELECT,     MF_BYCOMMAND | (g_Selection.empty () || g_bLoading ? MF_GRAYED : MF_ENABLED));

	// 'View' menu.
	::EnableMenuItem (hMenu, ID_ZOOM_IN,      MF_BYCOMMAND | (bEmptyMap        || g_bLoading ? MF_GRAYED : MF_ENABLED));
	::EnableMenuItem (hMenu, ID_ZOOM_OUT,     MF_BYCOMMAND | (bEmptyMap        || g_bLoading ? MF_GRAYED : MF_ENABLED));
	::EnableMenuItem (hMenu, ID_ZOOM_FULL,    MF_BYCOMMAND | (bEmptyMap        || g_bLoading ? MF_GRAYED : MF_ENABLED));
	::EnableMenuItem (hMenu, ID_SHOW_MINI,    MF_BYCOMMAND | (bEmptyMap        || g_bLoading ? MF_GRAYED : MF_ENABLED));
	::EnableMenuItem (hMenu, ID_SHOW_RULER,   MF_BYCOMMAND | (bEmptyMap        || g_bLoading ? MF_GRAYED : MF_ENABLED));
	::EnableMenuItem (hMenu, ID_SHOW_GRID,    MF_BYCOMMAND | (bEmptyMap        || g_bLoading ? MF_GRAYED : MF_ENABLED));
	::EnableMenuItem (hMenu, ID_SHOW_LABELS,  MF_BYCOMMAND | (bEmptyMap        || g_bLoading ? MF_GRAYED : MF_ENABLED));
	::EnableMenuItem (hMenu, ID_TRANSPARENT_POLYGONS,
	                                          MF_BYCOMMAND | (bEmptyMap        || g_bLoading ? MF_GRAYED : MF_ENABLED));
	::EnableMenuItem (hMenu, ID_REFRESH,      MF_BYCOMMAND | (bEmptyMap        || g_bLoading ? MF_GRAYED : MF_ENABLED));

	// 'Scale' sub-menu.
	::EnableMenuItem (hViewMenu, 3, MF_BYPOSITION | (bEmptyMap || g_bLoading ? MF_GRAYED : MF_ENABLED));
	if (g_hScaleCB)
		::EnableWindow (g_hScaleCB, bEmptyMap || g_bLoading ? FALSE : TRUE);

	// 'Levels' sub-menu.
	::EnableMenuItem (hViewMenu, 6, MF_BYPOSITION | (bEmptyMap || g_bLoading ? MF_GRAYED : MF_ENABLED));
	UpdateLevelsMenu ();

	::CheckMenuItem (hMenu, ID_SHOW_MINI,   MF_BYCOMMAND | (GetSetting ("ShowMiniMap", true) ? MF_CHECKED : MF_UNCHECKED));
	::CheckMenuItem (hMenu, ID_SHOW_RULER,  MF_BYCOMMAND | (GetSetting ("ShowRuler",   true) ? MF_CHECKED : MF_UNCHECKED));
	::CheckMenuItem (hMenu, ID_SHOW_GRID,   MF_BYCOMMAND | (GetSetting ("ShowGrid",    true) ? MF_CHECKED : MF_UNCHECKED));
	::CheckMenuItem (hMenu, ID_SHOW_LABELS, MF_BYCOMMAND | (GetSetting ("ShowLabels",  true) ? MF_CHECKED : MF_UNCHECKED));
	::CheckMenuItem (hMenu, ID_TRANSPARENT_POLYGONS, MF_BYCOMMAND | (GetSetting ("TransparentPolygons", true) ? MF_CHECKED : MF_UNCHECKED));

	::DrawMenuBar (g_hWnd);

	// Update toolbar.
	const size_t cBtns = ::SendMessage (g_hToolbarWnd, TB_BUTTONCOUNT, 0, 0);
	for (size_t cBtn = 0; cBtn < cBtns; ++ cBtn) {
		TBBUTTON tbb;
		::SendMessage (g_hToolbarWnd, TB_GETBUTTON, cBtn, reinterpret_cast<LPARAM> (& tbb));
		if (tbb.fsStyle == TBSTYLE_BUTTON || tbb.fsStyle == TBSTYLE_GROUP)
			::SendMessage (g_hToolbarWnd, TB_PRESSBUTTON,  tbb.idCommand, MAKELONG (::GetMenuState (hMenu, tbb.idCommand, MF_BYCOMMAND) & MF_CHECKED ? TRUE : FALSE, 0));
		if (tbb.fsStyle == TBSTYLE_CHECK)
			::SendMessage (g_hToolbarWnd, TB_CHECKBUTTON,  tbb.idCommand, MAKELONG (::GetMenuState (hMenu, tbb.idCommand, MF_BYCOMMAND) & MF_CHECKED ? TRUE : FALSE, 0));
		if (tbb.fsStyle != TBSTYLE_SEP)
			::SendMessage (g_hToolbarWnd, TB_ENABLEBUTTON, tbb.idCommand, MAKELONG (::GetMenuState (hMenu, tbb.idCommand, MF_BYCOMMAND) & MF_GRAYED  ? FALSE : TRUE, 0));
	}
}
