# include "StdAfx.h"
# include "Resource.h"
# include "ModeSelect.h"
# include "Selection.h"
# include "MapView.h"
# include "RgnProperties.h"
# include "Rename.h"
# include "ChangeType.h"
# include "Vectorize.h"
# include "Globals.h"

static HMENU g_hmenuRgnPopup;
static HMENU g_hmenuImagePopup;
static HMENU g_hmenuAttachmentPopup;

bool CSelectMode::Init () {
	g_hmenuRgnPopup = ::GetSubMenu (::LoadMenu (g_hInst, MAKEINTRESOURCE (IDM_POPUP)), 0);
	assert (g_hmenuRgnPopup);
	::SetMenuDefaultItem (g_hmenuRgnPopup, ID_RGN_PROPERTIES, MF_BYCOMMAND);

	g_hmenuImagePopup = ::GetSubMenu (::LoadMenu (g_hInst, MAKEINTRESOURCE (IDM_POPUP)), 1);
	assert (g_hmenuImagePopup);
	::SetMenuDefaultItem (g_hmenuImagePopup, ID_RGN_PROPERTIES, MF_BYCOMMAND);

	g_hmenuAttachmentPopup = ::GetSubMenu (::LoadMenu (g_hInst, MAKEINTRESOURCE (IDM_POPUP)), 2);
	assert (g_hmenuAttachmentPopup);
	::SetMenuDefaultItem (g_hmenuAttachmentPopup, ID_RGN_PROPERTIES, MF_BYCOMMAND);

	return true;
}

void CSelectMode::OnEnterMode () {
	SetPermanentCursor (g_hArrowCursor);
}

void CSelectMode::OnLDown (int _x, int _y, DWORD _dwKeys) {
	SelectObjectAt (_x, _y, (_dwKeys & MK_SHIFT) != 0, (_dwKeys & MK_CONTROL) != 0);
}

void CSelectMode::OnLUp (int _x, int _y, DWORD _dwKeys) {
	SetPermanentCursor (g_hArrowCursor);
}

void CSelectMode::OnMove (int _x, int _y, DWORD _dwKeys) {
	if (s_bDND) {
		// TODO: move selected objects.
	}
}

void CSelectMode::OnRUp (int _x, int _y, DWORD _dwKeys) {
	g_lRClickX = _x;
	g_lRClickY = _y;

	SelectObjectAt (_x, _y, (_dwKeys & MK_SHIFT) != 0, (_dwKeys & MK_CONTROL) != 0);

	if (! g_Selection.empty ()) {
		HMENU hMenu;
		if (g_Selection [0].pImage) {
			hMenu = g_hmenuImagePopup;
		} else {
			if (g_Selection [0].iRgn->dwFlags & rgn_t::fWaypointOrTrack) {
				hMenu = g_hmenuAttachmentPopup;
				const UINT uiFlags = g_Selection.size () == 1 ? MF_BYCOMMAND | MF_ENABLED : MF_BYCOMMAND | MF_GRAYED;
				::EnableMenuItem (hMenu, ID_RGN_PROPERTIES, uiFlags);
			} else {
				hMenu = g_hmenuRgnPopup;
				const UINT uiFlags = g_Selection.size () == 1 ? MF_BYCOMMAND | MF_ENABLED : MF_BYCOMMAND | MF_GRAYED;
				::EnableMenuItem (hMenu, ID_EDIT_LABEL, uiFlags);
				::EnableMenuItem (hMenu, ID_EDIT_TYPE,  uiFlags);
				::EnableMenuItem (hMenu, ID_MODE_MOVE_POINTS, uiFlags);
				::EnableMenuItem (hMenu, ID_RGN_PROPERTIES, uiFlags);
			}
		}

		POINT pt;
		::GetCursorPos (& pt);
		::TrackPopupMenu (
			hMenu,
			TPM_LEFTALIGN | TPM_TOPALIGN | TPM_LEFTBUTTON,
			pt.x, pt.y,
			0, g_hWnd, NULL
		);
	}
}

bool CSelectMode::OnCommand (WORD _wID) {
	switch (_wID) {
		case ID_EDIT_LABEL:
			if (g_Selection.size () == 1 && g_Selection [0].pLayer)
				OnEditLabelStart ();
			return true;

		case ID_EDIT_TYPE:
			if (g_Selection.size () == 1 && g_Selection [0].pLayer)
				ShowChangeType (g_hWnd);
			return true;

		case ID_EDIT_VECTORIZE:
			if (g_Selection.size () == 1 && g_Selection [0].pImage)
				Vectorize (g_lRClickX, g_lRClickY, g_Selection [0].pImage);
			return true;
	}
	return false;
}

void CSelectMode::OnLDblClk (int _x, int _y, DWORD _dwKeys) {
	SelectObjectAt (_x, _y, (_dwKeys & MK_SHIFT) != 0, (_dwKeys & MK_CONTROL) != 0);
	if (! g_Selection.empty ())
		ShowRgnProperties (g_Selection [0]);
}

