# include "StdAfx.h"
# include "Resource.h"
# include "ModeZoom.h"
# include "PtAlgo.h"
# include "Map.h"
# include "Position.h"
# include "Scale.h"
# include "MapView.h"
# include "Globals.h"

static HCURSOR g_hcurZoom;
static HCURSOR g_hcurZoomIn;
static HCURSOR g_hcurZoomOut;
static HCURSOR g_hcurZoomDrag;

static HBRUSH g_hbrZoom;

bool CZoomMode::Init () {
	assert (g_hInst);

	g_hcurZoom     = ::LoadCursor (g_hInst, MAKEINTRESOURCE (IDC_CursorZoom));
	g_hcurZoomIn   = ::LoadCursor (g_hInst, MAKEINTRESOURCE (IDC_CursorZoomIn));
	g_hcurZoomOut  = ::LoadCursor (g_hInst, MAKEINTRESOURCE (IDC_CursorZoomOut));
	g_hcurZoomDrag = ::LoadCursor (g_hInst, MAKEINTRESOURCE (IDC_CursorZoomDrag));

	g_hbrZoom = ::CreateSolidBrush (c_colGray);

	return true;
}

void CZoomMode::OnEnterMode () {
	SetPermanentCursor (g_hcurZoom);
}

void CZoomMode::OnLDown (int _x, int _y, DWORD _dwKeys) {
	SetPermanentCursor (g_hcurZoomIn);

	StartDND (_x, _y);
}

void CZoomMode::OnLUp (int _x, int _y, DWORD _dwKeys) {
	SetPermanentCursor (g_hcurZoom);
	if (CMouseMode::s_bDND) {
		StopDND ();
		if (
			::labs (_x - CMouseMode::s_lX_DND_Start) < ::GetSystemMetrics (SM_CXDRAG) &&
			::labs (_y - CMouseMode::s_lY_DND_Start) < ::GetSystemMetrics (SM_CYDRAG)
		) {
			SetScale (g_fScale/c_Sqrt4, _x, _y);
		} else {
			point_t point;
			ScreenToPt ((CMouseMode::s_lX_DND_Start + _x)/2, (CMouseMode::s_lY_DND_Start + _y)/2, point);
			SetToCenter (point);

			const float fKx = ::fabs (_x - CMouseMode::s_lX_DND_Start)/g_lWidth;
			const float fKy = ::fabs (_y - CMouseMode::s_lY_DND_Start)/g_lHeight;

			SetScale (fKx > fKy ? g_fScale*fKx : g_fScale*fKy);
		}
	}
}

void CZoomMode::OnMove (int _x, int _y, DWORD _dwKeys) {
	if (
		CMouseMode::s_bDND &&
		(
			::labs (_x - CMouseMode::s_lX_DND_Start) >= ::GetSystemMetrics (SM_CXDRAG) ||
			::labs (_y - CMouseMode::s_lY_DND_Start) >= ::GetSystemMetrics (SM_CYDRAG)
		)
	) {
		SetPermanentCursor (g_hcurZoomDrag);
		Redraw ();
	} else
		SetPermanentCursor (g_hcurZoom);
}

void CZoomMode::OnRDown (int _x, int _y, DWORD _dwKeys) {
	CMouseMode::s_lX_DND_Start = _x;
	CMouseMode::s_lY_DND_Start = _y;
	SetPermanentCursor (g_hcurZoomOut);
}

void CZoomMode::OnRUp (int _x, int _y, DWORD _dwKeys) {
	SetPermanentCursor (g_hcurZoom);
	if (
		::labs (_x - CMouseMode::s_lX_DND_Start) >= ::GetSystemMetrics (SM_CXDRAG) ||
		::labs (_y - CMouseMode::s_lY_DND_Start) >= ::GetSystemMetrics (SM_CYDRAG)
	)
		return;

	SetScale (g_fScale*c_Sqrt4, _x, _y);
}

void CZoomMode::OnEsc () {
	StopDND ();
	Redraw ();
}

void CZoomMode::OnPaint (HDC _hDC) {
	if (! CMouseMode::s_bDND)
		return;

	RECT rect;

	rect.left = CMouseMode::s_lX_DND_Start;
	rect.right = CMouseMode::s_lX_DND;
	if (rect.left > rect.right)
		std::swap (rect.left, rect.right);
	++ rect.left;

	rect.top = CMouseMode::s_lY_DND_Start;
	rect.bottom = CMouseMode::s_lY_DND;
	if (rect.top > rect.bottom)
		std::swap (rect.top, rect.bottom);
	++ rect.top;

	::FrameRect (_hDC, & rect, g_hbrZoom);
}