# include "stdafx.h"
# include "MouseMode.h"
# include "Menu.h"
# include "Settings.h"
# include "MapView.h"
# include "Globals.h"

HCURSOR g_hCursor;
HCURSOR g_hPermanentCursor;

HCURSOR g_hWaitCursor;
HCURSOR g_hArrowCursor;

CWaiter::CWaiter () {
	::SetCursor (g_hCursor = g_hWaitCursor);
}

CWaiter::~CWaiter () {
	::SetCursor (g_hCursor = g_hPermanentCursor);
}

///////////////////////////////////////////////

# include "DragMap.h"
static CDragMode g_modeDrag;

# include "ModeZoom.h"
static CZoomMode g_modeZoom;

# include "ModeSelect.h"
static CSelectMode g_modeSelect;

# include "MovePoints.h"
static CPointsMode g_modeMovePoints;

# include "AffTransform.h"
static CAffTransformMode g_modeAT;

CMouseMode * const g_pMouseModes [] = {
	& g_modeDrag,
	& g_modeZoom,
	& g_modeSelect,
	& g_modeMovePoints,
	& g_modeAT
};

const size_t g_cMouseModes = sizeof (g_pMouseModes)/sizeof (g_pMouseModes [0]);

CMouseMode * CMouseMode::s_pMode;

bool CMouseMode::_Init () {
	for (size_t cMode = 0; cMode < g_cMouseModes; ++ cMode)
		if (! g_pMouseModes [cMode]->Init ())
			return false;

	CMouseMode::Set (GetSetting ("SelectMode", (DWORD) 0));

	return true;
}

void CMouseMode::_OnNewMap () {
	for (size_t cMode = 0; cMode < g_cMouseModes; ++ cMode)
		g_pMouseModes [cMode]->OnNewMap ();
}

void CMouseMode::Set (size_t _cModeIndex) {
	if (g_pMouseModes [_cModeIndex] == s_pMode) {
		// NOTE: clicking toolbar group button being already in pressed state will unpress it.
		// make sure that the button will be still pressed.
		UpdateMenu ();
		return;
	}

	if (s_pMode)
		s_pMode->OnLeaveMode ();

	if (_cModeIndex >= g_cMouseModes)
		s_pMode = & g_modeDrag;
	else
		s_pMode = g_pMouseModes [_cModeIndex];

	SaveSetting ("SelectMode", (DWORD) _cModeIndex);
	UpdateMenu ();

	s_pMode->OnEnterMode ();

	Redraw ();
}

bool CMouseMode::s_bDND = false;
long CMouseMode::s_lX_DND_Start;
long CMouseMode::s_lY_DND_Start;
long CMouseMode::s_lX_DND;
long CMouseMode::s_lY_DND;

void CMouseMode::SetPermanentCursor (HCURSOR _hcur) {
	g_hCursor = g_hPermanentCursor = _hcur;

	POINT p;
	::GetCursorPos (& p);
	if (::WindowFromPoint (p) == g_hMapWnd)
		::SetCursor (_hcur);
}

void CMouseMode::StartDND (long _lX, long _lY) {
	if (CMouseMode::s_bDND)
		return;

	CMouseMode::s_bDND = true;
	CMouseMode::s_lX_DND_Start = _lX;
	CMouseMode::s_lY_DND_Start = _lY;
	CMouseMode::s_lX_DND = _lX;
	CMouseMode::s_lY_DND = _lY;
	::SetCapture (g_hMapWnd);
}

void CMouseMode::StopDND () {
	if (! CMouseMode::s_bDND)
		return;

	CMouseMode::s_bDND = false;
	::ReleaseCapture ();
}

void CMouseMode::_OnMove (int _x, int _y, DWORD _dwKeys) {
	if (CMouseMode::s_bDND) {
		CMouseMode::s_lX_DND = _x;
		CMouseMode::s_lY_DND = _y;
	}

	if (s_pMode)
		s_pMode->OnMove (_x, _y, _dwKeys);
}