# include "stdafx.h"
# include "RecentFiles.h"
# include "Load.h"
# include "Settings.h"
# include "Globals.h"

const size_t c_cMaxRFs = 10;

struct file_name_t {
	char str [MAX_PATH];

	file_name_t (const char * _str) {
		::strncpy (str, _str, MAX_PATH);
	}
};
typedef std::vector<file_name_t> rfs_t;
static rfs_t g_rfs;

static bool g_bSeparatorAdded = false;

void UpdateRecentFilesMenu () {
	// Get 'File' menu.
	const HMENU hFileMenu = ::GetSubMenu (::GetMenu (g_hWnd), 0);

	// Remove old recent files.
	size_t c;
	for (c = 0; c < c_cMaxRFs; ++ c)
		::RemoveMenu (hFileMenu, 2000 + c, MF_BYCOMMAND);

	// Add menu separator (first time only).
	if (! g_bSeparatorAdded) {
		::AppendMenu (hFileMenu, MF_SEPARATOR, 0, NULL);
		g_bSeparatorAdded = true;
	}

	// Add recent files items.
	const UINT uFlags = MF_STRING | (g_bLoading ? MF_GRAYED : MF_ENABLED);
	for (c = 0; c < g_rfs.size () && c < c_cMaxRFs; ++ c) {
		char strText [MAX_PATH + 10];
		::sprintf (strText, "&%d %s", c, g_rfs [c].str);
		::AppendMenu (hFileMenu, uFlags, 2000 + c, strText);
	}
}

void InitRecentFiles () {
	for (size_t c = 0; c < c_cMaxRFs; ++ c) {
		char strName [32];
		::sprintf (strName, "Recent%d", c);

		char strRF [MAX_PATH];
		::GetSetting (strName, strRF, MAX_PATH, "");
		if (strRF [0] == '\0')
			break;

		g_rfs.push_back (strRF);
	}

	if (g_rfs.empty ())
		return;

	UpdateRecentFilesMenu ();
}

void AddRecentFile (const char * _strFileName) {
	// Remove existing position, if any.
	for (rfs_t::iterator i = g_rfs.begin (); i != g_rfs.end (); ++ i)
		if (::strncmp (_strFileName, i->str, MAX_PATH) == 0) {
			g_rfs.erase (i);
			break;
		}

	// Insert to top.
	g_rfs.insert (g_rfs.begin (), _strFileName);

	// Save to registry.
	for (size_t c = 0; c < c_cMaxRFs && c < g_rfs.size (); ++ c) {
		char strName [32];
		::sprintf (strName, "Recent%d", c);
		::SaveSetting (strName, g_rfs [c].str);
	}

	// Update menu.
	UpdateRecentFilesMenu ();
}

void LoadRecentFile (size_t _cIndex) {
	if (_cIndex >= g_rfs.size ())
		return;

	LoadMap (g_rfs [_cIndex].str, 0, true);
}

bool IsFileAccessible (size_t _cIndex) {
	if (_cIndex >= g_rfs.size ())
		return false;
	return ::GetFileAttributes (g_rfs [_cIndex].str) != -1;
}

