# include "stdafx.h"
# include "UndoRedo.h"
# include "Map.h"
# include "Rename.h"
# include "Label.h"
# include "MapView.h"
# include "Position.h"
# include "Selection.h"
# include "Types.h"
# include "Globals.h"

class CRenameAction : public CEditAction {
  public:
	CRenameAction (layer_t * _pLayer, rgns_t::iterator _iRgn, const char * _strNewLabel) :
		m_pLayer (_pLayer),
		m_pRgn (& * _iRgn)
	{
		m_strNewLabel = _strNewLabel;
		m_strPrevLabel = m_pRgn->strLabel;
	}

	virtual void GetName (std::string & _str) const;
	virtual bool CanUndo () const;
	virtual bool Do ();
	virtual void Undo ();
	virtual void Redo ();

  private:
	layer_t * m_pLayer;
	rgn_t *   m_pRgn;

	std::string m_strNewLabel;
	std::string m_strPrevLabel;
};

void CRenameAction::GetName (std::string & _str) const {
	CCritSec::CLock lock (g_cs);

	_str.append ("rename [");
	GetHintText (* m_pRgn, _str);
	_str.append ("]");
}

bool CRenameAction::CanUndo () const {return true;}

bool CRenameAction::Do () {
	CCritSec::CLock lock (g_cs);

	// Update label' text.
	assert (m_pLayer);
	m_pRgn->strLabel = m_strNewLabel;

	// Update label attributes.
	m_pRgn->pTypeDesc->pfPrepareLabel (* m_pRgn);

	return true;
}

void CRenameAction::Undo () {
	CCritSec::CLock lock (g_cs);

	// Revert label' text
	assert (m_pLayer);
	m_pRgn->strLabel = m_strPrevLabel;

	// Update label attributes.
	m_pRgn->pTypeDesc->pfPrepareLabel (* m_pRgn);
}

void CRenameAction::Redo () {
	CRenameAction::Do ();
}

/////////////////////////////////////////////////////////////////

static HWND g_hLabelEditWnd;
static WNDPROC g_wpLabelEdit = NULL;
LRESULT APIENTRY LabelEditProc (HWND _hWnd, UINT _uMsg, WPARAM _wParam, LPARAM _lParam);

bool InitLabelEdit () {
	assert (g_hMapWnd);

	g_hLabelEditWnd = ::CreateWindow (
		"EDIT", NULL,
		ES_LEFT | ES_AUTOHSCROLL | WS_CHILDWINDOW | WS_CLIPSIBLINGS | WS_CLIPCHILDREN | WS_BORDER,
		0, 0, 1, 1,
		g_hMapWnd, NULL, g_hInst, NULL
	);
	if (! g_hLabelEditWnd)
		return false;

	::SendMessage (g_hLabelEditWnd, WM_SETFONT, (WPARAM) g_hFont, FALSE);
	// Subclass the edit box to hook ENTER and ESC.
	g_wpLabelEdit = reinterpret_cast<WNDPROC> (::GetWindowLong (g_hLabelEditWnd, GWL_WNDPROC));
	::SetWindowLong (g_hLabelEditWnd, GWL_WNDPROC, reinterpret_cast<long> (LabelEditProc));

	return true;
}

void OnEditLabelStart () {
	if (g_Selection.size () != 1 || ! g_Selection [0].pLayer)
		return;

	if (! (g_Selection [0].iRgn->dwFlags & rgn_t::fPtLabelDefined))
		g_Selection [0].iRgn->pTypeDesc->pfPrepareLabel (* g_Selection [0].iRgn);

	if (g_Selection [0].iRgn->dwFlags & rgn_t::fPtLabelDefined) {
		g_bDialogMode = true;

		::SetWindowText (g_hLabelEditWnd, g_Selection [0].iRgn->strLabel.c_str ());

		const size_t cLen = g_Selection [0].iRgn->strLabel.size ();

		const int w = cLen*5 + 20 < 200 ? cLen*5 + 20 : 200;
		const int h = 14;
		int x = XToScreen (g_Selection [0].iRgn->ptLabel.x);
		if (x + w > g_lWidth)
			x = g_lWidth - w;
		if (x < 0)
			x = 0;
		int y = YToScreen (g_Selection [0].iRgn->ptLabel.y);
		if (y + h > g_lHeight)
			y = g_lHeight - h;
		if (y < 0)
			y = 0;
		::SetWindowPos (g_hLabelEditWnd, NULL, x, y, w, h, SWP_NOZORDER | SWP_SHOWWINDOW);

		::SetFocus (g_hLabelEditWnd);
		::SendMessage (g_hLabelEditWnd, EM_SETSEL, 0, cLen);
		::SendMessage (g_hLabelEditWnd, EM_SCROLLCARET, 0, 0);
	}
}

static
void OnEditLabelComplete (bool _bDone = true) {
	if (! g_bDialogMode)
		return;
	g_bDialogMode = false;

	// Return keyboard input from the label edit box.
	::SetFocus (g_hWnd);

	::ShowWindow (g_hLabelEditWnd, SW_HIDE);

	if (g_Selection [0].pLayer && _bDone) {
		const size_t cLen = ::GetWindowTextLength (g_hLabelEditWnd) + 1;
		char * const strLabel = reinterpret_cast<char *> (::alloca (sizeof (char)*cLen));
		::GetWindowText (g_hLabelEditWnd, strLabel, cLen);

		if (g_Selection [0].iRgn->strLabel == strLabel)
			// NOTE: no changes.
			return;

		if (! IsLabelCorrect (strLabel)) {
			ErrorMsgBox ("The label includes symbols not allowed for 6-bit coding.");
			return;
		}

		PerformEditAction (new CRenameAction (g_Selection [0].pLayer, g_Selection [0].iRgn, strLabel));
	}
}

static
LRESULT APIENTRY LabelEditProc (HWND _hWnd, UINT _uMsg, WPARAM _wParam, LPARAM _lParam) {
	switch (_uMsg) {
		case WM_KILLFOCUS:
			OnEditLabelComplete ();
			break;

		case WM_CHAR:
			switch (_wParam) {
				case VK_RETURN: OnEditLabelComplete (); break;
				case VK_ESCAPE: OnEditLabelComplete (false); break;
			}
			break;
	}

	return ::CallWindowProc (g_wpLabelEdit, _hWnd, _uMsg, _wParam, _lParam);
}
