# include "stdafx.h"
# include "Resource.h"
# include "Map.h"
# include "Types.h"
# include "RgnProperties.h"
# include "Globals.h"

int CALLBACK RgnPropertiesProc (HWND _hWnd, UINT _uMsg, WPARAM _wParam, LPARAM);

void ShowRgnProperties (const find_info_t & _Selection) {
	if (! _Selection.pLayer)
		return;

	::DialogBoxParam (g_hInst, MAKEINTRESOURCE (IDD_RGN_PROPERTIES), g_hWnd, & RgnPropertiesProc, reinterpret_cast<LPARAM> (& * _Selection.iRgn));
}

static
int CALLBACK RgnPropertiesProc (HWND _hWnd, UINT _uMsg, WPARAM _wParam, LPARAM _lParam) {
	switch (_uMsg) {
		case WM_INITDIALOG: {
			const rgn_t * pRgn = reinterpret_cast<const rgn_t *> (_lParam);

			if (pRgn->dwFlags & rgn_t::fWaypointOrTrack)
				::SetWindowText (_hWnd, pRgn->ushKind == 0 ? "Waypoint properties" : "Track properties");

			size_t cPoints = 0;
			size_t cElements = 0;
			for (size_t c = 0; c < pRgn->elements.size (); ++ c) {
				const rgn_element_t & element = pRgn->elements [c];
				if (element.bDeleted)
					continue;

				cPoints += element.points.size ();
				++ cElements;
			}

			const size_t cObjectSize = pRgn->cEndPos - pRgn->cBeginPos;

			char strBuf [64];

			assert (pRgn->ushKind < 4);
			::sprintf (strBuf, "%s (0x%x, %s)", pRgn->pTypeDesc->strTypeName, pRgn->ushType, g_strKinds [pRgn->ushKind]);
			::SetDlgItemText (_hWnd, IDC_TYPE, strBuf);
			::sprintf (strBuf, "%s", pRgn->strLabel.c_str ());
			::SetDlgItemText (_hWnd, IDC_LABEL, strBuf);
			::sprintf (strBuf, "%d", pRgn->cOrdinal);
			::SetDlgItemText (_hWnd, IDC_ORDINAL, strBuf);

			::sprintf (strBuf, "%d", cElements);
			::SetDlgItemText (_hWnd, IDC_ELEMENTS, strBuf);
			::sprintf (strBuf, "%d", cPoints);
			::SetDlgItemText (_hWnd, IDC_POINTS, strBuf);

			::sprintf (strBuf, "%d", pRgn->cFileLine);
			::SetDlgItemText (_hWnd, IDC_LINE, strBuf);
			::sprintf (strBuf, "%d", pRgn->cBeginPos);
			::SetDlgItemText (_hWnd, IDC_OFFSET, strBuf);
			::sprintf (strBuf, "%d", cObjectSize);
			::SetDlgItemText (_hWnd, IDC_OSIZE, strBuf);
			
			//
			// Prepare source edit box.
			//
			if (pRgn->pFileRef) {
				::SetDlgItemText (_hWnd, IDC_FILE_NAME, pRgn->pFileRef->strFileName.c_str ());

				const CHandle hFile = ::CreateFile (pRgn->pFileRef->strFileName.c_str (), GENERIC_READ, 0, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
				if (hFile == NULL || hFile == INVALID_HANDLE_VALUE) {
					::SetDlgItemText (_hWnd, IDC_SOURCE, "(Can't open the file)");
					return TRUE;
				}

				char * const pSource = reinterpret_cast<char *> (alloca (sizeof (char)*cObjectSize + 1));

				::SetFilePointer (hFile, pRgn->cBeginPos, 0, FILE_BEGIN);
				DWORD dwFileSizeRead = 0;
				::ReadFile (hFile, pSource, cObjectSize, & dwFileSizeRead, NULL);
				if (dwFileSizeRead < cObjectSize) {
					::SetDlgItemText (_hWnd, IDC_SOURCE, "(Can't read the file)");
					return TRUE;
				}
				pSource [cObjectSize] = 0;

				::SetDlgItemText (_hWnd, IDC_SOURCE, pSource);
			} else {
				::SetDlgItemText (_hWnd, IDC_FILE_NAME, "(no associated file)");
				::SetDlgItemText (_hWnd, IDC_SOURCE, "(No associated file)");
			}

			return TRUE;
		}

		case WM_COMMAND:
			if (LOWORD (_wParam) == IDOK || LOWORD (_wParam) == IDCANCEL) {
				// On closing dialog.
				::EndDialog (_hWnd, 0);
				return TRUE;
			}
	}

	return FALSE;
}
