# include "stdafx.h"
# include "Ruler.h"
# include "Map.h"
# include "Load.h"
# include "Scale.h"
# include "Globals.h"

static
void DrawText2 (HDC _hDC, int _x, int _y, const char * _strText) {
	const size_t cLen = ::strlen (_strText);

	::SetTextColor (_hDC, c_colWhite);
	::SetBkMode (_hDC, TRANSPARENT);

	POLYTEXT pt [4];
	pt [0].x = _x - 1;
	pt [0].y = _y - 1;
	pt [1].x = _x + 1;
	pt [1].y = _y - 1;
	pt [2].x = _x - 1;
	pt [2].y = _y + 1;
	pt [3].x = _x + 1;
	pt [3].y = _y + 1;
	for (size_t c = 0; c < 4; ++ c) {
		pt [c].n = cLen;
		pt [c].lpstr = _strText;
		pt [c].uiFlags = 0;
		pt [c].pdx = NULL;
	}
	::PolyTextOut (_hDC, pt, 4);

	::SetTextColor (_hDC, c_colBlack);
	pt [0].x = _x;
	pt [0].y = _y;
	::PolyTextOut (_hDC, pt, 1);
}

void ShowRuler (HDC _hDC, int _x0, int _y0) {
	if (g_bLoading)
		return;
	{
		CCritSec::CLock lock (g_cs);

		if (g_map.rectBound.Invalid ())
			return;
	}

	char strText [20];
	const float fRulerLen = GetScale (g_fScale, strText);

	// Draw the ruler background.
	::SelectObject (_hDC, g_hWhiteThickPen);
	size_t c;
	for (c = 0; c < 4; ++ c) {
		::MoveToEx (_hDC, _x0 + fast_floor (.5 + fRulerLen*c),       _y0 - 5, NULL);
		::LineTo   (_hDC, _x0 + fast_floor (.5 + fRulerLen*c),       _y0);
		::LineTo   (_hDC, _x0 + fast_floor (.5 + fRulerLen*(c + 1)), _y0);
		::LineTo   (_hDC, _x0 + fast_floor (.5 + fRulerLen*(c + 1)), _y0 - 5);
	}

	// Draw the ruler foreground.
	::SelectObject (_hDC, g_hBlackThinPen);
	for (c = 0; c < 4; ++ c) {
		::MoveToEx (_hDC, _x0 + fast_floor (.5 + fRulerLen*c),       _y0 - 5, NULL);
		::LineTo   (_hDC, _x0 + fast_floor (.5 + fRulerLen*c),       _y0);
		::LineTo   (_hDC, _x0 + fast_floor (.5 + fRulerLen*(c + 1)), _y0);
		::LineTo   (_hDC, _x0 + fast_floor (.5 + fRulerLen*(c + 1)), _y0 - 5);
	}

	// Draw the ruler text.
	::SelectObject (_hDC, g_hFont);
	::SetBkMode (_hDC, TRANSPARENT);
	::SetTextAlign (_hDC, TA_BOTTOM | TA_CENTER);
	DrawText2 (_hDC, _x0 + fast_floor (fRulerLen/2), _y0 + 2, strText);
}

