# include "stdafx.h"
# include "UndoRedo.h"
# include "Map.h"
# include "Save.h"
# include "Log.h"
# include "Menu.h"
# include "Status.h"
# include "RecentFiles.h"
# include "Globals.h"

bool RequestSave () {
	if (g_bModified) {
		switch (::MessageBox (g_hWnd, "The map is modified.\nDo you want to save it?", "Save", MB_YESNOCANCEL | MB_ICONQUESTION)) {
			case IDYES:
				if (g_bTypeCanBeSaved)
					SaveMap (g_strFileName);
				else
					SaveMapAs ();
				return true;
			case IDNO: return true;
			case IDCANCEL: return false;
		}
	}
	return true;
}

void SaveMapAs () {
	char strFileName [MAX_PATH + 1];
	::strncpy (strFileName, g_strFileName, MAX_PATH);

	OPENFILENAME ofn;
	::ZeroMemory (& ofn, sizeof (OPENFILENAME));
	ofn.lStructSize = sizeof (OPENFILENAME);
	ofn.hwndOwner = g_hWnd;
	ofn.hInstance = g_hInst;
	ofn.lpstrFile = strFileName;
	ofn.nMaxFile = sizeof (strFileName);
	ofn.lpstrFilter = "Polish format (*.txt)\0*.txt\0All files (*.*)\0*.*\0";
	ofn.nFilterIndex = 1;
	ofn.Flags = OFN_EXPLORER | OFN_PATHMUSTEXIST | OFN_ENABLESIZING | OFN_NOREADONLYRETURN | OFN_OVERWRITEPROMPT;
	ofn.lpstrDefExt = "txt";
	if (::GetSaveFileName (& ofn) == TRUE)
		SaveMap (strFileName);
}

void SaveMap (const char * _strFileName) {
	CWaiter w;

	const long lStartTime = ::GetTickCount ();

	//
	// Open the file.
	//

	const CHandle hFile = ::CreateFile (_strFileName, GENERIC_WRITE, 0, NULL, CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL | FILE_FLAG_SEQUENTIAL_SCAN, NULL);
	if (hFile == NULL || hFile == INVALID_HANDLE_VALUE) {
		ErrorMsgBox ("Can't open file\n%s.", _strFileName);
		return;
	}

	AddRecentFile (_strFileName);

	g_bTypeCanBeSaved = true;

	//
	// Save the map.
	//

	ClearLog ();
	ReportText ("Saving map to file '%s'...", _strFileName);

	SetStatus ("Saving map...");

	CTxtSaver saver;
	{
		CCritSec::CLock lock (g_cs);
		saver.Save (hFile, _strFileName, & g_map);
	}

	//
	// Report about results.
	//

	if (GetWarningsCount ())
		ReportText ("%d warning(s).", GetWarningsCount ());

	char strTime [64];
	::sprintf (strTime, "Save time %.2f s", .001*(::GetTickCount () - lStartTime));
	ReportText (strTime);

	//
	// Update GUI elements.
	//

	ResetUndoRedo ();
	g_bModified = false;

	::strncpy (g_strFileName, _strFileName, MAX_PATH);
	UpdateTitle ();

	UpdateMenu ();

	if (GetWarningsCount ())
		SetStatus ("Ready*");
	else
		SetStatus ("Ready");
}