# include "stdafx.h"
# include "Scale.h"
# include "Position.h"
# include "Map.h"
# include "Status.h"
# include "MapView.h"
# include "Globals.h"

	// The Earth radius in meters.
const float c_ER = 6378137;

	// Meters per pixel.
float g_fScale = c_Scale_20km;

	// Pixels per degree.
float g_kX, g_kY;

float GetKx (float _Scale) {return c_ER*(c_PI/180.)/_Scale*g_map.fCosY;}
float GetKy (float _Scale) {return c_ER*(c_PI/180.)/_Scale;}

void RefreshScales () {
	CCritSec::CLock lock (g_cs);

	g_kX = GetKx (g_fScale);
	g_kY = GetKy (g_fScale);

	g_lX = g_X*g_kX;
	g_lY = g_Y*g_kY;

	g_lTotalWidth  = g_map.rectBound.Width  () <= 0 ? g_lWidth  : g_kX*g_map.rectBound.Width ();
	g_lTotalHeight = g_map.rectBound.Height () <= 0 ? g_lHeight : g_kY*g_map.rectBound.Height ();
}

float RoundScale (float _f) {
	assert (_f > 0);
	const float l = ::log10 (_f*c_pixels_per_cm);
	assert (l >= 0);
	const float m = ::pow (10, ::fmod (l, 1));
	const float o = ::pow (10, fast_floor (l));
	assert (1 <= m && m < 10);

	const float m2 = fast_floor (m*10 + .5)*.1;

	return m2*o/c_pixels_per_cm;
}

void SetScale (float _fNewScale) {
	SetScale (_fNewScale, g_lWidth/2, g_lHeight/2);
}

void SetScale (float _fNewScale, long _lX0, long _lY0) {
	if (_fNewScale > c_ScaleMax)
		_fNewScale = c_ScaleMax;
	else if (_fNewScale < c_ScaleMin)
		_fNewScale = c_ScaleMin;

	_fNewScale = RoundScale (_fNewScale);

	if (_fNewScale == g_fScale)
		return;

	SetScaleStatus (_fNewScale);

	const float fNewKx = GetKx (_fNewScale);
	const float fNewKy = GetKy (_fNewScale);

	g_X += _lX0/g_kX - _lX0/fNewKx;
	_lY0 = g_lHeight - _lY0;
	g_Y += _lY0/g_kY - _lY0/fNewKy;

	g_fScale = _fNewScale;
	RefreshScales ();
	Refresh ();
}

float GetFullMapScale () {
	CCritSec::CLock lock (g_cs);

	const float Scale1 = c_ER*(c_PI/180.)*g_map.rectBound.Width  ()/g_lWidth*g_map.fCosY;
	const float Scale2 = c_ER*(c_PI/180.)*g_map.rectBound.Height ()/g_lHeight;
	return Scale1 > Scale2 ? Scale1 : Scale2;
}

void SetFullMapScale () {
	g_fScale = GetFullMapScale ();
	SetScaleStatus (g_fScale);
	RefreshScales ();

	{
		CCritSec::CLock lock (g_cs);
		g_X = g_map.rectBound.Width ()/2  - g_lWidth/(2*g_kX);
		g_Y = g_map.rectBound.Height ()/2 - g_lHeight/(2*g_kY);
	}
	RefreshScales ();

	Refresh ();
}

float GetScale (float _fScale, char * _strText) {
	// Calculate the scale's order & mantissa (in meters).
	assert (_fScale > 0);
	const float k_log = ::log10 (_fScale*c_pixels_per_cm);
	assert (k_log >= 0);
	const float k_mant  = ::pow (10, ::fmod (k_log, 1));
	const float k_order = ::pow (10, fast_floor (k_log));
	assert (1 <= k_mant && k_mant < 10);

	// Get the exact ruler length.
	const float fRulerLen = fast_floor (k_mant*k_order/_fScale + .5);

	const bool bInKm = k_order >= 1000;
	const float fValue = bInKm ? k_mant*k_order/1000 : k_mant*k_order;

	// Get the ruler text.
	if (fValue >= 0.1) {
		const int iValue_int = fast_floor (fValue + .001);
		const int iValue_rat = fast_floor (::fmod (fValue + 0.01, 1)*10 + .5);

		if (iValue_rat == 0)
			::sprintf (_strText, "%d %s", iValue_int, bInKm ? "km" : "m");
		else
			::sprintf (_strText, "%d.%d %s", iValue_int, iValue_rat, bInKm ? "km" : "m");
	} else
		::sprintf (_strText, "%.1f %s", fValue, bInKm ? "km" : "m");

	return fRulerLen;
}