# include "stdafx.h"
# include "resource.h"
# include "Toolbar.h"
# include "Status.h"
# include "Scale.h"
# include "Globals.h"

HWND g_hToolbarWnd;
HWND g_hToolbarTipWnd;

HWND g_hScaleCB;
HWND g_hScaleEB;

static WNDPROC g_wpScaleEdit = NULL;
LRESULT APIENTRY ScaleEditProc (HWND _hWnd, UINT _uMsg, WPARAM _wParam, LPARAM _lParam);

bool InitToolbar () {
	assert (g_hWnd);

	TBBUTTON btns [] = {
		{0,  ID_FILE_NEW,             TBSTATE_ENABLED, TBSTYLE_BUTTON, 0, 0},
		{1,  ID_FILE_ADD,             TBSTATE_ENABLED, TBSTYLE_BUTTON, 0, 0},
		{2,  ID_FILE_SAVE,            TBSTATE_ENABLED, TBSTYLE_BUTTON, 0, 0},

		{0,  -1,                      TBSTATE_ENABLED, TBSTYLE_SEP,    0, 0},
		{3,  ID_PRINT,                TBSTATE_ENABLED, TBSTYLE_BUTTON, 0, 0},
		{4,  ID_MAP_PROPERTIES,       TBSTATE_ENABLED, TBSTYLE_BUTTON, 0, 0},

		{0,  -1,                      TBSTATE_ENABLED, TBSTYLE_SEP,    0, 0},
		{5,  ID_MODE_DRAG,            TBSTATE_ENABLED, TBSTYLE_GROUP,  0, 0},
		{6,  ID_MODE_ZOOM,            TBSTATE_ENABLED, TBSTYLE_GROUP,  0, 0},
		{7,  ID_MODE_SELECT,          TBSTATE_ENABLED, TBSTYLE_GROUP,  0, 0},
		{8,  ID_MODE_MOVE_POINTS,     TBSTATE_ENABLED, TBSTYLE_GROUP,  0, 0},
		{9,  ID_MODE_AT,              TBSTATE_ENABLED, TBSTYLE_GROUP,  0, 0},

		{0,  -1,                      TBSTATE_ENABLED, TBSTYLE_SEP,    0, 0},
		// NOTE: this is just the placeholder for the 'scales' combobox
		// which will be placed above.
		{0,  ID_SCALES_CB,            TBSTATE_ENABLED, TBSTYLE_SEP,    0, 0},
		{10, ID_ZOOM_IN,              TBSTATE_ENABLED, TBSTYLE_BUTTON, 0, 0},
		{11, ID_ZOOM_OUT,             TBSTATE_ENABLED, TBSTYLE_BUTTON, 0, 0},

		{0,  -1,                      TBSTATE_ENABLED, TBSTYLE_SEP,    0, 0},
		{12, ID_SHOW_LABELS,          TBSTATE_ENABLED, TBSTYLE_CHECK,  0, 0},
		{13, ID_SHOW_RULER,           TBSTATE_ENABLED, TBSTYLE_CHECK,  0, 0},
		{14, ID_SHOW_GRID,            TBSTATE_ENABLED, TBSTYLE_CHECK,  0, 0},
		{15, ID_TRANSPARENT_POLYGONS, TBSTATE_ENABLED, TBSTYLE_CHECK,  0, 0},

		{0,  -1,                      TBSTATE_ENABLED, TBSTYLE_SEP,    0, 0},
		{16, ID_EDIT_UNDO,            TBSTATE_ENABLED, TBSTYLE_DROPDOWN, 0, 0},
		{17, ID_EDIT_REDO,            TBSTATE_ENABLED, TBSTYLE_DROPDOWN, 0, 0},

		{0,  -1,                      TBSTATE_ENABLED, TBSTYLE_SEP,    0, 0},
		{18, ID_EDIT_CUT,             TBSTATE_ENABLED, TBSTYLE_BUTTON, 0, 0},
		{19, ID_EDIT_COPY,            TBSTATE_ENABLED, TBSTYLE_BUTTON, 0, 0},
		{20, ID_EDIT_PASTE,           TBSTATE_ENABLED, TBSTYLE_BUTTON, 0, 0},
		{21, ID_EDIT_DELETE,          TBSTATE_ENABLED, TBSTYLE_BUTTON, 0, 0},
	};
	g_hToolbarWnd = ::CreateToolbarEx (
		g_hWnd, WS_CHILD | WS_VISIBLE | WS_CLIPSIBLINGS | CCS_TOP | TBSTYLE_FLAT | TBSTYLE_TOOLTIPS, 100,
		17, g_hInst, IDT_TOOLBAR,
		btns, sizeof (btns)/sizeof (btns [0]), 0, 0, 16, 16, sizeof (TBBUTTON)
	);
	// Enable downarrows for Undo/Redo.
	::SendMessage (g_hToolbarWnd, TB_SETEXTENDEDSTYLE, 0, TBSTYLE_EX_DRAWDDARROWS);

	g_hToolbarTipWnd = reinterpret_cast<HWND> (::SendMessage (g_hToolbarWnd, TB_GETTOOLTIPS, 0, 0));

	//
	// Add 'scale' combobox.
	//

	const size_t cBtnIndex = ::SendMessage (g_hToolbarWnd, TB_COMMANDTOINDEX, ID_SCALES_CB, 0);
	if (! g_bNT) {
		// NOTE: combobox' behaviour appears unpredictable under Win9x (up to system crash).
		::SendMessage (g_hToolbarWnd, TB_DELETEBUTTON, cBtnIndex, 0);
		return true;
	}

	RECT rectCB;
	::SendMessage (g_hToolbarWnd, TB_GETITEMRECT, cBtnIndex, reinterpret_cast<LPARAM> (& rectCB));

	g_hScaleCB = ::CreateWindow (
		WC_COMBOBOXEX, NULL, WS_CHILD | WS_VISIBLE | WS_VSCROLL | WS_CLIPSIBLINGS | CBS_DROPDOWN,
		rectCB.left, 1, 60, 200,
		g_hToolbarWnd, NULL, g_hInst, NULL
	);
	::SendMessage (g_hScaleCB, CBEM_SETEXTENDEDSTYLE, CBES_EX_NOEDITIMAGE, CBES_EX_NOEDITIMAGE);
	::SetWindowLong (g_hScaleCB, GWL_ID, ID_SCALES_CB);
	::SendMessage (g_hScaleCB, WM_SETFONT, (WPARAM) g_hFont, FALSE);

	// Change width of separator on the toolbar being below the scales combobox.
	::GetWindowRect (g_hScaleCB, & rectCB);
	TBBUTTONINFO bi;
	bi.cbSize = sizeof (bi);
    bi.dwMask = TBIF_SIZE;
	bi.cx = rectCB.right - rectCB.left;
	::SendMessage (g_hToolbarWnd, TB_SETBUTTONINFO, ID_SCALES_CB, reinterpret_cast<LPARAM> (& bi));

	HWND hScaleCB2 = (HWND) ::SendMessage (g_hScaleCB, CBEM_GETCOMBOCONTROL, 0, 0);
	::SetWindowLong (hScaleCB2, GWL_ID, ID_SCALES_CB);

	// Get the edit box.
	g_hScaleEB = (HWND) ::SendMessage (g_hScaleCB, CBEM_GETEDITCONTROL, 0, 0);
	::SetWindowLong (g_hScaleEB, GWL_ID, ID_SCALES_CB);
	::SendMessage (g_hScaleEB, EM_SETMARGINS, EC_LEFTMARGIN, 3);

	// Fill combobox' list with scales menu items.
	const HMENU hScalesMenu = ::GetSubMenu (::GetSubMenu (::GetMenu (g_hWnd), 2), 3);
	assert (hScalesMenu);
	for (size_t cMenuItem = 0; cMenuItem < (size_t)::GetMenuItemCount (hScalesMenu); ++ cMenuItem) {
		char strMenuItemText [64];
		if (::GetMenuString (hScalesMenu, cMenuItem, strMenuItemText, sizeof (strMenuItemText), MF_BYPOSITION)) {
			COMBOBOXEXITEM cbi;
			cbi.mask = CBEIF_TEXT | CBEIF_LPARAM;
			cbi.iItem = -1;
			cbi.pszText = strMenuItemText;
			cbi.cchTextMax = sizeof (strMenuItemText);
			cbi.lParam = ::GetMenuItemID (hScalesMenu, cMenuItem);
			::SendMessage (g_hScaleCB, CBEM_INSERTITEM, 0, (LPARAM) & cbi);
		}
	}

	// Subclass the edit box to hook ENTER and ESC.
	g_wpScaleEdit = reinterpret_cast<WNDPROC> (::GetWindowLong (g_hScaleEB, GWL_WNDPROC));
	::SetWindowLong (g_hScaleEB, GWL_WNDPROC, reinterpret_cast<long> (ScaleEditProc));

	return true;
}

static
void OnEditScaleComplete (bool _bDone = true) {
	if (! g_bDialogMode)
		return;
	g_bDialogMode = false;

	// Return keyboard input from the Scale combobox.
	::SetFocus (g_hWnd);

	if (_bDone) {
		char strScale [20];
		::GetWindowText (g_hScaleEB, strScale, sizeof (strScale));

		char * strEnd = NULL;
		float fScale = ::strtod (strScale, & strEnd);
		if (fScale != 0) {
			if (strEnd != NULL && * strEnd != '\0') {
				// Omit spaces.
				while (* strEnd == ' ')
					++ strEnd;

				// Check unit.
				if (::strnicmp (strEnd, "km", 2) == 0) {
					SetScale (1000*fScale/c_pixels_per_cm);
					return;
				} else if (::strnicmp (strEnd, "m", 2) == 0) {
					SetScale (fScale/c_pixels_per_cm);
					return;
				}
			}
		}
	}

	// Restore scale text.
	SetScaleStatus (g_fScale);
}

static
LRESULT APIENTRY ScaleEditProc (HWND _hWnd, UINT _uMsg, WPARAM _wParam, LPARAM _lParam) {
	switch (_uMsg) {
		case WM_SETFOCUS:
			g_bDialogMode = true;
			break;

		case WM_KILLFOCUS:
			OnEditScaleComplete ();
			break;

		case WM_CHAR:
			switch (_wParam) {
				case VK_RETURN: OnEditScaleComplete (); break;
				case VK_ESCAPE: OnEditScaleComplete (false); break;
			}
			break;
	}
		
	return ::CallWindowProc (g_wpScaleEdit, _hWnd, _uMsg, _wParam, _lParam);
}

