# include "stdafx.h"
# include "Resource.h"
# include "Types.h"
# include "Map.h"
# include "Levels.h"
# include "MapView.h"
# include "Position.h"
# include "Globals.h"

const char * g_strKinds [] = {"POI", "point", "polyline", "polygon"};

HFONT g_hCityFont;

HBRUSH g_hWaterBrush;
HPEN g_hWaterPen;
HBRUSH g_hDryWaterBrush;
HPEN g_hDryWaterPen;
HPEN g_hCoastPen;
HBRUSH g_hGlacierBrush;

HBRUSH g_hMarchesBrush;
HBRUSH g_hForestBrush;

HBRUSH g_hCemeteryBrush;

HBRUSH g_hSettlementBrush;
HBRUSH g_hHousingBrush;
HBRUSH g_hCityParkBrush;

HPEN g_hGroundRoadPen;
HPEN g_hThickRoadPen;
HPEN g_hThinRoadPen;
HPEN g_hEnhancedRoadPen;
HPEN g_hHighwayPen1;
HPEN g_hHighwayPen2;
HPEN g_hRailroadPen1;
HPEN g_hRailroadPen2;

HPEN g_hBoundaryPen;
HPEN g_hHeightPen;
HPEN g_hDepthPen;

HCURSOR g_hcurAutoRental;
HCURSOR g_hcurAutoWash;
HCURSOR g_hcurBar;
HCURSOR g_hcurDrinkingWater;
HCURSOR g_hcurExit;
HCURSOR g_hcurInformation;
HCURSOR g_hcurLibrary;
HCURSOR g_hcurPolice;
HCURSOR g_hcurRestroom;
HCURSOR g_hcurRock;
HCURSOR g_hcurSpring;
HCURSOR g_hcurAutoRepair;
HCURSOR g_hcurBank;
HCURSOR g_hcurBeacon;
HCURSOR g_hcurBusStation;
HCURSOR g_hcurCinema;
HCURSOR g_hcurDanger;
HCURSOR g_hcurDining;
HCURSOR g_hcurHotel;
HCURSOR g_hcurMuseum;
HCURSOR g_hcurParking;
HCURSOR g_hcurPetrol;
HCURSOR g_hcurPharmacie;
HCURSOR g_hcurPostOffice;
HCURSOR g_hcurShopping;
HCURSOR g_hcurSports;
HCURSOR g_hcurSwimming;
HCURSOR g_hcurTelephone;
HCURSOR g_hcurTheatre;
HCURSOR g_hcurSettlement;
HCURSOR g_hcurCitySmall;
HCURSOR g_hcurCityMedium;
HCURSOR g_hcurCityBig;
HCURSOR g_hcurTop;
HCURSOR g_hcurAirport;
HCURSOR g_hcurWarf;
HCURSOR g_hcurBuilding;
HCURSOR g_hcurChurch;
HCURSOR g_hcurPassing;
HCURSOR g_hcurTree;
HCURSOR g_hcurStation;
HCURSOR g_hcurBridge;
HCURSOR g_hcurCemetery;
HCURSOR g_hcurHill;
HCURSOR g_hcurZoo;
HCURSOR g_hcurSchool;
HCURSOR g_hcurPicnic;
HCURSOR g_hcurHospital;
HCURSOR g_hcurCamping;
HCURSOR g_hcurCliff;
HCURSOR g_hcurFalls;
HCURSOR g_hcurIsland;
HCURSOR g_hcurReservoir;
HCURSOR g_hcurNamedpoint;
HCURSOR g_hcurMilitary;
HCURSOR g_hcurTower;
HCURSOR g_hcurFish;
HCURSOR g_hcurReef;

HPEN g_hTrackPen;
HCURSOR g_hcurWaypoint;
int g_nCursorShiftX_Waypoint, g_nCursorShiftY_Waypoint;

HPEN g_hSelectionPen;
HBRUSH g_hSelectionBrush;

HPEN    g_hUnknownPen;
HBRUSH  g_hUnknownBrush;
HCURSOR g_hcurUnknown;

const COLORREF c_colWater      = c_colBlue;
const COLORREF c_colWater2     = c_colDkBlue;
const COLORREF c_colGlacier    = c_colCyan;
const COLORREF c_colForest     = RGB ((153+213)/2, (226+240)/2, (153+153)/2);
const COLORREF c_colCityPark   = RGB (144, 192, 0);
const COLORREF c_colSettlement = c_colGray;
const COLORREF c_colHousing    = c_colRed;
const COLORREF c_colHeight     = RGB (137, 80, 24);
const COLORREF c_colDepth      = c_colCyan;
const COLORREF c_colBoundary   = RGB (0, 200, 100);

HFONT g_hRotatedFont [181];

void InitTypes () {
	for (int nAngle = -90; nAngle <= 90; ++ nAngle) {
		const HFONT hFont = ::CreateFont (12, 0, nAngle*10, nAngle*10, FW_DONTCARE, FALSE, FALSE, FALSE, DEFAULT_CHARSET, OUT_DEFAULT_PRECIS, CLIP_DEFAULT_PRECIS, DEFAULT_QUALITY, DEFAULT_PITCH | FF_DONTCARE, NULL);
		g_hRotatedFont [nAngle + 90] = hFont;
	}

	g_hCityFont = ::CreateFont (14, 0, 0, 0, FW_BOLD, FALSE, FALSE, FALSE, DEFAULT_CHARSET, OUT_DEFAULT_PRECIS, CLIP_DEFAULT_PRECIS, DEFAULT_QUALITY, DEFAULT_PITCH | FF_SWISS, NULL);

	g_hUnknownPen      = ::CreatePen (PS_DOT, 1, c_colRed);
	g_hUnknownBrush    = ::CreateHatchBrush (HS_BDIAGONAL, c_colRed);
	g_hcurUnknown      = ::LoadCursor (g_hInst, MAKEINTRESOURCE (IDC_Unknown));

	const HBITMAP hSelectionBmp = ::LoadBitmap (g_hInst, MAKEINTRESOURCE (IDB_SelectPattern));
	g_hSelectionPen    = ::CreatePen (PS_DOT, 1, RGB (0, 0, 0));
	g_hSelectionBrush  = ::CreatePatternBrush (hSelectionBmp);

	g_hWaterBrush      = ::CreateSolidBrush (c_colWater);
	g_hWaterPen        = ::CreatePen (PS_SOLID, 1, c_colWater);
	g_hDryWaterBrush   = ::CreateHatchBrush (HS_DIAGCROSS, c_colWater);
	g_hDryWaterPen     = ::CreatePen (PS_DASH,  1, c_colWater);
	g_hCoastPen        = ::CreatePen (PS_SOLID, 1, c_colWater2);
	g_hGlacierBrush    = ::CreateSolidBrush (c_colGlacier);

	g_hHeightPen       =  ::CreatePen (PS_SOLID, 1, c_colHeight);
	g_hDepthPen        =  ::CreatePen (PS_SOLID, 1, c_colDepth);

	g_hMarchesBrush    = ::CreateHatchBrush (HS_BDIAGONAL, c_colWater);
	g_hForestBrush     = ::CreateSolidBrush (c_colForest);

	const HBITMAP hCemeteryBmp = ::LoadBitmap (g_hInst, MAKEINTRESOURCE (IDB_Cemetery));
	g_hCemeteryBrush   = ::CreatePatternBrush (hCemeteryBmp);

	g_hSettlementBrush = ::CreateSolidBrush (c_colSettlement);
	g_hHousingBrush    = ::CreateSolidBrush (c_colRed);
	g_hCityParkBrush   = ::CreateSolidBrush (c_colCityPark);

	g_hTrackPen        = ::CreatePen (PS_SOLID, 3, c_colGray);
	g_hGroundRoadPen   = ::CreatePen (PS_DASH,  1, c_colBlack);
	g_hThinRoadPen     = ::CreatePen (PS_SOLID, 1, c_colBlack);
	g_hThickRoadPen    = ::CreatePen (PS_SOLID, 2, c_colBlack);
	g_hEnhancedRoadPen = ::CreatePen (PS_SOLID, 3, RGB (100, 50, 50));
	g_hHighwayPen1     = ::CreatePen (PS_SOLID, 4, RGB (100, 50, 50));
	g_hHighwayPen2     = ::CreatePen (PS_SOLID, 2, RGB (196, 100, 66));
	
	g_hRailroadPen1    = ::CreatePen (PS_SOLID, 3, c_colBlack);
	g_hRailroadPen2    = ::CreatePen (PS_DOT,   1, c_colWhite);

	g_hBoundaryPen     = ::CreatePen (PS_DOT,   1, c_colBoundary);

	g_hcurWaypoint     = ::LoadCursor (g_hInst, MAKEINTRESOURCE (IDC_Waypoint));
	g_hcurAutoRental   = ::LoadCursor (g_hInst, MAKEINTRESOURCE (IDC_AutoRental));
	g_hcurAutoWash     = ::LoadCursor (g_hInst, MAKEINTRESOURCE (IDC_AutoWash));
	g_hcurBar          = ::LoadCursor (g_hInst, MAKEINTRESOURCE (IDC_Bar));
	g_hcurDrinkingWater = ::LoadCursor (g_hInst, MAKEINTRESOURCE (IDC_DrinkingWater));
	g_hcurExit         = ::LoadCursor (g_hInst, MAKEINTRESOURCE (IDC_Exit));
	g_hcurInformation  = ::LoadCursor (g_hInst, MAKEINTRESOURCE (IDC_Information));
	g_hcurLibrary      = ::LoadCursor (g_hInst, MAKEINTRESOURCE (IDC_Library));
	g_hcurPolice       = ::LoadCursor (g_hInst, MAKEINTRESOURCE (IDC_Police));
	g_hcurRestroom     = ::LoadCursor (g_hInst, MAKEINTRESOURCE (IDC_Restroom));
	g_hcurRock         = ::LoadCursor (g_hInst, MAKEINTRESOURCE (IDC_Rock));
	g_hcurSpring       = ::LoadCursor (g_hInst, MAKEINTRESOURCE (IDC_Spring));
	g_hcurAutoRepair   = ::LoadCursor (g_hInst, MAKEINTRESOURCE (IDC_AutoRepair));
	g_hcurBank         = ::LoadCursor (g_hInst, MAKEINTRESOURCE (IDC_Bank));
	g_hcurBeacon       = ::LoadCursor (g_hInst, MAKEINTRESOURCE (IDC_Beacon));
	g_hcurBusStation   = ::LoadCursor (g_hInst, MAKEINTRESOURCE (IDC_BusStation));
	g_hcurCinema       = ::LoadCursor (g_hInst, MAKEINTRESOURCE (IDC_Cinema));
	g_hcurDanger       = ::LoadCursor (g_hInst, MAKEINTRESOURCE (IDC_Danger));
	g_hcurDining       = ::LoadCursor (g_hInst, MAKEINTRESOURCE (IDC_Dining));
	g_hcurHotel        = ::LoadCursor (g_hInst, MAKEINTRESOURCE (IDC_Hotel));
	g_hcurMuseum       = ::LoadCursor (g_hInst, MAKEINTRESOURCE (IDC_Museum));
	g_hcurParking      = ::LoadCursor (g_hInst, MAKEINTRESOURCE (IDC_Parking));
	g_hcurPetrol       = ::LoadCursor (g_hInst, MAKEINTRESOURCE (IDC_Petrol));
	g_hcurPharmacie    = ::LoadCursor (g_hInst, MAKEINTRESOURCE (IDC_Pharmacie));
	g_hcurPostOffice   = ::LoadCursor (g_hInst, MAKEINTRESOURCE (IDC_PostOffice));
	g_hcurShopping     = ::LoadCursor (g_hInst, MAKEINTRESOURCE (IDC_Shopping));
	g_hcurSports       = ::LoadCursor (g_hInst, MAKEINTRESOURCE (IDC_Sports));
	g_hcurSwimming     = ::LoadCursor (g_hInst, MAKEINTRESOURCE (IDC_Swimming));
	g_hcurTelephone    = ::LoadCursor (g_hInst, MAKEINTRESOURCE (IDC_Telephone));
	g_hcurTheatre      = ::LoadCursor (g_hInst, MAKEINTRESOURCE (IDC_Theatre));
	g_hcurSettlement   = ::LoadCursor (g_hInst, MAKEINTRESOURCE (IDC_Settlement));
	g_hcurCitySmall    = ::LoadCursor (g_hInst, MAKEINTRESOURCE (IDC_CitySmall));
	g_hcurCityMedium   = ::LoadCursor (g_hInst, MAKEINTRESOURCE (IDC_CityMedium));
	g_hcurCityBig      = ::LoadCursor (g_hInst, MAKEINTRESOURCE (IDC_CityBig));
	g_hcurTop          = ::LoadCursor (g_hInst, MAKEINTRESOURCE (IDC_Topc));
	g_hcurAirport      = ::LoadCursor (g_hInst, MAKEINTRESOURCE (IDC_Airport));
	g_hcurWarf         = ::LoadCursor (g_hInst, MAKEINTRESOURCE (IDC_Warf));
	g_hcurBuilding     = ::LoadCursor (g_hInst, MAKEINTRESOURCE (IDC_Building));
	g_hcurChurch       = ::LoadCursor (g_hInst, MAKEINTRESOURCE (IDC_Church));
	g_hcurPassing      = ::LoadCursor (g_hInst, MAKEINTRESOURCE (IDC_Passing));
	g_hcurTree         = ::LoadCursor (g_hInst, MAKEINTRESOURCE (IDC_Tree));
	g_hcurStation      = ::LoadCursor (g_hInst, MAKEINTRESOURCE (IDC_Station));
	g_hcurBridge       = ::LoadCursor (g_hInst, MAKEINTRESOURCE (IDC_Bridge));
	g_hcurCemetery     = ::LoadCursor (g_hInst, MAKEINTRESOURCE (IDC_Cemetery));
	g_hcurHill         = ::LoadCursor (g_hInst, MAKEINTRESOURCE (IDC_Hill));
	g_hcurZoo          = ::LoadCursor (g_hInst, MAKEINTRESOURCE (IDC_Zoo));
	g_hcurSchool       = ::LoadCursor (g_hInst, MAKEINTRESOURCE (IDC_School));
	g_hcurPicnic       = ::LoadCursor (g_hInst, MAKEINTRESOURCE (IDC_Picnic));
	g_hcurHospital     = ::LoadCursor (g_hInst, MAKEINTRESOURCE (IDC_Hospital));
	g_hcurCamping      = ::LoadCursor (g_hInst, MAKEINTRESOURCE (IDC_Camping));
	g_hcurCliff        = ::LoadCursor (g_hInst, MAKEINTRESOURCE (IDC_Cliff));
	g_hcurFalls        = ::LoadCursor (g_hInst, MAKEINTRESOURCE (IDC_Falls));
	g_hcurIsland       = ::LoadCursor (g_hInst, MAKEINTRESOURCE (IDC_Island));
	g_hcurReservoir    = ::LoadCursor (g_hInst, MAKEINTRESOURCE (IDC_Reservoir));
	g_hcurNamedpoint   = ::LoadCursor (g_hInst, MAKEINTRESOURCE (IDC_Namedpoint));
	g_hcurMilitary     = ::LoadCursor (g_hInst, MAKEINTRESOURCE (IDC_Military));
	g_hcurTower        = ::LoadCursor (g_hInst, MAKEINTRESOURCE (IDC_Tower));
	g_hcurFish         = ::LoadCursor (g_hInst, MAKEINTRESOURCE (IDC_Fish));
	g_hcurReef         = ::LoadCursor (g_hInst, MAKEINTRESOURCE (IDC_Reef));

	for (size_t cType = 0; cType < g_cTypes; ++ cType) {
		// Check metainfo.
		if (cType < g_cTypes - 1) {
			assert (g_types [cType].ulType < g_types [cType + 1].ulType);
			assert (g_types [cType].phPen != NULL);
		}

		type_desc_t & td = g_types [cType];

		// Prepare hot spots for cursors.
		ICONINFO ii;
		::GetIconInfo (* td.phCursor, & ii);
		const_cast<int &> (td.nCursorShiftX) = ii.xHotspot;
		const_cast<int &> (td.nCursorShiftY) = ii.yHotspot;
		::DeleteObject (ii.hbmColor);
		::DeleteObject (ii.hbmMask);

		// Create brushes for transparent polygons.
		if (td.ushKind == 3 || cType == 0) {
			assert (td.phBrush);
			LOGBRUSH lb;
			::GetObject (* td.phBrush, sizeof (lb), & lb);

			td.hbrTransparentPolygon = ::CreateHatchBrush (HS_DIAGCROSS, lb.lbColor);
		} else
			td.hbrTransparentPolygon = NULL;
	}

	ICONINFO ii;
	::GetIconInfo (g_hcurWaypoint, & ii);
	g_nCursorShiftX_Waypoint = ii.xHotspot;
	g_nCursorShiftY_Waypoint = ii.yHotspot;
	::DeleteObject (ii.hbmColor);
	::DeleteObject (ii.hbmMask);
}

///////////////////////////////////////////////

static
size_t GetTextWidth (const rgn_t & _rgn) {
	if (_rgn.strLabel.empty ())
		return 0;

	const char * str = _rgn.strLabel.c_str ();
	if (::strncmp (_rgn.strLabel.c_str (), "~[0x2", 5) == 0)
		str = _rgn.strLabel.c_str () + 5;

	if (_rgn.fSize2 != 0) {
		const CClientDC hDC (g_hMapWnd);
		::SelectObject (hDC, _rgn.pTypeDesc->phFont);
		RECT rect;
		::memset (& rect, 0, sizeof (RECT));
		::DrawText (hDC, str, -1, & rect, DT_CALCRECT | DT_NOPREFIX | DT_SINGLELINE);

		return rect.right - rect.left;
	} else
		return ::strlen (str)*6;
}

///////////////////////////////////////////////

void PreparePointLabel (rgn_t & _rgn) {
	if (! _rgn.origins.empty ()) {
		_rgn.ptLabel = _rgn.origins [0].point;
		_rgn.dwFlags |= rgn_t::fPtLabelDefined;
		return;
	}

	float ave_x = 0, ave_y = 0;
	size_t ave_N = 0;
	for (size_t cElement = 0; cElement < _rgn.elements.size (); ++ cElement) {
		const rgn_element_t & element = _rgn.elements [cElement];
		if (element.bDeleted)
			continue;

		const points_t & points = element.points;
		const size_t cPoints = points.size ();

		for (size_t cPoint = 0; cPoint < cPoints; ++ cPoint) {
			const point_t & pt = points [cPoint];
			ave_x += pt.x;
			ave_y += pt.y;
			++ ave_N;
		}
	}
	if (ave_N) {
		_rgn.ptLabel.x = ave_x/ave_N;
		_rgn.ptLabel.y = ave_y/ave_N;
		_rgn.fLabelAngle = 0;
		_rgn.dwFlags |= rgn_t::fPtLabelDefined;
	}

	_rgn.cLabelWidth = GetTextWidth (_rgn);
}

void PreparePolylineLabel (rgn_t & _rgn) {
	if (_rgn.elements.empty ())
		return;

	// Get middle element.
	const rgn_element_t & element = _rgn.elements [_rgn.elements.size ()/2];
	if (element.bDeleted)
		return;

	const points_t & points = element.points;
	const size_t cPoints = points.size ();
	if (cPoints == 0)
		return;

	const float fCosY = ::cos (points [0].y*c_PI/180.);

	// Estimate its length.
	float part_l = 0;
	size_t cPoint;
	for (cPoint = 0; cPoint < cPoints - 1; ++ cPoint) {
		const point_t & p0 = points [cPoint];
		const point_t & p1 = points [cPoint + 1];

		part_l += ::sqrt ((p1.x - p0.x)*(p1.x - p0.x)*fCosY*fCosY + (p1.y - p0.y)*(p1.y - p0.y));
	}

	// Find the middle segment.
	float part_l2 = 0;
	for (cPoint = 0; cPoint < cPoints - 1; ++ cPoint) {
		const point_t & p0 = points [cPoint];
		const point_t & p1 = points [cPoint + 1];

		const float l = ::sqrt ((p1.x - p0.x)*(p1.x - p0.x)*fCosY*fCosY + (p1.y - p0.y)*(p1.y - p0.y));
		if (part_l2 + l > part_l/2) {
			_rgn.ptLabel.x = (p0.x + p1.x)/2;
			_rgn.ptLabel.y = (p0.y + p1.y)/2;
			_rgn.fLabelAngle = ::atan2 (p1.y - p0.y, (p1.x - p0.x)*fCosY);
			_rgn.dwFlags |= rgn_t::fPtLabelDefined;
			break;
		} else
			part_l2 += l;
	}

	_rgn.cLabelWidth = GetTextWidth (_rgn);
}

void PreparePolygonLabel (rgn_t & _rgn) {
	if (_rgn.elements.empty ())
		return;

	// Get some element of the region.
	const rgn_element_t & element = _rgn.elements [0];
	if (element.bDeleted)
		return;

	const points_t & points = element.points;
	const size_t cPoints = points.size ();
	if (cPoints == 0)
		return;

	const float fCosY = ::cos (points [0].y*c_PI/180.);

	// Find first most distant point (relative to 0-th one).
	float fMaxD2 = 0;
	size_t cMostDistantPoint = 0;
	size_t cPoint;
	const point_t & pt0 = points [0];
	for (cPoint = 1; cPoint < cPoints; ++ cPoint) {
		const point_t & pt = points [cPoint];

		float d2 = (pt.x - pt0.x)*(pt.x - pt0.x)*fCosY*fCosY + (pt.y - pt0.y)*(pt.y - pt0.y);
		if (d2 > fMaxD2) {
			fMaxD2 = d2;
			cMostDistantPoint = cPoint;
		}
	}

	// Find second most distant point (relative to first found one).
	fMaxD2 = 0;
	size_t cMostDistantPoint2 = 0;
	const point_t & pt1 = points [cMostDistantPoint];
	for (cPoint = 1; cPoint < cPoints; ++ cPoint) {
		const point_t & pt = points [cPoint];

		const float d2 = (pt.x - pt1.x)*(pt.x - pt1.x)*fCosY*fCosY + (pt.y - pt1.y)*(pt.y - pt1.y);
		if (d2 > fMaxD2) {
			fMaxD2 = d2;
			cMostDistantPoint = cPoint;
		}
	}

	// Get middle segment of the range.
	cPoint = (cMostDistantPoint + cMostDistantPoint2)/2;
	const point_t & p0 = points [cPoint];
	const point_t & p1 = points [cPoint + 1];
	_rgn.ptLabel.x = (p0.x + p1.x)/2;
	_rgn.ptLabel.y = (p0.y + p1.y)/2;
	_rgn.fLabelAngle = ::atan2 (p1.y - p0.y, (p1.x - p0.x)*fCosY);
	_rgn.dwFlags |= rgn_t::fPtLabelDefined;

	_rgn.cLabelWidth = GetTextWidth (_rgn);
}

/////////////////////////////////////////

void DrawLabelImpl (HDC _hDC, const rgn_t & _rgn, const char * _strLabel) {
	assert (! _rgn.strLabel.empty ());

	bool bXYDefined = false;
	int x, y;

	if (_rgn.dwFlags & rgn_t::fPtLabelDefined) {
		x = XToScreen (_rgn.ptLabel.x);
		y = YToScreen (_rgn.ptLabel.y);
		bXYDefined = true;
	} else {
		for (size_t cOrigin = 0; cOrigin < _rgn.origins.size (); ++ cOrigin) {
			const rgn_origin_t & origin = _rgn.origins [cOrigin];
			if (origin.bDeleted)
				continue;

			if (! (_rgn.dwFlags & rgn_t::fWaypointOrTrack) && ! LevelIsVisible (origin.cLevel))
				continue;

			x = XToScreen (origin.point.x);
			y = YToScreen (origin.point.y);
			bXYDefined = true;

			break;
		}
	}

	if (bXYDefined) {
		const type_desc_t * const pTD = _rgn.pTypeDesc;

		::SetTextColor (_hDC, pTD->colText);
		::SetTextAlign (_hDC, TA_TOP | TA_LEFT);
		::SetBkMode (_hDC, TRANSPARENT); 
		::SelectObject (_hDC, pTD->phFont ? * (pTD->phFont) : g_hFont);
		::TextOut (_hDC, x + 2, y + 2, _strLabel, ::strlen (_strLabel));
	}
}

void DrawLabel (HDC _hDC, const rgn_t & _rgn) {
	DrawLabelImpl (_hDC, _rgn, _rgn.strLabel.c_str ());
}

void DrawMetersLabel (HDC _hDC, const rgn_t & _rgn) {
	assert (! _rgn.strLabel.empty ());

	float fFeets = ::atof (_rgn.strLabel.c_str ());
	if (fFeets == 0)
		DrawLabelImpl (_hDC, _rgn, _rgn.strLabel.c_str ());
	else {
		bool bMinus = false; 
		if (fFeets < 0) {
			fFeets = -fFeets;
			bMinus = true;
		}

		// Convert feets to meters.
		char strMeters [20];
		::sprintf (strMeters, bMinus ? "-%d" : "%d", static_cast<int> (fast_floor (fFeets/c_feet_per_m + .5)));
		DrawLabelImpl (_hDC, _rgn, strMeters);
	}
}

void DrawRotatedLabel (HDC _hDC, const rgn_t & _rgn, const char * _strLabel) {
	assert (! _rgn.strLabel.empty ());
	if (! (_rgn.dwFlags & rgn_t::fPtLabelDefined))
		return;

	float fLabelAngle = _rgn.fLabelAngle;
	if (fLabelAngle >= c_PI/2)
		fLabelAngle -= c_PI;
	if (fLabelAngle < -c_PI/2)
		fLabelAngle += c_PI;
	assert (-c_PI/2 <= fLabelAngle && fLabelAngle < c_PI/2);

	const int x = XToScreen (_rgn.ptLabel.x) - 3*::sin (fLabelAngle);
	const int y = YToScreen (_rgn.ptLabel.y) - 3*::cos (fLabelAngle);

	const int nAngle = fast_floor (fLabelAngle/c_PI*180 + .5);
	assert (-90 <= nAngle && nAngle <= 90);
	const HFONT hFont = g_hRotatedFont [nAngle + 90];

	// Draw the label.
	const type_desc_t * const pTD = _rgn.pTypeDesc;
	::SetTextColor (_hDC, pTD->colText);
	::SetTextAlign (_hDC, TA_BASELINE | TA_CENTER);
	::SetBkMode (_hDC, TRANSPARENT); 
	const HFONT hfontPrev = (HFONT) ::SelectObject (_hDC, hFont);
	::TextOut (_hDC, x + 2, y + 2, _strLabel, _rgn.strLabel.size ());

	::SelectObject (_hDC, hfontPrev);
}

/////////////////////////////////////////

void PrepareLakeLabel (rgn_t & _rgn) {
	PreparePointLabel (_rgn);
}

void DrawLakeLabel (HDC _hDC, const rgn_t & _rgn) {
	assert (! _rgn.strLabel.empty ());

	const int x = XToScreen (_rgn.ptLabel.x);
	const int y = YToScreen (_rgn.ptLabel.y);

	const type_desc_t * const pTD = _rgn.pTypeDesc;

	::SetTextColor (_hDC, pTD->colText);
	::SetTextAlign (_hDC, TA_BASELINE | TA_CENTER);
	::SetBkMode (_hDC, TRANSPARENT); 
	::SelectObject (_hDC, pTD->phFont ? * (pTD->phFont) : g_hFont);
	::TextOut (_hDC, x + 2, y + 2, _rgn.strLabel.c_str (), _rgn.strLabel.size ());
}

/////////////////////////////////////////

void PreparePolyRiverLabel (rgn_t & _rgn) {
	PreparePolygonLabel (_rgn);
}

void DrawPolyRiverLabel (HDC _hDC, const rgn_t & _rgn) {
	DrawRotatedLabel (_hDC, _rgn, _rgn.strLabel.c_str ());
}

/////////////////////////////////////////

void PrepareRiverLabel (rgn_t & _rgn) {
	PreparePolylineLabel (_rgn);
}

void DrawRiverLabel (HDC _hDC, const rgn_t & _rgn) {
	DrawRotatedLabel (_hDC, _rgn, _rgn.strLabel.c_str ());
}

/////////////////////////////////////////

void PrepareBoundaryLabel (rgn_t & _rgn) {
	PreparePolylineLabel (_rgn);
}

void DrawBoundaryLabel (HDC _hDC, const rgn_t & _rgn) {
	DrawRotatedLabel (_hDC, _rgn, _rgn.strLabel.c_str ());
}

/////////////////////////////////////////

void PrepareHeightLabel (rgn_t & _rgn) {
	PreparePolylineLabel (_rgn);
}

void DrawHeightLabel (HDC _hDC, const rgn_t & _rgn) {
	assert (! _rgn.strLabel.empty ());

	float fFeets = ::atof (_rgn.strLabel.c_str ());
	if (fFeets == 0)
		DrawRotatedLabel (_hDC, _rgn, _rgn.strLabel.c_str ());
	else {
		bool bMinus = false; 
		if (fFeets < 0) {
			fFeets = -fFeets;
			bMinus = true;
		}

		// Convert feets to meters.
		char strMeters [20];
		::sprintf (strMeters, bMinus ? "-%d" : "%d", static_cast<int> (fast_floor (fFeets/c_feet_per_m + .5)));
		DrawRotatedLabel (_hDC, _rgn, strMeters);
	}
}

////////////////////////////////////////

void PrepareRoadLabel (rgn_t & _rgn) {
	PreparePolylineLabel (_rgn);
}

void DrawRoadLabel (HDC _hDC, const rgn_t & _rgn) {
	assert (! _rgn.strLabel.empty ());
	if (! (_rgn.dwFlags & rgn_t::fPtLabelDefined))
		return;

	bool bDrawInRect = false;
	const char * strLabel = _rgn.strLabel.c_str ();
	if (strLabel [0] == '~' && strLabel [1] == '[' && strLabel [2] == '0' && strLabel [3] == 'x') {
		const char * const p = ::strchr (strLabel + 4, ']');
		if (p && p [1])
			strLabel = p + 1;

		char * pEnd;
		const int iCode = ::strtoul (_rgn.strLabel.c_str () + 4, & pEnd, 16);
		if (0x2a <= iCode && iCode <= 0x2f)
			bDrawInRect = true;
	}

	if (! bDrawInRect) {
		DrawRotatedLabel (_hDC, _rgn, strLabel);
		return;
	}
	
	// TODO: 
	//    ~[0x2a] - Interstate highway
	//    ~[0x2b] - Higway - shield
	//    ~[0x2c] - Higway - round symbol
	//    ~[0x2d] - Higway - small symbol - blue/red
	//    ~[0x2e] - Higway - big symbol
	//    ~[0x2f] - Higway - small symbol - white

	::SetBkMode (_hDC, TRANSPARENT);
	::SetTextColor (_hDC, c_colBlack);
	::SetTextAlign (_hDC, TA_TOP | TA_LEFT);
	::SelectObject (_hDC, g_hFont);

	RECT rect;
	memset (& rect, 0, sizeof (RECT));
	::DrawText (_hDC, strLabel, -1, & rect, DT_CALCRECT | DT_NOPREFIX | DT_SINGLELINE);

	const int w = rect.right - rect.left + 4;
	const int h = rect.bottom - rect.top;

	const int x = XToScreen (_rgn.ptLabel.x);
	const int y = YToScreen (_rgn.ptLabel.y);

	rect.left = x - w/2;
	rect.right = x + w/2;
	rect.top = y - h/2;
	rect.bottom = y + h/2;

	::FillRect (_hDC, & rect, g_hWhiteBrush);
	::FrameRect (_hDC, & rect, g_hBlackBrush);
	::DrawText (_hDC, strLabel, -1, & rect, DT_CENTER | DT_VCENTER | DT_NOPREFIX | DT_SINGLELINE);
}

	// NOTE: should be 'const' but can't be due to nCursorShiftX & nCursorShiftY members
type_desc_t g_types [] = {
//  |---------|------------------------|---------------------|-------------------------------------|---------------------|------------------|----------------------------------|---------------|--------------
//  |Type     |Text                    |Brush                |Pen, Pen2                            |Font, Font color     |Cursor            |Draw label                        |Scale          |Label scale
//  |---------|------------------------|---------------------|-------------------------------------|---------------------|------------------|----------------------------------|---------------|--------------
	{0x00000, "Unknown",                & g_hUnknownBrush,    & g_hUnknownPen, NULL,                NULL, c_colBlack,    & g_hcurUnknown,    PreparePointLabel, DrawLabel},

//	{0x00001, "City ???",               NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurSettlement, PreparePointLabel, DrawLabel},

	// POIs
	{0x00020, "Exit",                   NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurExit,       PreparePointLabel, DrawLabel},
	{0x00022, "Exit (restroom)",        NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurExit,       PreparePointLabel, DrawLabel},
	{0x00023, "Exit (convenience store)", NULL,               & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurExit,       PreparePointLabel, DrawLabel},
	{0x00024, "Exit (weigh station)",   NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurExit,       PreparePointLabel, DrawLabel},
	{0x00025, "Exit (toll booth)",      NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurExit,       PreparePointLabel, DrawLabel},
	{0x00026, "Exit (information)",     NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurExit,       PreparePointLabel, DrawLabel},
//	{0x00027, "Exit",                   NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurExit,       PreparePointLabel, DrawLabel},
	{0x00027, "Stop/station",           NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurStation,    PreparePointLabel, DrawLabel},
	{0x0002a, "Dining (other)",         NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurUnknown,    PreparePointLabel, DrawLabel},
	{0x0002f, "Special",                NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurUnknown,    PreparePointLabel, DrawLabel},
	{0x00100, "Namedpoint (big)",       NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurNamedpoint, PreparePointLabel, DrawLabel},
	{0x00600, "Namedpoint (big point)", NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurNamedpoint, PreparePointLabel, DrawLabel},
	{0x00b00, "Namedpoint (small point)",    NULL,            & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurNamedpoint, PreparePointLabel, DrawLabel},
	{0x00c00, "Namedpoint (small point)",    NULL,            & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurNamedpoint, PreparePointLabel, DrawLabel},
	{0x00d00, "Namedpoint (small point)",    NULL,            & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurNamedpoint, PreparePointLabel, DrawLabel},
	{0x00e00, "Namedpoint (big point)", NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurNamedpoint, PreparePointLabel, DrawLabel},
	{0x01400, "Namedpoint (big)",       NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurNamedpoint, PreparePointLabel, DrawLabel},
	{0x01601, "Fog Horn",               NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurUnknown,    PreparePointLabel, DrawLabel},
	{0x01602, "Radio Beacon",           NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurUnknown,    PreparePointLabel, DrawLabel},
	{0x01603, "Racon",                  NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurUnknown,    PreparePointLabel, DrawLabel},
	{0x01604, "Daybeacon, red Triangle",     NULL,            & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurBeacon,     PreparePointLabel, DrawLabel},
	{0x01605, "Daybeacon, green Square",     NULL,            & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurBeacon,     PreparePointLabel, DrawLabel},
	{0x01606, "Daybeacon, white Diamond",    NULL,            & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurBeacon,     PreparePointLabel, DrawLabel},
	{0x01607, "Unlit navaid, white",    NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurBeacon,     PreparePointLabel, DrawLabel},
	{0x01608, "Unlit navaid, red",      NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurBeacon,     PreparePointLabel, DrawLabel},
	{0x01609, "Unlit navaid, green",    NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurBeacon,     PreparePointLabel, DrawLabel},
	{0x0160a, "Unlit navaid, black",    NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurBeacon,     PreparePointLabel, DrawLabel},
	{0x0160b, "Unlit navaid, yellow/amber",  NULL,            & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurBeacon,     PreparePointLabel, DrawLabel},
	{0x0160c, "Unlit navaid, orange",   NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurBeacon,     PreparePointLabel, DrawLabel},
	{0x0160d, "Unlit navaid, multi colored", NULL,            & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurBeacon,     PreparePointLabel, DrawLabel},
	{0x0160e, "Navaid, unknown",        NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurBeacon,     PreparePointLabel, DrawLabel},
	{0x0160f, "Navaid, white",          NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurBeacon,     PreparePointLabel, DrawLabel},
	{0x01610, "Navaid, red",            NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurBeacon,     PreparePointLabel, DrawLabel},
	{0x01611, "Navaid, green",          NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurBeacon,     PreparePointLabel, DrawLabel},
	{0x01612, "Navaid, yellow/amber",   NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurBeacon,     PreparePointLabel, DrawLabel},
	{0x01613, "Navaid, orange",         NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurBeacon,     PreparePointLabel, DrawLabel},
	{0x01614, "Navaid, violet",         NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurBeacon,     PreparePointLabel, DrawLabel},
	{0x01615, "Navaid, blue",           NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurBeacon,     PreparePointLabel, DrawLabel},
	{0x01616, "Navaid, milti colored",  NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurBeacon,     PreparePointLabel, DrawLabel},
	{0x01701, "Fog Horn",               NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurUnknown,    PreparePointLabel, DrawLabel},
	{0x01702, "Radio Beacon",           NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurUnknown,    PreparePointLabel, DrawLabel},
	{0x01703, "Racon",                  NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurUnknown,    PreparePointLabel, DrawLabel},
	{0x01704, "Daybeacon, red Triangle",     NULL,            & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurBeacon,     PreparePointLabel, DrawLabel},
	{0x01705, "Daybeacon, green Square",     NULL,            & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurBeacon,     PreparePointLabel, DrawLabel},
	{0x01706, "Daybeacon, white Diamond",    NULL,            & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurBeacon,     PreparePointLabel, DrawLabel},
	{0x01707, "Unlit navaid, white",    NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurBeacon,     PreparePointLabel, DrawLabel},
	{0x01708, "Unlit navaid, red",      NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurBeacon,     PreparePointLabel, DrawLabel},
	{0x01709, "Unlit navaid, green",    NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurBeacon,     PreparePointLabel, DrawLabel},
	{0x0170a, "Unlit navaid, black",    NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurBeacon,     PreparePointLabel, DrawLabel},
	{0x0170b, "Unlit navaid, yellow/amber",  NULL,            & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurBeacon,     PreparePointLabel, DrawLabel},
	{0x0170c, "Unlit navaid, orange",   NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurBeacon,     PreparePointLabel, DrawLabel},
	{0x0170d, "Unlit navaid, multi colored", NULL,            & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurBeacon,     PreparePointLabel, DrawLabel},
	{0x0170e, "Navaid, unknown",        NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurBeacon,     PreparePointLabel, DrawLabel},
	{0x0170f, "Navaid, white",          NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurBeacon,     PreparePointLabel, DrawLabel},
	{0x01710, "Navaid, red",            NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurBeacon,     PreparePointLabel, DrawLabel},
	{0x01711, "Navaid, green",          NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurBeacon,     PreparePointLabel, DrawLabel},
	{0x01712, "Navaid, yellow/amber",   NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurBeacon,     PreparePointLabel, DrawLabel},
	{0x01713, "Navaid, orange",         NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurBeacon,     PreparePointLabel, DrawLabel},
	{0x01714, "Navaid, violet",         NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurBeacon,     PreparePointLabel, DrawLabel},
	{0x01715, "Navaid, blue",           NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurBeacon,     PreparePointLabel, DrawLabel},
	{0x01716, "Navaid, milti colored",  NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurBeacon,     PreparePointLabel, DrawLabel},
	{0x01801, "Fog Horn",               NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurUnknown,    PreparePointLabel, DrawLabel},
	{0x01802, "Radio Beacon",           NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurUnknown,    PreparePointLabel, DrawLabel},
	{0x01803, "Racon",                  NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurUnknown,    PreparePointLabel, DrawLabel},
	{0x01804, "Daybeacon, red Triangle",     NULL,            & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurBeacon,     PreparePointLabel, DrawLabel},
	{0x01805, "Daybeacon, green Square",     NULL,            & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurBeacon,     PreparePointLabel, DrawLabel},
	{0x01806, "Daybeacon, white Diamond",    NULL,            & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurBeacon,     PreparePointLabel, DrawLabel},
	{0x01807, "Unlit navaid, white",    NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurBeacon,     PreparePointLabel, DrawLabel},
	{0x01808, "Unlit navaid, red",      NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurBeacon,     PreparePointLabel, DrawLabel},
	{0x01809, "Unlit navaid, green",    NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurBeacon,     PreparePointLabel, DrawLabel},
	{0x0180a, "Unlit navaid, black",    NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurBeacon,     PreparePointLabel, DrawLabel},
	{0x0180b, "Unlit navaid, yellow/amber",  NULL,            & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurBeacon,     PreparePointLabel, DrawLabel},
	{0x0180c, "Unlit navaid, orange",   NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurBeacon,     PreparePointLabel, DrawLabel},
	{0x0180d, "Unlit navaid, multi colored", NULL,            & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurBeacon,     PreparePointLabel, DrawLabel},
	{0x0180e, "Navaid, unknown",        NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurBeacon,     PreparePointLabel, DrawLabel},
	{0x0180f, "Navaid, white",          NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurBeacon,     PreparePointLabel, DrawLabel},
	{0x01810, "Navaid, red",            NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurBeacon,     PreparePointLabel, DrawLabel},
	{0x01811, "Navaid, green",          NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurBeacon,     PreparePointLabel, DrawLabel},
	{0x01812, "Navaid, yellow/amber",   NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurBeacon,     PreparePointLabel, DrawLabel},
	{0x01813, "Navaid, orange",         NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurBeacon,     PreparePointLabel, DrawLabel},
	{0x01814, "Navaid, violet",         NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurBeacon,     PreparePointLabel, DrawLabel},
	{0x01815, "Navaid, blue",           NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurBeacon,     PreparePointLabel, DrawLabel},
	{0x01816, "Navaid, milti colored",  NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurBeacon,     PreparePointLabel, DrawLabel},
	{0x01901, "Fog Horn",               NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurUnknown,    PreparePointLabel, DrawLabel},
	{0x01902, "Radio Beacon",           NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurUnknown,    PreparePointLabel, DrawLabel},
	{0x01903, "Racon",                  NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurUnknown,    PreparePointLabel, DrawLabel},
	{0x01904, "Daybeacon, red Triangle",     NULL,            & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurBeacon,     PreparePointLabel, DrawLabel},
	{0x01905, "Daybeacon, green Square",     NULL,            & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurBeacon,     PreparePointLabel, DrawLabel},
	{0x01906, "Daybeacon, white Diamond",    NULL,            & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurBeacon,     PreparePointLabel, DrawLabel},
	{0x01907, "Unlit navaid, white",    NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurBeacon,     PreparePointLabel, DrawLabel},
	{0x01908, "Unlit navaid, red",      NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurBeacon,     PreparePointLabel, DrawLabel},
	{0x01909, "Unlit navaid, green",    NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurBeacon,     PreparePointLabel, DrawLabel},
	{0x0190a, "Unlit navaid, black",    NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurBeacon,     PreparePointLabel, DrawLabel},
	{0x0190b, "Unlit navaid, yellow/amber",  NULL,            & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurBeacon,     PreparePointLabel, DrawLabel},
	{0x0190c, "Unlit navaid, orange",   NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurBeacon,     PreparePointLabel, DrawLabel},
	{0x0190d, "Unlit navaid, multi colored", NULL,            & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurBeacon,     PreparePointLabel, DrawLabel},
	{0x0190e, "Navaid, unknown",        NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurBeacon,     PreparePointLabel, DrawLabel},
	{0x0190f, "Navaid, white",          NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurBeacon,     PreparePointLabel, DrawLabel},
	{0x01910, "Navaid, red",            NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurBeacon,     PreparePointLabel, DrawLabel},
	{0x01911, "Navaid, green",          NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurBeacon,     PreparePointLabel, DrawLabel},
	{0x01912, "Navaid, yellow/amber",   NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurBeacon,     PreparePointLabel, DrawLabel},
	{0x01913, "Navaid, orange",         NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurBeacon,     PreparePointLabel, DrawLabel},
	{0x01914, "Navaid, violet",         NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurBeacon,     PreparePointLabel, DrawLabel},
	{0x01915, "Navaid, blue",           NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurBeacon,     PreparePointLabel, DrawLabel},
	{0x01916, "Navaid, milti colored",  NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurBeacon,     PreparePointLabel, DrawLabel},
	{0x01a01, "Fog Horn",               NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurUnknown,    PreparePointLabel, DrawLabel},
	{0x01a02, "Radio Beacon",           NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurUnknown,    PreparePointLabel, DrawLabel},
	{0x01a03, "Racon",                  NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurUnknown,    PreparePointLabel, DrawLabel},
	{0x01a04, "Daybeacon, red Triangle",     NULL,            & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurBeacon,     PreparePointLabel, DrawLabel},
	{0x01a05, "Daybeacon, green Square",     NULL,            & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurBeacon,     PreparePointLabel, DrawLabel},
	{0x01a06, "Daybeacon, white Diamond",    NULL,            & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurBeacon,     PreparePointLabel, DrawLabel},
	{0x01a07, "Unlit navaid, white",    NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurBeacon,     PreparePointLabel, DrawLabel},
	{0x01a08, "Unlit navaid, red",      NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurBeacon,     PreparePointLabel, DrawLabel},
	{0x01a09, "Unlit navaid, green",    NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurBeacon,     PreparePointLabel, DrawLabel},
	{0x01a0a, "Unlit navaid, black",    NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurBeacon,     PreparePointLabel, DrawLabel},
	{0x01a0b, "Unlit navaid, yellow/amber",  NULL,            & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurBeacon,     PreparePointLabel, DrawLabel},
	{0x01a0c, "Unlit navaid, orange",   NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurBeacon,     PreparePointLabel, DrawLabel},
	{0x01a0d, "Unlit navaid, multi colored", NULL,            & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurBeacon,     PreparePointLabel, DrawLabel},
	{0x01a0e, "Navaid, unknown",        NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurBeacon,     PreparePointLabel, DrawLabel},
	{0x01a0f, "Navaid, white",          NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurBeacon,     PreparePointLabel, DrawLabel},
	{0x01a10, "Navaid, red",            NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurBeacon,     PreparePointLabel, DrawLabel},
	{0x01a11, "Navaid, green",          NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurBeacon,     PreparePointLabel, DrawLabel},
	{0x01a12, "Navaid, yellow/amber",   NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurBeacon,     PreparePointLabel, DrawLabel},
	{0x01a13, "Navaid, orange",         NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurBeacon,     PreparePointLabel, DrawLabel},
	{0x01a14, "Navaid, violet",         NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurBeacon,     PreparePointLabel, DrawLabel},
	{0x01a15, "Navaid, blue",           NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurBeacon,     PreparePointLabel, DrawLabel},
	{0x01a16, "Navaid, milti colored",  NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurBeacon,     PreparePointLabel, DrawLabel},
	{0x01b01, "Fog Horn",               NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurUnknown,    PreparePointLabel, DrawLabel},
	{0x01b02, "Radio Beacon",           NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurUnknown,    PreparePointLabel, DrawLabel},
	{0x01b03, "Racon",                  NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurUnknown,    PreparePointLabel, DrawLabel},
	{0x01b04, "Daybeacon, red Triangle",     NULL,            & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurBeacon,     PreparePointLabel, DrawLabel},
	{0x01b05, "Daybeacon, green Square",     NULL,            & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurBeacon,     PreparePointLabel, DrawLabel},
	{0x01b06, "Daybeacon, white Diamond",    NULL,            & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurBeacon,     PreparePointLabel, DrawLabel},
	{0x01b07, "Unlit navaid, white",    NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurBeacon,     PreparePointLabel, DrawLabel},
	{0x01b08, "Unlit navaid, red",      NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurBeacon,     PreparePointLabel, DrawLabel},
	{0x01b09, "Unlit navaid, green",    NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurBeacon,     PreparePointLabel, DrawLabel},
	{0x01b0a, "Unlit navaid, black",    NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurBeacon,     PreparePointLabel, DrawLabel},
	{0x01b0b, "Unlit navaid, yellow/amber",  NULL,            & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurBeacon,     PreparePointLabel, DrawLabel},
	{0x01b0c, "Unlit navaid, orange",   NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurBeacon,     PreparePointLabel, DrawLabel},
	{0x01b0d, "Unlit navaid, multi colored", NULL,            & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurBeacon,     PreparePointLabel, DrawLabel},
	{0x01b0e, "Navaid, unknown",        NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurBeacon,     PreparePointLabel, DrawLabel},
	{0x01b0f, "Navaid, white",          NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurBeacon,     PreparePointLabel, DrawLabel},
	{0x01b10, "Navaid, red",            NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurBeacon,     PreparePointLabel, DrawLabel},
	{0x01b11, "Navaid, green",          NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurBeacon,     PreparePointLabel, DrawLabel},
	{0x01b12, "Navaid, yellow/amber",   NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurBeacon,     PreparePointLabel, DrawLabel},
	{0x01b13, "Navaid, orange",         NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurBeacon,     PreparePointLabel, DrawLabel},
	{0x01b14, "Navaid, violet",         NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurBeacon,     PreparePointLabel, DrawLabel},
	{0x01b15, "Navaid, blue",           NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurBeacon,     PreparePointLabel, DrawLabel},
	{0x01b16, "Navaid, milti colored",  NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurBeacon,     PreparePointLabel, DrawLabel},
	{0x01c00, "Unclasified obstruction", NULL,                & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurUnknown,    PreparePointLabel, DrawLabel},
	{0x01c01, "Wreck",                  NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurUnknown,    PreparePointLabel, DrawLabel},
	{0x01c02, "Submerged wreck, dangerous",     NULL,         & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurUnknown,    PreparePointLabel, DrawLabel},
	{0x01c03, "Submerged wreck, non-dangerous", NULL,         & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurUnknown,    PreparePointLabel, DrawLabel},
	{0x01c04, "Wreck, cleared by Wire-drag",    NULL,         & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurUnknown,    PreparePointLabel, DrawLabel},
	{0x01c05, "Obstr., visible at high Water",  NULL,         & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurUnknown,    PreparePointLabel, DrawLabel},
	{0x01c06, "Obstruction, awash",     NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurUnknown,    PreparePointLabel, DrawLabel},
	{0x01c07, "Obstruction, submerged", NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurUnknown,    PreparePointLabel, DrawLabel},
	{0x01c08, "Obstr., cleared by Wire-drag",   NULL,         & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurUnknown,    PreparePointLabel, DrawLabel},
	{0x01c09, "Rock, awash",            NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurReef,       PreparePointLabel, DrawLabel},
	{0x01c0a, "Rock at low Water",      NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurReef,       PreparePointLabel, DrawLabel},
	{0x01c0b, "Sounding",               NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurUnknown,    PreparePointLabel, DrawLabel},
	{0x01e00, "Namedpoint, middle",     NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurNamedpoint, PreparePointLabel, DrawLabel},
	{0x02100, "Exit with facilities",   NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurExit,       PreparePointLabel, DrawLabel},
	{0x02800, "Namedpoint (small point)", NULL,               & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurNamedpoint, PreparePointLabel, DrawLabel},
	{0x02a01, "Dining (American)",      NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurDining,     PreparePointLabel, DrawLabel},
	{0x02a02, "Dining (Asian)",         NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurDining,     PreparePointLabel, DrawLabel},
	{0x02a03, "Dining (barbecue)",      NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurDining,     PreparePointLabel, DrawLabel},
	{0x02a04, "Dining (Chinese)",       NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurDining,     PreparePointLabel, DrawLabel},
	{0x02a05, "Dining (Deli/Bakery)",   NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurDining,     PreparePointLabel, DrawLabel},
	{0x02a06, "Dining (international)", NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurDining,     PreparePointLabel, DrawLabel},
	{0x02a07, "Fast food",              NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurDining,     PreparePointLabel, DrawLabel},
	{0x02a08, "Dining (Italian)",       NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurDining,     PreparePointLabel, DrawLabel},
	{0x02a09, "Dining (Mexican)",       NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurDining,     PreparePointLabel, DrawLabel},
	{0x02a0a, "Dining (pizza)",         NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurDining,     PreparePointLabel, DrawLabel},
	{0x02a0b, "Dining (sea food)",      NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurDining,     PreparePointLabel, DrawLabel},
	{0x02a0c, "Dining (steak/grill)",   NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurDining,     PreparePointLabel, DrawLabel},
	{0x02a0d, "Dining (bagel/donut)",   NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurDining,     PreparePointLabel, DrawLabel},
	{0x02a0e, "Dining (cafe/diner)",    NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurDining,     PreparePointLabel, DrawLabel},
	{0x02a0f, "Dining (French)",        NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurDining,     PreparePointLabel, DrawLabel},
	{0x02a10, "Dining (German)",        NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurDining,     PreparePointLabel, DrawLabel},
	{0x02a11, "Dining (British Isle)",  NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurDining,     PreparePointLabel, DrawLabel},
	{0x02b00, "Hotel (Other)",          NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurHotel,      PreparePointLabel, DrawLabel},
	{0x02b01, "Hotel/Motel",            NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurHotel,      PreparePointLabel, DrawLabel},
	{0x02b02, "Bed&Breakfast inn",      NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurHotel,      PreparePointLabel, DrawLabel},
	{0x02b03, "Camping/RV-Park",        NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurUnknown,    PreparePointLabel, DrawLabel},
	{0x02b04, "Resort",                 NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurUnknown,    PreparePointLabel, DrawLabel},
	{0x02c01, "Amusment park",          NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurUnknown,    PreparePointLabel, DrawLabel},
	{0x02c02, "Museum/History",         NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurMuseum,     PreparePointLabel, DrawLabel},
	{0x02c03, "Libraries",              NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurLibrary,    PreparePointLabel, DrawLabel},
	{0x02c04, "Land Mark",              NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurUnknown,    PreparePointLabel, DrawLabel},
	{0x02c05, "School",                 NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurSchool,     PreparePointLabel, DrawLabel},
	{0x02c06, "Park",                   NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurUnknown,    PreparePointLabel, DrawLabel},
	{0x02c07, "Zoo",                    NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurZoo,        PreparePointLabel, DrawLabel},
	{0x02c08, "Sportcenter, stadion, race", NULL,             & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurSports,     PreparePointLabel, DrawLabel},
	{0x02c09, "Conference",             NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurUnknown,    PreparePointLabel, DrawLabel},
	{0x02c0a, "Vine",                   NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurBar,        PreparePointLabel, DrawLabel},
	{0x02d01, "Theatre",                NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurTheatre,    PreparePointLabel, DrawLabel},
	{0x02d02, "Bar",                    NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurBar,        PreparePointLabel, DrawLabel},
	{0x02d03, "Cinema",                 NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurCinema,     PreparePointLabel, DrawLabel},
	{0x02d04, "Casino",                 NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurUnknown,    PreparePointLabel, DrawLabel},
	{0x02d05, "Golf",                   NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurUnknown,    PreparePointLabel, DrawLabel},
	{0x02d06, "Skiing center",          NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurUnknown,    PreparePointLabel, DrawLabel},
	{0x02d07, "Bowling",                NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurUnknown,    PreparePointLabel, DrawLabel},
	{0x02d08, "Ice-skate",              NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurUnknown,    PreparePointLabel, DrawLabel},
	{0x02d09, "Swimming",               NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurSwimming,   PreparePointLabel, DrawLabel},
	{0x02d0a, "Sports",                 NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurSports,     PreparePointLabel, DrawLabel},
	{0x02d0b, "Sport airport",          NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurUnknown,    PreparePointLabel, DrawLabel},
	{0x02e01, "Department store",       NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurShopping,   PreparePointLabel, DrawLabel},
	{0x02e02, "Grocery",                NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurUnknown,    PreparePointLabel, DrawLabel},
	{0x02e03, "General merchandiser",   NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurUnknown,    PreparePointLabel, DrawLabel},
	{0x02e04, "Shopping-center",        NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurShopping,   PreparePointLabel, DrawLabel},
	{0x02e05, "Pharmacie",              NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurPharmacie,  PreparePointLabel, DrawLabel},
	{0x02e06, "Convinience",            NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurUnknown,    PreparePointLabel, DrawLabel},
	{0x02f01, "Petrol",                 NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurPetrol,     PreparePointLabel, DrawLabel},
	{0x02f02, "Auto rental",            NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurAutoRental, PreparePointLabel, DrawLabel},
	{0x02f03, "Auto repair",            NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurAutoRepair, PreparePointLabel, DrawLabel},
	{0x02f04, "Airport",                NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurAirport,    PreparePointLabel, DrawLabel},
	{0x02f05, "Post office",            NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurPostOffice, PreparePointLabel, DrawLabel},
	{0x02f06, "Bank",                   NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurBank,       PreparePointLabel, DrawLabel},
	{0x02f07, "Autohandel",             NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurUnknown,    PreparePointLabel, DrawLabel},
	{0x02f08, "Bus station",            NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurBusStation, PreparePointLabel, DrawLabel},
	{0x02f09, "Marina",                 NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurWarf,       PreparePointLabel, DrawLabel},
	{0x02f0a, "Wrecker service",        NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurUnknown,    PreparePointLabel, DrawLabel},
	{0x02f0b, "Parking",                NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurParking,    PreparePointLabel, DrawLabel},
	{0x02f0c, "Restroom",               NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurRestroom,   PreparePointLabel, DrawLabel},
	{0x02f0d, "Automobil club",         NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurUnknown,    PreparePointLabel, DrawLabel},
	{0x02f0e, "Car Wash",               NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurAutoWash,   PreparePointLabel, DrawLabel},
	{0x03001, "Policestation",          NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurPolice,     PreparePointLabel, DrawLabel},
	{0x03002, "Hospital",               NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurHospital,   PreparePointLabel, DrawLabel},
	{0x03003, "Public office",          NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurUnknown,    PreparePointLabel, DrawLabel},
	{0x03004, "Court",                  NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurUnknown,    PreparePointLabel, DrawLabel},
	{0x03005, "Disco",                  NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurUnknown,    PreparePointLabel, DrawLabel},
	{0x03006, "Zoll",                   NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurUnknown,    PreparePointLabel, DrawLabel},
	{0x04000, "Golf",                   NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurUnknown,    PreparePointLabel, DrawLabel},
	{0x04100, "Fish",                   NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurFish,       PreparePointLabel, DrawLabel},
	{0x04200, "Wreck",                  NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurUnknown,    PreparePointLabel, DrawLabel},
	{0x04300, "Marina",                 NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurWarf,       PreparePointLabel, DrawLabel},
	{0x04400, "Gas station",            NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurPetrol,     PreparePointLabel, DrawLabel},
	{0x04500, "Restaurant",             NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurDining,     PreparePointLabel, DrawLabel},
	{0x04600, "Bar",                    NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurBar,        PreparePointLabel, DrawLabel},
	{0x04700, "Boat ramp",              NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurUnknown,    PreparePointLabel, DrawLabel},
	{0x04800, "Camping",                NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurCamping,    PreparePointLabel, DrawLabel},
	{0x04900, "Park",                   NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurUnknown,    PreparePointLabel, DrawLabel},
	{0x04a00, "Picnic area",            NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurPicnic,     PreparePointLabel, DrawLabel},
	{0x04b00, "Hospital",               NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurHospital,   PreparePointLabel, DrawLabel},
	{0x04c00, "Information",            NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurInformation,PreparePointLabel, DrawLabel},
	{0x04d00, "Parking",                NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurParking,    PreparePointLabel, DrawLabel},
	{0x04e00, "Restroom",               NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurRestroom,   PreparePointLabel, DrawLabel},
	{0x04f00, "Shower",                 NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurUnknown,    PreparePointLabel, DrawLabel},
	{0x05000, "Drinking water",         NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurDrinkingWater, PreparePointLabel, DrawLabel},
	{0x05100, "Telephone",              NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurTelephone,  PreparePointLabel, DrawLabel},
	{0x05200, "Scenic area",            NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurUnknown,    PreparePointLabel, DrawLabel},
	{0x05300, "Skiing",                 NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurUnknown,    PreparePointLabel, DrawLabel},
	{0x05400, "Swimming",               NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurSwimming,   PreparePointLabel, DrawLabel},
	{0x05500, "Dam",                    NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurUnknown,    PreparePointLabel, DrawLabel},
	{0x05700, "Danger area",            NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurDanger,     PreparePointLabel, DrawLabel},
	{0x05900, "Airport",                NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurAirport,    PreparePointLabel, DrawLabel},
	{0x05904, "Heliport",               NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurAirport,    PreparePointLabel, DrawLabel},
	{0x05905, "Airport",                NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurAirport,    PreparePointLabel, DrawLabel},
	{0x05d00, "Daymark, green square",  NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurUnknown,    PreparePointLabel, DrawLabel},
	{0x05e00, "Daymark, red triangle",  NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurUnknown,    PreparePointLabel, DrawLabel},
	{0x06401, "Bridge",                 NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurBridge,     PreparePointLabel, DrawLabel},
	{0x06402, "Building",               NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurBuilding,   PreparePointLabel, DrawLabel},
	{0x06403, "Cemetery",               NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurCemetery,   PreparePointLabel, DrawLabel},
	{0x06404, "Church",                 NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurChurch,     PreparePointLabel, DrawLabel},
	{0x06405, "Civil",                  NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurUnknown,    PreparePointLabel, DrawLabel},
	{0x06406, "Crossing",               NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurPassing,    PreparePointLabel, DrawLabel},
	{0x06407, "Dam",                    NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurUnknown,    PreparePointLabel, DrawLabel},
	{0x06408, "Hospital",               NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurHospital,   PreparePointLabel, DrawLabel},
	{0x06409, "Levee",                  NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurUnknown,    PreparePointLabel, DrawLabel},
	{0x0640a, "Locale",                 NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurUnknown,    PreparePointLabel, DrawLabel},
	{0x0640b, "Military",               NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurMilitary,   PreparePointLabel, DrawLabel},
	{0x0640c, "Mine",                   NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurUnknown,    PreparePointLabel, DrawLabel},
	{0x0640d, "Oil field",              NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurUnknown,    PreparePointLabel, DrawLabel},
	{0x0640e, "Park",                   NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurUnknown,    PreparePointLabel, DrawLabel},
	{0x0640f, "Post",                   NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurUnknown,    PreparePointLabel, DrawLabel},
	{0x06410, "School",                 NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurSchool,     PreparePointLabel, DrawLabel},
	{0x06411, "Tower",                  NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurTower,      PreparePointLabel, DrawLabel},
	{0x06412, "Trial",                  NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurUnknown,    PreparePointLabel, DrawLabel},
	{0x06413, "Tunnel",                 NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurUnknown,    PreparePointLabel, DrawLabel},
	{0x06414, "Drinking water",         NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurDrinkingWater, PreparePointLabel, DrawLabel},
	{0x06415, "Ghost town",             NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurUnknown,    PreparePointLabel, DrawLabel},
	{0x06416, "Subdivision",            NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurUnknown,    PreparePointLabel, DrawLabel},
	{0x06501, "Arroyo",                 NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurUnknown,    PreparePointLabel, DrawLabel},
	{0x06502, "Sand bar",               NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurUnknown,    PreparePointLabel, DrawLabel},
	{0x06503, "Bay",                    NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurUnknown,    PreparePointLabel, DrawLabel},
	{0x06504, "Bend",                   NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurUnknown,    PreparePointLabel, DrawLabel},
	{0x06505, "Canal",                  NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurUnknown,    PreparePointLabel, DrawLabel},
	{0x06506, "Channel",                NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurUnknown,    PreparePointLabel, DrawLabel},
	{0x06507, "Cove",                   NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurUnknown,    PreparePointLabel, DrawLabel},
	{0x06508, "Falls",                  NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurFalls,      PreparePointLabel, DrawLabel},
	{0x06509, "Geyser",                 NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurUnknown,    PreparePointLabel, DrawLabel},
	{0x0650a, "Glacier",                NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurUnknown,    PreparePointLabel, DrawLabel},
	{0x0650b, "Harbor",                 NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurWarf,       PreparePointLabel, DrawLabel},
	{0x0650c, "Island",                 NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurIsland,     PreparePointLabel, DrawLabel},
	{0x0650d, "Lake",                   NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurReservoir,  PreparePointLabel, DrawLabel},
	{0x0650e, "Rapids",                 NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurFalls,      PreparePointLabel, DrawLabel},
	{0x0650f, "Reservoir",              NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurReservoir,  PreparePointLabel, DrawLabel},
	{0x06510, "Sea",                    NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurReservoir,  PreparePointLabel, DrawLabel},
	{0x06511, "Spring",                 NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurSpring,     PreparePointLabel, DrawLabel},
	{0x06512, "Stream",                 NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurUnknown,    PreparePointLabel, DrawLabel},
	{0x06513, "Swamp",                  NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurUnknown,    PreparePointLabel, DrawLabel},
	{0x06601, "Arch",                   NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurUnknown,    PreparePointLabel, DrawLabel},
	{0x06602, "Area",                   NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurUnknown,    PreparePointLabel, DrawLabel},
	{0x06603, "Basin",                  NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurReservoir,  PreparePointLabel, DrawLabel},
	{0x06604, "Beach",                  NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurUnknown,    PreparePointLabel, DrawLabel},
	{0x06605, "Beach",                  NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurUnknown,    PreparePointLabel, DrawLabel},
	{0x06606, "Cape",                   NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurUnknown,    PreparePointLabel, DrawLabel},
	{0x06607, "Cliff",                  NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurCliff,      PreparePointLabel, DrawLabel},
	{0x06608, "Crater",                 NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurUnknown,    PreparePointLabel, DrawLabel},
	{0x06609, "Flat",                   NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurUnknown,    PreparePointLabel, DrawLabel},
	{0x0660a, "Forest",                 NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurTree,       PreparePointLabel, DrawLabel},
	{0x0660b, "Gap",                    NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurUnknown,    PreparePointLabel, DrawLabel},
	{0x0660c, "Gut",                    NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurUnknown,    PreparePointLabel, DrawLabel},
	{0x0660d, "Isthmus",                NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurUnknown,    PreparePointLabel, DrawLabel},
	{0x0660e, "Lava",                   NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurUnknown,    PreparePointLabel, DrawLabel},
	{0x0660f, "Pillar",                 NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurUnknown,    PreparePointLabel, DrawLabel},
	{0x06610, "Plain",                  NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurUnknown,    PreparePointLabel, DrawLabel},
	{0x06611, "Range",                  NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurUnknown,    PreparePointLabel, DrawLabel},
	{0x06612, "Reserve",                NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurUnknown,    PreparePointLabel, DrawLabel},
	{0x06613, "Ridge",                  NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurRock,       PreparePointLabel, DrawLabel},
	{0x06614, "Rock",                   NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurRock,       PreparePointLabel, DrawLabel},
	{0x06615, "Slope",                  NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurUnknown,    PreparePointLabel, DrawLabel},
	{0x06616, "Summit",                 NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurHill,       PreparePointLabel, DrawLabel},
	{0x06617, "Valley",                 NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurUnknown,    PreparePointLabel, DrawLabel},
	{0x06618, "Woods",                  NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurTree,       PreparePointLabel, DrawLabel},

	{0x10001, "City1",                  NULL,                 & g_hNullPen, NULL,                   & g_hCityFont, c_colBlack, & g_hcurCityBig,    PreparePointLabel, DrawLabel},
	{0x10002, "City2",                  NULL,                 & g_hNullPen, NULL,                   & g_hCityFont, c_colBlack, & g_hcurCityBig,    PreparePointLabel, DrawLabel},
	{0x10003, "City3",                  NULL,                 & g_hNullPen, NULL,                   & g_hCityFont, c_colBlack, & g_hcurCityBig,    PreparePointLabel, DrawLabel},
	{0x10004, "City4",                  NULL,                 & g_hNullPen, NULL,                   & g_hCityFont, c_colBlack, & g_hcurCityMedium, PreparePointLabel, DrawLabel},
	{0x10005, "City5",                  NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurCitySmall,  PreparePointLabel, DrawLabel},
	{0x10006, "City6",                  NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurCitySmall,  PreparePointLabel, DrawLabel},
	{0x10007, "City7",                  NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurSettlement, PreparePointLabel, DrawLabel},
	{0x10008, "City8",                  NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurSettlement, PreparePointLabel, DrawLabel},
	{0x10009, "City9",                  NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurSettlement, PreparePointLabel, DrawLabel},
	{0x1000a, "City10",                 NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurSettlement, PreparePointLabel, DrawLabel},
	{0x1000b, "City11",                 NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurSettlement, PreparePointLabel, DrawLabel},
	{0x1000c, "City12",                 NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurSettlement, PreparePointLabel, DrawLabel},
	{0x1000d, "City13",                 NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurSettlement, PreparePointLabel, DrawLabel},
	{0x1000e, "City14",                 NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurSettlement, PreparePointLabel, DrawLabel},
	{0x1000f, "City15",                 NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurSettlement, PreparePointLabel, DrawLabel},
	{0x10010, "City16",                 NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurSettlement, PreparePointLabel, DrawLabel},
	{0x10011, "City17",                 NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurSettlement, PreparePointLabel, DrawLabel},
	{0x10012, "City18",                 NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurSettlement, PreparePointLabel, DrawLabel},
	{0x10013, "City19",                 NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurSettlement, PreparePointLabel, DrawLabel},
	{0x10014, "City20",                 NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurSettlement, PreparePointLabel, DrawLabel},
	{0x10015, "Region",                 NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurUnknown,    PreparePointLabel, DrawLabel},
	{0x10040, "Big region",             NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurUnknown,    PreparePointLabel, DrawLabel},
	{0x10041, "Fish",                   NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurFish,       PreparePointLabel, DrawLabel},
	{0x10042, "Wreck",                  NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurUnknown,    PreparePointLabel, DrawLabel},
	{0x10043, "Marina",                 NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurWarf,       PreparePointLabel, DrawLabel},
	{0x10044, "Fuel",                   NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurPetrol,     PreparePointLabel, DrawLabel},
	{0x10047, "Ramp",                   NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurUnknown,    PreparePointLabel, DrawLabel},
	{0x10048, "Campground",             NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurUnknown,    PreparePointLabel, DrawLabel},
	{0x10049, "Park",                   NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurUnknown,    PreparePointLabel, DrawLabel},
	{0x1004a, "Picnic area",            NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurUnknown,    PreparePointLabel, DrawLabel},
	{0x1004b, "First aid",              NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurUnknown,    PreparePointLabel, DrawLabel},
	{0x1004c, "Information",            NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurInformation,PreparePointLabel, DrawLabel},
	{0x1004e, "Restroom",               NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurRestroom,   PreparePointLabel, DrawLabel},
	{0x1004f, "Showers",                NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurUnknown,    PreparePointLabel, DrawLabel},
	{0x10050, "Drinking water",         NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurDrinkingWater, PreparePointLabel, DrawLabel},
	{0x10051, "Telephone",              NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurTelephone,  PreparePointLabel, DrawLabel},
	{0x10052, "Scenic area",            NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurUnknown,    PreparePointLabel, DrawLabel},
	{0x10053, "Skiing",                 NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurUnknown,    PreparePointLabel, DrawLabel},
	{0x10054, "Swimming",               NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurSwimming,   PreparePointLabel, DrawLabel},
	{0x10055, "Dam",                    NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurUnknown,    PreparePointLabel, DrawLabel},
	{0x10057, "Danger area",            NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurDanger,     PreparePointLabel, DrawLabel},
	{0x10059, "Airport",                NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurAirport,    PreparePointLabel, DrawLabel},
	{0x1005d, "Daymark, green square",  NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurUnknown,    PreparePointLabel, DrawLabel},
	{0x1005e, "Daymark, red triangle",  NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurUnknown,    PreparePointLabel, DrawLabel},
	{0x10062, "Height point",           NULL,                 & g_hNullPen, NULL,                   NULL, c_colHeight,   & g_hcurTop,        PreparePointLabel, DrawMetersLabel},
	{0x10063, "Height point",           NULL,                 & g_hNullPen, NULL,                   NULL, c_colHeight,   & g_hcurTop,        PreparePointLabel, DrawMetersLabel},
	{0x10064, "(blue point)",           NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurUnknown,    PreparePointLabel, DrawLabel},
	{0x10065, "(blue point)",           NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurUnknown,    PreparePointLabel, DrawLabel},
	{0x10066, "(blue point)",           NULL,                 & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurUnknown,    PreparePointLabel, DrawLabel},

	// Polylines.
	{0x20001, "Major HWY, thick",       NULL,                 & g_hHighwayPen1, & g_hHighwayPen2,   NULL, c_colDkGray,   & g_hcurUnknown,    PrepareRoadLabel, DrawRoadLabel},
	{0x20002, "Principal HWY, thick",   NULL,                 & g_hEnhancedRoadPen, NULL,           NULL, c_colDkGray,   & g_hcurUnknown,    PrepareRoadLabel, DrawRoadLabel},
	{0x20003, "Principal HWY, medium",  NULL,                 & g_hEnhancedRoadPen, NULL,           NULL, c_colDkGray,   & g_hcurUnknown,    PrepareRoadLabel, DrawRoadLabel},
	{0x20004, "Arterial road, medium",  NULL,                 & g_hThickRoadPen, NULL,              NULL, c_colDkGray,   & g_hcurUnknown,    PrepareRoadLabel, DrawRoadLabel},
	{0x20005, "Arterial road, thin",    NULL,                 & g_hThickRoadPen, NULL,              NULL, c_colDkGray,   & g_hcurUnknown,    PrepareRoadLabel, DrawRoadLabel},
	{0x20006, "Road, thin",             NULL,                 & g_hThinRoadPen, NULL,               NULL, c_colDkGray,   & g_hcurUnknown,    PrepareRoadLabel, DrawRoadLabel},
	{0x20007, "Alley, thin",            NULL,                 & g_hThinRoadPen, NULL,               NULL, c_colDkGray,   & g_hcurUnknown,    PrepareRoadLabel, DrawRoadLabel},
	{0x20008, "Ramp, thin",             NULL,                 & g_hThinRoadPen, NULL,               NULL, c_colDkGray,   & g_hcurUnknown,    PrepareRoadLabel, DrawRoadLabel},
	{0x20009, "Ramp, thin",             NULL,                 & g_hThinRoadPen, NULL,               NULL, c_colDkGray,   & g_hcurUnknown,    PrepareRoadLabel, DrawRoadLabel},
	{0x2000a, "Unpaved road, thin",     NULL,                 & g_hThinRoadPen, NULL,               NULL, c_colDkGray,   & g_hcurUnknown,    PrepareRoadLabel, DrawRoadLabel},
	{0x2000b, "Major HWY connector, thick", NULL,             & g_hHighwayPen1, & g_hHighwayPen2,   NULL, c_colDkGray,   & g_hcurUnknown,    PrepareRoadLabel, DrawRoadLabel},
	{0x2000c, "Roundabout, thin",       NULL,                 & g_hThickRoadPen, NULL,              NULL, c_colDkGray,   & g_hcurUnknown,    PrepareRoadLabel, DrawRoadLabel},
	{0x20014, "Railroad",               NULL,                 & g_hRailroadPen1, & g_hRailroadPen2, NULL, c_colBlack,    & g_hcurUnknown,    PreparePointLabel, DrawLabel},
	{0x20015, "Shoreline",              NULL,                 & g_hCoastPen, NULL,                  NULL, c_colBlack,    & g_hcurUnknown,    PreparePointLabel, DrawLabel},
	{0x20016, "Trail",                  NULL,                 & g_hGroundRoadPen, NULL,             NULL, c_colBlack,    & g_hcurUnknown,    PreparePointLabel, DrawLabel},
	{0x20018, "Stream",                 NULL,                 & g_hWaterPen, NULL,                  NULL, c_colWater,    & g_hcurUnknown,    PrepareRiverLabel, DrawRiverLabel},
	{0x20019, "Time-zone",              NULL,                 & g_hUnknownPen, NULL,                NULL, c_colBlack,    & g_hcurUnknown,    PreparePointLabel, DrawLabel},
	{0x2001a, "Ferry",                  NULL,                 & g_hGroundRoadPen, NULL,             NULL, c_colBlack,    & g_hcurUnknown,    PreparePointLabel, DrawLabel},
	{0x2001b, "Ferry",                  NULL,                 & g_hGroundRoadPen, NULL,             NULL, c_colBlack,    & g_hcurUnknown,    PreparePointLabel, DrawLabel},
	{0x2001c, "Political boundary",     NULL,                 & g_hBoundaryPen, NULL,               NULL, c_colBoundary, & g_hcurUnknown,    PrepareBoundaryLabel, DrawBoundaryLabel},
	{0x2001d, "County boundary",        NULL,                 & g_hBoundaryPen, NULL,               NULL, c_colBoundary, & g_hcurUnknown,    PrepareBoundaryLabel, DrawBoundaryLabel},
	{0x2001e, "Intl. boundary",         NULL,                 & g_hBoundaryPen, NULL,               NULL, c_colBoundary, & g_hcurUnknown,    PrepareBoundaryLabel, DrawBoundaryLabel},
	{0x2001f, "River",                  NULL,                 & g_hWaterPen, NULL,                  NULL, c_colWater,    & g_hcurUnknown,    PrepareRiverLabel, DrawRiverLabel},
	{0x20020, "Land contour, thin",     NULL,                 & g_hHeightPen, NULL,                 NULL, c_colHeight,   & g_hcurUnknown,    PrepareHeightLabel, DrawHeightLabel},
	{0x20021, "Land contour, medium",   NULL,                 & g_hHeightPen, NULL,                 NULL, c_colHeight,   & g_hcurUnknown,    PrepareHeightLabel, DrawHeightLabel},
	{0x20022, "Land contour, thick",    NULL,                 & g_hHeightPen, NULL,                 NULL, c_colHeight,   & g_hcurUnknown,    PrepareHeightLabel, DrawHeightLabel},
	{0x20023, "Depth contour, thin",    NULL,                 & g_hDepthPen, NULL,                  NULL, c_colDepth,    & g_hcurUnknown,    PrepareHeightLabel, DrawHeightLabel},
	{0x20024, "Depth contour, medium",  NULL,                 & g_hDepthPen, NULL,                  NULL, c_colDepth,    & g_hcurUnknown,    PrepareHeightLabel, DrawHeightLabel},
	{0x20025, "Depth contour, thick",   NULL,                 & g_hDepthPen, NULL,                  NULL, c_colDepth,    & g_hcurUnknown,    PrepareHeightLabel, DrawHeightLabel},
	{0x20026, "Intermittent river",     NULL,                 & g_hDryWaterPen, NULL,               NULL, c_colWater,    & g_hcurUnknown,    PrepareRiverLabel, DrawRiverLabel},
	{0x20027, "Airport runway",         NULL,                 & g_hHighwayPen1, & g_hHighwayPen2,   NULL, c_colBlack,    & g_hcurUnknown,    PrepareRoadLabel, DrawRoadLabel},
	{0x20028, "Pipeline",               NULL,                 & g_hUnknownPen, NULL,                NULL, c_colBlack,    & g_hcurUnknown,    PreparePointLabel, DrawLabel},
	{0x20029, "Power line",             NULL,                 & g_hUnknownPen, NULL,                NULL, c_colBlack,    & g_hcurUnknown,    PreparePointLabel, DrawLabel},
	{0x2002a, "Marine boundary",        NULL,                 & g_hCoastPen, NULL,                  NULL, c_colBlack,    & g_hcurUnknown,    PreparePointLabel, DrawLabel},
	{0x2002b, "Marine hazard",          NULL,                 & g_hUnknownPen, NULL,                NULL, c_colBlack,    & g_hcurReef,       PreparePointLabel, DrawLabel},

	// Polygons.
	{0x30001, "City",                   & g_hSettlementBrush, & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurSettlement, PreparePointLabel, DrawLabel},
	{0x30002, "City",                   & g_hSettlementBrush, & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurSettlement, PreparePointLabel, DrawLabel},
	{0x30003, "City",                   & g_hSettlementBrush, & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurSettlement, PreparePointLabel, DrawLabel},
	{0x30004, "Military",               & g_hGrayBrush,       & g_hBlackThinPen, NULL,              NULL, c_colBlack,    & g_hcurUnknown,    PreparePointLabel, DrawLabel},
	{0x30005, "Parking lot",            & g_hGrayBrush,       & g_hBlackThinPen, NULL,              NULL, c_colBlack,    & g_hcurUnknown,    PreparePointLabel, DrawLabel},
	{0x30006, "Parking garage",         & g_hGrayBrush,       & g_hBlackThinPen, NULL,              NULL, c_colBlack,    & g_hcurUnknown,    PreparePointLabel, DrawLabel},
	{0x30007, "Airport",                & g_hGrayBrush,       & g_hBlackThinPen, NULL,              NULL, c_colBlack,    & g_hcurUnknown,    PreparePointLabel, DrawLabel},
	{0x30008, "Shopping center",        & g_hHousingBrush,    & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurUnknown,    PreparePointLabel, DrawLabel},
	{0x30009, "Marina",                 & g_hHousingBrush,    & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurUnknown,    PreparePointLabel, DrawLabel},
	{0x3000a, "University",             & g_hHousingBrush,    & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurUnknown,    PreparePointLabel, DrawLabel},
	{0x3000b, "Hospital",               & g_hHousingBrush,    & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurUnknown,    PreparePointLabel, DrawLabel},
	{0x3000c, "Industrial",             & g_hGrayBrush,       & g_hBlackThinPen, NULL,              NULL, c_colBlack,    & g_hcurUnknown,    PreparePointLabel, DrawLabel},
	{0x3000d, "Reservation",            & g_hForestBrush,     & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurUnknown,    PreparePointLabel, DrawLabel},
	{0x3000e, "Airport runway",         & g_hWhiteBrush,      & g_hBlackThinPen, NULL,              NULL, c_colBlack,    & g_hcurUnknown,    PreparePointLabel, DrawLabel},
	{0x30014, "National park",          & g_hForestBrush,     & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurUnknown,    PreparePointLabel, DrawLabel},
	{0x30015, "National park",          & g_hForestBrush,     & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurUnknown,    PreparePointLabel, DrawLabel},
	{0x30016, "National park",          & g_hForestBrush,     & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurUnknown,    PreparePointLabel, DrawLabel},
	{0x30017, "City park",              & g_hCityParkBrush,   & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurUnknown,    PreparePointLabel, DrawLabel},
	{0x30018, "Golf",                   & g_hUnknownBrush,    & g_hUnknownPen, NULL,                NULL, c_colBlack,    & g_hcurUnknown,    PreparePointLabel, DrawLabel},
	{0x30019, "Sport",                  & g_hGrayBrush,       & g_hBlackThinPen, NULL,              NULL, c_colBlack,    & g_hcurUnknown,    PreparePointLabel, DrawLabel},
	{0x3001a, "Cemetery",               & g_hCemeteryBrush,   & g_hBlackThinPen, NULL,              NULL, c_colBlack,    & g_hcurUnknown,    PreparePointLabel, DrawLabel},
	{0x3001e, "State park",             & g_hForestBrush,     & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurUnknown,    PreparePointLabel, DrawLabel},
	{0x3001f, "State park",             & g_hForestBrush,     & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurUnknown,    PreparePointLabel, DrawLabel},
	{0x30028, "Ocean",                  & g_hWaterBrush,      & g_hCoastPen, NULL,                  NULL, c_colWater2,   & g_hcurUnknown,    PrepareLakeLabel, DrawLakeLabel},
	{0x30032, "Sea",                    & g_hWaterBrush,      & g_hCoastPen, NULL,                  NULL, c_colWater2,   & g_hcurUnknown,    PrepareLakeLabel, DrawLakeLabel},
	{0x3003c, "Lake",                   & g_hWaterBrush,      & g_hCoastPen, NULL,                  NULL, c_colWater2,   & g_hcurUnknown,    PrepareLakeLabel, DrawLakeLabel},
	{0x3003d, "Lake",                   & g_hWaterBrush,      & g_hCoastPen, NULL,                  NULL, c_colWater2,   & g_hcurUnknown,    PrepareLakeLabel, DrawLakeLabel},
	{0x3003e, "Lake",                   & g_hWaterBrush,      & g_hCoastPen, NULL,                  NULL, c_colWater2,   & g_hcurUnknown,    PrepareLakeLabel, DrawLakeLabel},
	{0x3003f, "Lake",                   & g_hWaterBrush,      & g_hCoastPen, NULL,                  NULL, c_colWater2,   & g_hcurUnknown,    PrepareLakeLabel, DrawLakeLabel},
	{0x30040, "Lake",                   & g_hWaterBrush,      & g_hCoastPen, NULL,                  NULL, c_colWater2,   & g_hcurUnknown,    PrepareLakeLabel, DrawLakeLabel},
	{0x30041, "Lake",                   & g_hWaterBrush,      & g_hCoastPen, NULL,                  NULL, c_colWater2,   & g_hcurUnknown,    PrepareLakeLabel, DrawLakeLabel},
	{0x30042, "Lake",                   & g_hWaterBrush,      & g_hCoastPen, NULL,                  NULL, c_colWater2,   & g_hcurUnknown,    PrepareLakeLabel, DrawLakeLabel},
	{0x30043, "Lake",                   & g_hWaterBrush,      & g_hCoastPen, NULL,                  NULL, c_colWater2,   & g_hcurUnknown,    PrepareLakeLabel, DrawLakeLabel},
	{0x30044, "Lake",                   & g_hWaterBrush,      & g_hCoastPen, NULL,                  NULL, c_colWater2,   & g_hcurUnknown,    PrepareLakeLabel, DrawLakeLabel},
	{0x30046, "River",                  & g_hWaterBrush,      & g_hCoastPen, NULL,                  NULL, c_colWater2,   & g_hcurUnknown,    PreparePolyRiverLabel, DrawPolyRiverLabel},
	{0x30047, "River",                  & g_hWaterBrush,      & g_hCoastPen, NULL,                  NULL, c_colWater2,   & g_hcurUnknown,    PreparePolyRiverLabel, DrawPolyRiverLabel},
	{0x30048, "River",                  & g_hWaterBrush,      & g_hCoastPen, NULL,                  NULL, c_colWater2,   & g_hcurUnknown,    PreparePolyRiverLabel, DrawPolyRiverLabel},
	{0x30049, "River",                  & g_hWaterBrush,      & g_hCoastPen, NULL,                  NULL, c_colWater2,   & g_hcurUnknown,    PreparePolyRiverLabel, DrawPolyRiverLabel},
	{0x3004c, "Intermittent river/lake", & g_hDryWaterBrush,  & g_hDryWaterPen, NULL,               NULL, c_colWater2,   & g_hcurUnknown,    PrepareLakeLabel, DrawLakeLabel},
	{0x3004d, "Glacier",                & g_hGlacierBrush,    & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurUnknown,    PreparePointLabel, DrawLabel},
	{0x30050, "Forest",                 & g_hForestBrush,     & g_hNullPen, NULL,                   NULL, c_colBlack,    & g_hcurUnknown,    PreparePointLabel, DrawLabel},
	{0x30051, "Wetland",                & g_hMarchesBrush,    & g_hNullPen, NULL,                   NULL, c_colWater2,   & g_hcurUnknown,    PrepareLakeLabel, DrawLakeLabel},
	{0x30053, "Island",                 & g_hWhiteBrush,      & g_hCoastPen, NULL,                  NULL, c_colBlack,    & g_hcurUnknown,    PreparePointLabel, DrawLabel},
};

const size_t g_cTypes = sizeof (g_types)/sizeof (g_types [0]);

const type_desc_t * GetTypeDesc (USHORT _shType, USHORT _ushKind) {
	const ULONG ulType = _shType + (ULONG (_ushKind) << 16);
	
	size_t a = 1, b = g_cTypes - 1;
	while (true) {
		const ULONG ulA = g_types [a].ulType;
		if (ulA == ulType)
			return g_types + a;

		const ULONG ulB = g_types [b].ulType;
		if (ulB == ulType)
			return g_types + b;

		if (b - a <= 1)
			// NOTE: 0-th item is "Unknown".
			return g_types + 0;

		const size_t m = (a + b)/2;

		const ULONG ulM = g_types [m].ulType;
		if (ulM < ulType)
			a = m;
		else
			b = m;
	}
}
