# ifndef __Types_H_
# define __Types_H_

//
// Types.h
//
// Declarations of map region types.
//

void InitTypes ();

struct rgn_t;
typedef void (* pfPrepareLabel_t) (rgn_t & _rgn);
typedef void (* pfDrawLabel_t) (HDC _hDC, const rgn_t & _rgn);

extern const char * g_strKinds [];

//
// The metainfo.
//
struct type_desc_t {
	union {
		// Unique key for search.
		ULONG        ulType;

		struct {
			// Garmin type of region.
			USHORT   ushType;
			// 0 - POI (RGN10), 1 - point (RGN20), 2 - polyline (RGN40), 3 - polygon (RGN80).
			USHORT   ushKind;
		};
	};
		// English description of the type.
	const char *     strTypeName;

		// The brush to fill polygons.
	HBRUSH *         phBrush;

		// The pens to draw polylines .
	HPEN *           phPen;
		// The pen for second pass. NOTE: NULL means not to draw 2nd pass.
	HPEN *           phPen2;

		// The font and color for labels. NOTE: NULL means default font.
	HFONT *          phFont;
	COLORREF         colText;

		// The icon for point/POI. NOTE: can't be NULL; use g_hcurUnknown for default.
	HCURSOR *        phCursor;

		// The function used to prepare region for label output.
	pfPrepareLabel_t pfPrepareLabel;
		// The function used for label output. NOTE: can't be NULL; use DrawLabel for default.
	pfDrawLabel_t    pfDrawLabel;

	// TODO: function to draw complex lines (power line, bridge, etc)

	// Hot spot of phCursor.
	int nCursorShiftX, nCursorShiftY;

	// Hatched bush for transparent polygon mode.
	HBRUSH           hbrTransparentPolygon;
};

	// The type descriptions.
	// NOTE: g_types[0] is reserved for all unknown types.
extern type_desc_t g_types [];
	// Total number of types.
extern const size_t g_cTypes;

	// Search for type.
	// NOTE: if not found, g_types[0] is returned (which corresponds to "Unknown"). 
const type_desc_t * GetTypeDesc (USHORT _ushType, USHORT _ushKind);


//
// Shared graphics objects.
//

	// The default icon for unknown types.
extern HCURSOR g_hcurUnknown;

	// The default icon for waypoint.
extern HCURSOR g_hcurWaypoint;
	// The hot spot of it.
extern int g_nCursorShiftX_Waypoint, g_nCursorShiftY_Waypoint;

	// The pen for attached tracks.
extern HPEN g_hTrackPen;

	// The pen & brush used to outline selection.
extern HPEN g_hSelectionPen;
extern HBRUSH g_hSelectionBrush;


# endif // __Types_H_