# include "stdafx.h"
# include "UndoRedo.h"
# include "Menu.h"
# include "MapView.h"
# include "Position.h"
# include "Globals.h"

bool g_bModified = false;

actions_t g_Undo;
actions_t g_Redo;

bool CanUndo () {return ! g_Undo.empty ();}
bool CanRedo () {return ! g_Redo.empty ();}
bool CanSave () {return g_bModified && g_bTypeCanBeSaved;} 

void ResetUndoRedo () {
	actions_t::iterator i;
	for (i = g_Undo.begin (); i != g_Undo.end (); ++ i)
		delete * i;
	g_Undo.clear ();

	for (i = g_Redo.begin (); i != g_Redo.end (); ++ i)
		delete * i;
	g_Redo.clear ();
}

bool PerformEditAction (CEditAction * _pAction) {
	const bool bCanUndo = _pAction->CanUndo ();
	if (! bCanUndo && ! g_Undo.empty ()) {
		if (::MessageBox (g_hWnd, "The action cannot be undone.\nDo you want to proceed?", "Confirmation", MB_YESNO | MB_ICONQUESTION) == IDNO) {
			delete _pAction;
			return false;
		}
	}

	_pAction->m_Selection = g_Selection;
	if (! _pAction->Do ()) {
		::ErrorMsgBox ("The action failed.");

		delete _pAction;
		return false;
	}
	g_bModified = true;

	if (bCanUndo) {
		// Reset Redo stack.
		for (actions_t::iterator i = g_Redo.begin (); i != g_Redo.end (); ++ i)
			delete * i;
		g_Redo.clear ();

		g_Undo.push_back (_pAction);
	} else {
		ResetUndoRedo ();
		delete _pAction;
	}

	Refresh ();
	UpdateMenu ();

	return true;
}

void UndoEditAction () {
	if (g_Undo.empty ())
		return;

	CEditAction * const pAction = g_Undo.back ();
	g_Undo.pop_back ();

	pAction->Undo ();
	g_Selection = pAction->m_Selection;

	g_Redo.push_back (pAction);

	// TODO: ? restore the flag ?
	g_bModified = true;
}

void RedoEditAction () {
	if (g_Redo.empty ())
		return;

	CEditAction * const pAction = g_Redo.back ();
	g_Redo.pop_back ();

	pAction->m_Selection = g_Selection;
	pAction->Redo ();

	g_Undo.push_back (pAction);

	g_bModified = true;
}

