# ifndef __UndoRedo_H_
# define __UndoRedo_H_

# include "Selection.h"

//
// The base class for all map changes.
//
class CEditAction {
  public:
	// Append action description to _str.
	virtual void GetName (std::string & _str) const = 0;

	virtual bool CanUndo () const = 0;
	// Return true if success.
	virtual bool Do () = 0;

	virtual void Undo () = 0;
	virtual void Redo () = 0;

	// Depth of nesting of combined actions.
	// NOTE: ordinary actions should not override this.
	virtual int GetDepthLevel () const {return 0;}

	// The data used to restore state after undo/redo.
	find_info_m_t m_Selection;
};

//
// The undo/redo stack.
//
typedef std::list<CEditAction *> actions_t;
extern actions_t g_Undo;
extern actions_t g_Redo;

extern bool g_bModified;

bool PerformEditAction (CEditAction * _pAction);
void UndoEditAction ();
void RedoEditAction ();
void ResetUndoRedo ();

bool CanUndo ();
bool CanRedo ();
bool CanSave ();

# endif // __UndoRedo_H_