# include "stdafx.h"
# include "UndoRedo.h"
# include "Map.h"
# include "Types.h"
# include "MapView.h"
# include "Position.h"
# include "Globals.h"

class CVectorize : public CEditAction {
  public:
	CVectorize (points_t & _points) : m_pRgn (NULL) {m_points.swap (_points);}

	virtual void GetName (std::string & _str) const;
	virtual bool CanUndo () const;
	virtual bool Do ();
	virtual void Undo ();
	virtual void Redo ();

  private:
	points_t m_points;
	rgn_t * m_pRgn;
};

void CVectorize::GetName (std::string & _str) const {
	_str.append ("vectorize here");
}

bool CVectorize::CanUndo () const {return true;}

bool CVectorize::Do () {
	CCritSec::CLock lock (g_cs);

	// Create new rgn in layer 2 (polylines).
	g_map.layers [2].rgns.push_back (rgn_t ());
	++ g_map.layers [2].cRgns;

	rgn_t & rgn = g_map.layers [2].rgns.back ();

	g_map.InsertRgnToSaveList (rgn);

	// Initialize the rgn.
	rgn.cBeginPos = -1;
	rgn.cEndPos = -1;
	rgn.cFileLine = -1;
	rgn.cOrdinal = -1;
	rgn.pFileRef = NULL;

	rgn.strLabel = "";

	rgn.pTypeDesc = & g_types [0];
	rgn.ushKind = 2;

	// Create the element of the rgn.
	rgn.elements.push_back (rgn_element_t ());
	rgn_element_t & element = rgn.elements.back ();

	// Initialize the element of rgn.
	element.cFileLine = -1;
	element.cFileOffset = -1;
	element.cLevel = 0;
	element.points.swap (m_points);

	for (size_t cPoint = 0; cPoint < element.points.size (); ++ cPoint)
		element.rectBound.Extend (element.points [cPoint]);
	element.fLength = GetLength (element.points);
	element.fArea = 0;
	if (g_map.cMaxPointsInElement < element.points.size ())
		g_map.cMaxPointsInElement = element.points.size ();

	rgn.rectBound = element.rectBound;
	rgn.fSize2 = GetRgnSize (rgn);

	// TODO: initialize the rest...

	m_pRgn = & rgn;

	return true;
}

void CVectorize::Undo () {
	CCritSec::CLock lock (g_cs);

	assert (m_pRgn);
	++ g_map.layers [2].cDeletedRgns;
	m_pRgn->dwFlags |= rgn_t::fDeleted;
}

void CVectorize::Redo () {
	CCritSec::CLock lock (g_cs);

	assert (m_pRgn);
	m_pRgn->dwFlags &= ~rgn_t::fDeleted;
	-- g_map.layers [2].cDeletedRgns;
}

///////////////////////////////////////////////////////

float DColor (const BYTE * _pPoint1, const BYTE * _pPoint2) {
	const float fDRed   = _pPoint1 [0] - _pPoint2 [0];
	const float fDGreen = _pPoint1 [1] - _pPoint2 [1];
	const float fDBlue  = _pPoint1 [2] - _pPoint2 [2];
	return ::sqrt (fDRed*fDRed + fDGreen*fDGreen + fDBlue*fDBlue)/1.7f/255.f;
	//return (::fabs (fDRed) + ::fabs (fDGreen) + ::fabs (fDBlue))/3.f/255.f;
}

float DColor (COLORREF _col1, COLORREF _col2) {
	const float fDRed   = GetRValue (_col1) - GetRValue (_col2);
	const float fDGreen = GetGValue (_col1) - GetGValue (_col2);
	const float fDBlue  = GetBValue (_col1) - GetBValue (_col2);
	return ::sqrt (fDRed*fDRed + fDGreen*fDGreen + fDBlue*fDBlue)/1.7f/255.f;
	//return (::fabs (fDRed) + ::fabs (fDGreen) + ::fabs (fDBlue))/3.f/255.f;
}

COLORREF AveColor (const BYTE * _pBits, size_t _cScanWidth, size_t _cHeight, size_t _x, size_t _y) {
	float r = 0, g = 0, b = 0;

	const size_t cA = 1;
	for (size_t x = _x - cA; x <= _x + cA; ++ x)
		for (size_t y = _y - cA; y <= _y + cA; ++ y) {
			const BYTE * const pPoint = _pBits + x*3 + (_cHeight - 1 - y)*_cScanWidth;
			r += pPoint [0];
			g += pPoint [1];
			b += pPoint [2];
		}

	return RGB (r/(1 + cA*2), g/(1 + cA*2), b/(1 + cA*2));
}

float AveIntensity (const BYTE * _pBits, size_t _cScanWidth, size_t _cHeight, size_t _x, size_t _y) {
	float fRet = 0;

	const size_t cA = 2;
	for (size_t x = _x - cA; x <= _x + cA; ++ x)
		for (size_t y = _y - cA; y <= _y + cA; ++ y) {
			const BYTE * const pPoint = _pBits + x*3 + (_cHeight - 1 - y)*_cScanWidth;
			fRet += pPoint [0] + pPoint [1] + pPoint [2];
		}

	return fRet;
}

void Vectorize (long _lX, long _lY, const image_t * _pImage) {
	assert (_pImage);

	// Get coordinates of the start point.
	const rect_t & rect = _pImage->rect;
	const long lX0 = XToScreen (rect.x0);
	const long lX1 = XToScreen (rect.x1);
	const long lY0 = YToScreen (rect.y0);
	const long lY1 = YToScreen (rect.y1);
	const size_t cWidth  = _pImage->lWidth;
	const size_t cHeight = _pImage->lHeight;
	const size_t cImageX = cWidth*(_lX - lX0)/(lX1 - lX0);
	const size_t cImageY = cHeight*(_lY - lY1)/(lY0 - lY1);

	// Access source bits to perform projection's transform.
	BITMAPINFOHEADER bmih;
	::memset (& bmih, 0, sizeof (bmih));
	bmih.biSize = sizeof (bmih);
	bmih.biWidth = cWidth;
	bmih.biHeight = cHeight;
	bmih.biPlanes = 1;
	bmih.biBitCount = 24;
	bmih.biCompression = BI_RGB;
	bmih.biSizeImage = 0;
	bmih.biClrUsed = 0;
	bmih.biClrImportant = 0;

	const size_t cScanWidth = Align4 (3*cWidth);
	BYTE * const pBits = new BYTE [cScanWidth*cHeight];
	if (pBits == NULL)
		return;
	CByteArrayHolder pBits_Holder (pBits);
	::GetDIBits (_pImage->hMemDC, _pImage->hBitmap, 0, cHeight, pBits, reinterpret_cast <BITMAPINFO *> (& bmih), DIB_RGB_COLORS);

	float x = cImageX, y = cImageY;
	points_t points;

# define GET_POINT(_x,_y) (pBits + (_x)*3 + (cHeight - _y)*cScanWidth)

//	const BYTE * const pPixel0 = GET_POINT (cImageX, cImageY);

	const COLORREF colAve = AveColor (pBits, cScanWidth, cHeight, cImageX, cImageY);

	// Get sigma x.
	float t;
	float tx1_best = 0;
	for (t = 0; t < 10; t += 1) {
		const float xt = x - t;
		const float yt = y;
		const COLORREF colT = AveColor (pBits, cScanWidth, cHeight, ::fast_floor (xt + .5), ::fast_floor (yt + .5));
		const float d = DColor (colAve, colT);
		if (d > 0.3) {
			tx1_best = t;
			break;
		}
	}
	float tx2_best = 0;
	for (t = 0; t < 10; t += 1) {
		const float xt = x + t;
		const float yt = y;
		const COLORREF colT = AveColor (pBits, cScanWidth, cHeight, ::fast_floor (xt + .5), ::fast_floor (yt + .5));
		const float d = DColor (colAve, colT);
		if (d > 0.2) {
			tx2_best = t;
			break;
		}
	}
	const float sx = tx1_best + tx2_best;

	// Get sigma y.
	float ty1_best = 0;
	for (t = 0; t < 10; t += 1) {
		const float xt = x;
		const float yt = y - t;
		const COLORREF colT = AveColor (pBits, cScanWidth, cHeight, ::fast_floor (xt + .5), ::fast_floor (yt + .5));
		const float d = DColor (colAve, colT);
		if (d > 0.3) {
			ty1_best = t;
			break;
		}
	}
	float ty2_best = 0;
	for (t = 0; t < 10; t += 1) {
		const float xt = x;
		const float yt = y + t;
		const COLORREF colT = AveColor (pBits, cScanWidth, cHeight, ::fast_floor (xt + .5), ::fast_floor (yt + .5));
		const float d = DColor (colAve, colT);
		if (d > 0.2) {
			ty2_best = t;
			break;
		}
	}
	const float sy = ty1_best + ty2_best;

	// Sigma*2
	const float s = 1/::sqrt (1/(sx*sx) + 1/(sy*sy));

	float a = 0;

	for (size_t c = 0; c < 1000 && x < cWidth && y < cHeight; ++ c) {
		// Save the point.
		point_t point;
		point.x = rect.x0 + x*rect.Width  ()/cWidth;
		point.y = rect.y1 - y*rect.Height ()/cHeight;
		points.push_back (point);

		// Search the next point.
		const size_t cX = fast_floor (x + .5);
		const size_t cY = fast_floor (y + .5);
/*
		const float R = s*5;

		size_t cc = 0;
		bool bOK = false;
		for (float t = - 2; t <= 2; t += 1/R) {
			const float xt = x + R*::cos (a + t);
			const float yt = y + R*::sin (a + t);
			const COLORREF colT = AveColor (pBits, cScanWidth, cHeight, ::fast_floor (xt + .5), ::fast_floor (yt + .5));
			const float d = DColor (colAve, colT);
			if (d <= 0.1) {*/
/*				++ cc;
			} else
				cc = 0;

			if (cc >= s) {*/
/*				bOK = true;

//				t -= cc*.5/R;
				x += R*::cos (a + t);
				y += R*::sin (a + t);
				a = t;

				break;
			}
		}
		if (! bOK)
			break;
*/		
		const float fI00 = AveIntensity (pBits, cScanWidth, cHeight, cX, cY);
		const float fI10 = AveIntensity (pBits, cScanWidth, cHeight, cX + 3, cY);
		const float fI01 = AveIntensity (pBits, cScanWidth, cHeight, cX, cY + 3);

		float fDx = fI10 - fI00;
		float fDy = fI01 - fI00;
		const float fDn = ::sqrt (fDx*fDx + fDy*fDy);
		if (fDn == 0)
			break;
		fDx /= fDn;
		fDy /= fDn;

		x += 3*fDy;
		y -= 3*fDx;
/*
		float d_min = 2;
		float t_best = 0;
		for (float t = -10; t <= 10; t += 1) {
			const float xt = x + fDx*t;
			const float yt = y + fDy*t;

			const BYTE * const pPixelT = GET_POINT (fast_floor (xt + .5), fast_floor (yt + .5));
			const float d = DColor (pPixelT, pPixel0);
			if (d < d_min) {
				d_min = d;
				t_best = t;
			}
		}
		if (d_min > .05)
			break;

		x += fDx*t_best;
		y += fDy*t_best;*/
	}

	PerformEditAction (new CVectorize (points));
}